
#include "termios.h"
#include "sys/time.h"

#include "ARToolKitPlus/TrackerSingleMarkerImpl.h"
#include <map>
#include "GestioneConfigurazione.h"
#include "MatrixFunctions.h"
#include "Matrix3DRotoTraslation.h"
#include "Param3DRotoTraslation.h"
#include "VideoAcquisitionWheelchair.h"
#include "FileRawAcquisition.h"
#include "VisionSystemCalibHeader.h"

using namespace std;


void calcMatrixAndWriteFile(int markerId,ARToolKitPlus::TrackerSingleMarker* tracker,FILE *fout){
	Matrix3DRotoTraslation *matrix=new Matrix3DRotoTraslation();
	if(markerId>=0){
		printf("found marker %d\n",markerId);
		bool allZero=true;
		for(int j=0; j<3; j++){//faccio solo le prime tre righe, la quarta è [0 0 0 1]
			for(int i=0; i<4; i++){
				matrix->setValue(j,i,tracker->getModelViewMatrix()[i*4+j]);
				if(!(i==3 && j==3)){
					if(fabs(tracker->getModelViewMatrix()[i*4+j])>1e-6){
						//printf("artoolkit::allzero=false\n");
						allZero=false;
					}
				}
			}
		}
		//se allZero=true, la matrice non è valida! (anche se markerId è >=0 capita)
		if(!allZero){
			//matrice valida
			Param3DRotoTraslation *p=toParam3DRotoTraslation(matrix);
			fprintf(fout,"%d\t%.16f\t%.16f\t%.16f\t%.16f\t%.16f\n",markerId,p->getZ(),p->getSinB(),p->getCosB(),p->getSinC(),p->getCosC());
			delete p;
		}
	}
	delete matrix;
}

int createRelationFile(){
	FILE *fout=fopen(GestioneConfigurazione::getMatrixInversionParameterFileTmp().c_str(),"w");
	if(fout==NULL){
		cout<<"ERRORE APERTURA FILE "<<GestioneConfigurazione::getMatrixInversionParameterFileTmp().c_str()<<endl;
		return -1;
	}

	const int     width = 640, height = 480;
	/*size_t        numPixels = width*height*bpp;
	size_t        numBytesRead;*/
	struct timeval tv,tv2,r;
	struct timezone tz;
	bool exit=false;
	unsigned int count=0;
	char c;

	MyLogger      logger;
	ARToolKitPlus::TrackerSingleMarker* tracker = new ARToolKitPlus::TrackerSingleMarkerImpl<6,6,6, 1, 8>(width,height);

	//inizializzazione tracker
	tracker->setLogger(&logger);
	tracker->setPixelFormat(ARToolKitPlus::PIXEL_FORMAT_LUM);
	if(!tracker->init(GestioneConfigurazione::getArtoolkitCalibFile().c_str(), 1.0f, 1000.0f)){
		printf("ERROR: artoolkit init\n");
		delete tracker;
		return -1;	
	}
	// define size of the marker
	tracker->setPatternWidth(GestioneConfigurazione::getArtoolkitTagDimension());
	tracker->setBorderWidth( 0.250f);
	tracker->activateAutoThreshold(true);
	tracker->setUndistortionMode(ARToolKitPlus::UNDIST_LUT);
	tracker->setPoseEstimator(ARToolKitPlus::POSE_ESTIMATOR_RPP);
	tracker->setMarkerMode( ARToolKitPlus::MARKER_ID_SIMPLE);
	
	VideoAcquisitionWheelchair * camera;
	camera=new FileRawAcquisition(GestioneConfigurazione::getImageNamePrefix(),GestioneConfigurazione::getImageNameSubfix(),GestioneConfigurazione::getImageNameFormat());

	camera->Start();
	set_keypress();
	unsigned char * img;
	gettimeofday(&tv,&tz);
	while(!exit){
		c=getchar();
		if(c==27)exit=true;
		img=camera->CaptureFrame();
		if(img==NULL){
			printf("END!\n");
			exit =true;
		}
		if(!exit){
			ARToolKitPlus::ARMarkerInfo* markerInfo;
			int numMarkers;
			//così trova il marker migliore
			int markerId = tracker->calc(img,-1,true, &markerInfo,&numMarkers);
			calcMatrixAndWriteFile(markerId,tracker,fout);

			//ora tratto gli altri marker trovati
			vector<int> tagIdentificati;
			for(int j = 0; j < numMarkers; j++){
				if(markerInfo[j].id!=-1){
					bool exist=false;
					for(int i=0;i<tagIdentificati.size();i++){
						if(tagIdentificati[i]==markerInfo[j].id){
							exist=true;
							break;
						}
					}
					if(!exist && markerInfo[j].id!=markerId){
						printf("altro id trovato: %d\n",markerInfo[j].id);
						tagIdentificati.push_back(markerInfo[j].id);
					}
				}
			}
			
			for(int i=0;i<tagIdentificati.size();i++){
				//ricerca esplicita del tag
				int markerId = tracker->calc(img,tagIdentificati[i]);
				calcMatrixAndWriteFile(markerId,tracker,fout);
			}

			
						
			printf("COUNT=%d\n",count);
			count++;
		}
		
	}
	gettimeofday(&tv2,&tz);//temporizzazione del ciclo di controllo
	timersub(&tv2,&tv,&r);
	printf("tempo trascorso: %u.%.6u\n",r.tv_sec,r.tv_usec);
	double fps=((double)count)/((double)((double)r.tv_sec+(double)r.tv_usec/(double)1000000.0));
	printf("fps: %lf\n",fps);

	reset_keypress();

	fclose(fout);
	camera->Stop();
	delete tracker;
	delete camera;
	return 0;	
}


void mediaRelationFile(){
	map<int,double> zm;
	map<int,double> sbm;
	map<int,double> cbm;
	map<int,double> sgm;
	map<int,double> cgm;
	double tmp;
	map<int,double>  count;
	int index;

	FILE *f=fopen(GestioneConfigurazione::getMatrixInversionParameterFileTmp().c_str(),"r");

	while(!feof(f)){
		fscanf(f,"%d",&index);
		if(feof(f))break;
		//printf("read %d\n",index);
		map<int,double>::iterator iter = count.find(index);
		if( iter == count.end() ) {
			count[index]=0;
			zm[index]=0;
			sbm[index]=0;
			cbm[index]=0;
			sgm[index]=0;
			cgm[index]=0;
		}
		fscanf(f,"%lf",&tmp);
		zm[index]+=tmp;
		fscanf(f,"%lf",&tmp);
		sbm[index]+=tmp;
		fscanf(f,"%lf",&tmp);
		cbm[index]+=tmp;
		fscanf(f,"%lf",&tmp);
		sgm[index]+=tmp;
		fscanf(f,"%lf",&tmp);
		cgm[index]+=tmp;
		count[index]=count[index]+1;
		
	
	}	
	fclose(f);

	FILE *fout=fopen(GestioneConfigurazione::getMatrixInversionParameterFile().c_str(),"w");
	if(f==NULL){
		printf("Error opening file %s\n",GestioneConfigurazione::getMatrixInversionParameterFile().c_str());

	}

	for(map<int,double>::iterator i=count.begin();i!=count.end();i++){
		if(count[i->first]>0){
			fprintf(fout,"tagNumber \t = %d\n",i->first);
			fprintf(fout,"tagZeta \t = %.16f\n",zm[i->first]/count[i->first]);
			fprintf(fout,"tagBeta \t = %.16f\n",atan2(sbm[i->first]/count[i->first],cbm[i->first]/count[i->first]));
			fprintf(fout,"tagGamma \t = %.16f\n\n",atan2(sgm[i->first]/count[i->first],cgm[i->first]/count[i->first]));
			printf("tagNumber \t = %d\n",i->first);
			printf("tagZeta \t = %.16f\n",zm[i->first]/count[i->first]);
			printf("tagBeta \t = %.16f\n",atan2(sbm[i->first]/count[i->first],cbm[i->first]/count[i->first]));
			printf("tagGamma \t = %.16f\n\n",atan2(sgm[i->first]/count[i->first],cgm[i->first]/count[i->first]));
		}
	}
	printf("end\n\n");
	fclose(fout);

}

int main(int argc,char *argv[]){

	if(GestioneConfigurazione::init("config/config.ini",argc,argv)!=0)exit(0);
	cout<<"analisi immagini..."<<endl;
	if(createRelationFile()==0){
		string s;
		s="cp "+GestioneConfigurazione::getMatrixInversionParameterFile()+" "+ GestioneConfigurazione::getMatrixInversionParameterFile() +".old";
		cout<<s<<endl;
		system(s.c_str());
		mediaRelationFile();
	}

	


	return 0;
}

