/**********************************************
 * DCDT                       September 2002  *  
 *                                            *
 * interface.cpp                              *
 *                                            *
 * originally developed by M.Matteucci        *
 * matteucc@airlab.elet.polimi.it             *
 **********************************************/


#include "interface.h"

Interface::Interface (DCDT_Agora* dis)
  :DCDT_Member(dis,100000)
{
}

Interface::~Interface()
{
}

void
Interface::Init()
{
  // Section to request messages on ethnos
  // you have to put here the msg's type
  // you are interested in
  
  SubscribeMsgType(MSG_FROM_PINGER);
  //   [...]

  // End section to request messages

  int argc=0;        // You can use argc and argv to set
  char **argv=NULL;  // QApplication parameters [qt-doc] 

  QApplication a(argc,argv);  // Create the QT application
  Panel = new QInterface(this,0,"QI"); // Create the QInterface Object
  a.setMainWidget(Panel); // Set QInterface Object as main widget
  Panel->show(); // Show the widget
  a.exec();  // Execute the QT application
}

void
Interface::DoYourJob (int par)
{             
  // The DoYourJob cycle is an infinite loop whaiting 
  // the preemption of DCDT_Member. During this time the
  // controll is given to the QT application "Panel"

  while(!par)   
    {;}       
}        

void
Interface::Close()
{
  printf("\n Closing Interface");fflush(stdout);
}

/**

The following code is intended as a schema to produce
a qt application inside a DCDT_Member, using its methods.
   
This application send a message MSG_FROM_QINTERFA to 
a simple Pinger expert that reply with MSG_FROM_PINGER.
An internal counter is used to trace and display the 
number of recived messages.
   
To Implement your personal application insert your code
and use DCDT_Member methods by calling: QTMember->method()

**/
                                   
QInterface::QInterface(Interface *MyMember,QWidget *parent,const char *nome)
  : QWidget(parent,nome)
{
  QTMember=MyMember; // HERE is the trick to use DCDT_Member methods
  
  NumMsg=0; // memory of recived messages

  setBackgroundColor(lightGray);  // set background color
  resize(200,90);		  // resize this widget
  
  TopLayout = new QVBoxLayout(this,15);  // create a layout to manage widgets positioning

  Label = new QLabel("",this);  // Create the label to display the messages arrived from pinger
  Label->setFrameStyle(QFrame::Panel|QFrame::Sunken); // set the frame style
  Label->setFont(QFont("Fixed",20,80,0)); // set the label font
  Label->setAlignment(AlignCenter); //set the label alignment
  Label->setNum(NumMsg); // display the number of messages arrived
  Label->setMinimumSize(Label->sizeHint()); // set the minimum size as the actual vertical
  TopLayout->addWidget(Label,10);           // size of the label and add it to the layout

  Buttons = new QHBoxLayout();
  TopLayout->addLayout(Buttons);

  Send = new QPushButton("Send !",this);
  Send->setMinimumSize(Send->sizeHint());  
  connect(Send,SIGNAL(clicked()),SLOT(SendClicked()));
  Buttons->addWidget(Send,10);

  End = new QPushButton("End !",this);
  End->setMinimumSize(End->sizeHint());  
  connect(End,SIGNAL(clicked()),SLOT(EndClicked()));
  Buttons->addWidget(End,10);

  TopLayout->activate();

  timer = new QTimer(this);
  connect(timer,SIGNAL(timeout()),this,SLOT(timerEvent()));
  timer->start(0,FALSE);        
}



void QInterface::SendClicked()
{
  DCDT_Msg *Msg = QTMember->CreateMsg(MSG_FROM_QINTERFA);
  QTMember->ShareMsg(Msg);
}

void QInterface::EndClicked()
{
  QTMember->Shutdown(0);    
  //  ETControlMsg *Msg = new ETControlMsg(ENDSESSION);
  //  QTMember->ShareMsg(Msg,ETPRIO_MIN);
}

void QInterface::timerEvent()
{
  DCDT_Msg *rxMsg;
  
  while( (rxMsg = QTMember->ReceiveMsg()))
    {
      //     const DCDT_Msg *rxMsg = QTMember->ReceiveMsg();
      if ( (rxMsg->ReadType() == MSG_FROM_PINGER) )
	{
	  NumMsg++;
	  Label->setNum(NumMsg);
	}
      QTMember->RemoveCurrMsg();
    }    
}





