/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_Channel.h

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_CHANNEL_H
#define DCDT_CHANNEL_H

#include <DCDT_Defs.h>
#include <DCDT_Msg.h>

class DCDT_Msg;

/*! \brief Virtual class that describes a generic communication channel
 */
class DCDT_Channel {
 public:
  DCDT_Channel(int AgoraID);
  virtual ~DCDT_Channel();
  virtual void Open(int link_status = 0) = 0;
  virtual void Reopen(int link_status) = 0;
  /*! \brief Prepare the channel's data structure before opening the link
   *
   *  This function is responsible for the allocated area of the two classes passed as parameters\n
   */
  virtual void Prepare(CommData *local, CommData *remote) = 0;
  virtual void ChooseChannel(int channel) = 0;
  virtual void WaitConn() = 0;
  virtual void StartConn() = 0;
  virtual void Send(const DCDT_Msg *msg) = 0;
  virtual void SendNotify() = 0;
  virtual void SendAnswer(CommData* to, CommData* local) = 0;
  virtual DCDT_Msg* Receive() = 0;
  virtual void ReceiveHS(HSMsgHeader*& header, CommData*& remote) = 0;
  virtual void Dispose() = 0;
  virtual void Close() = 0;
  virtual void KeepAlive() = 0;

  /** Returns a heap memory area of which is no more responsible */
  virtual CommData* GetStartingCD() = 0;
  virtual unsigned int ReadStartingProfile() = 0;
  virtual CommData* GetCommData() = 0;
  virtual void SetTimers() = 0;
  virtual void SetLostTimer(int usec) = 0;
  virtual void LostTimerOn() = 0;
  virtual void LostTimerOff() = 0;
  virtual void SetConnTimer(int usec) = 0;
  virtual void ConnTimerOn() = 0;
  virtual void ConnTimerOff() = 0;
  virtual void SetSendTimer(int usec) = 0;
  virtual void SendTimerOn() = 0;
  virtual void SendTimerOff() = 0;
  virtual void UnblockSend() = 0;
  virtual void SetReceiveTimer(int usec) = 0;
  virtual void ReceiveTimerOn() = 0;
  virtual void ReceiveTimerOff() = 0;
  virtual void UnblockReceive() = 0;
  virtual void Restart(int link_status) = 0;
  virtual void Stop() = 0;

  /*! \brief Generic channel error */
  class ChannelErr
  {
  public:
    int errval;
    ChannelErr() { errval = 0; };
    ChannelErr(int err) { errval = err; };
  };

  /*! \brief Timeout experied */
  class TimeOut : public ChannelErr {};

 protected:
  HSMsgHeader notify, answer, recvd;
  int channel_type;
};


#endif
