/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_ChannelIP.h

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_CHANNEL_IP_H
#define DCDT_CHANNEL_IP_H

#include <UDPSocket.h>
#include <TCPSocket.h>
#include <DCDT_Channel.h>
#include <DCDT_ctrace.h>

#define IP_LOST_TIMER_VAL 10000000
#define IP_SEND_TIMER_VAL  1000000
#define IP_RECV_TIMER_VAL   100000
#define IP_CONN_TIMER_VAL 10000000

#define SNDLOWAT_VAL 500
#define SNDBUF_VAL 10*SNDLOWAT_VAL
#define MSS_VAL 500

typedef struct _Data_IP {
  int tcp_addr, udp_addr;
  int tcp_port, udp_port;
} Data_IP;
#define DATA_IP_LEN sizeof(Data_IP)

class CommData_IP;

class DCDT_Channel_IP : public DCDT_Channel {
 public:
  DCDT_Channel_IP(int AgoraID);
  virtual ~DCDT_Channel_IP();

  inline void Open(int link_status = 0);
  inline void Reopen(int link_status);
  inline void Prepare(CommData *local, CommData *remote);
  inline void ChooseChannel(int channel){channel=0;};
  inline void DynamicPrepare(int local_addr, int mc_addr, short port);
  inline void StaticPrepare(unsigned int prof, char *local_addr, short local_port, char *remote_addr, short remote_port);
  inline void WaitConn();
  inline void StartConn();
  inline void Send(const DCDT_Msg *msg);
  inline void SendNotify();
  inline void SendAnswer(CommData* to, CommData* local);
  inline DCDT_Msg* Receive();
  inline void ReceiveHS(HSMsgHeader*& header, CommData*& remote);
  inline void Dispose();
  inline void Close();
  inline void KeepAlive(){};
  inline CommData* GetCommData();
  inline CommData* GetStartingCD();
  inline unsigned int ReadStartingProfile();
  inline void SetTimers();
  inline void SetLostTimer(int usec);
  inline void LostTimerOn();
  inline void LostTimerOff();
  inline void SetConnTimer(int usec);
  inline void ConnTimerOn();
  inline void ConnTimerOff();
  inline void SetSendTimer(int usec);
  inline void SendTimerOn();
  inline void SendTimerOff();
  inline void UnblockSend();
  inline void SetReceiveTimer(int usec);
  inline void ReceiveTimerOn();
  inline void ReceiveTimerOff();
  inline void UnblockReceive();
  void Restart(int link_status);
  void Stop();

 private:
  bool
    /// Flag: 1 -> dynamic channel  !=1 static channel
    dynamic,
    /// Flag: 1 -> UDP message received but not processed
    udp_ready;

  unsigned int profile;
  UDPSocket *udp_sock;
  TCPSocket 
    /// Socket used in LISTEN state 
    *tcp_sock1,
    /// Socket created with accept()
    *tcp_sock2,
    /// Actual socket in use can be tcp_sock1 or tcp_sock2
    *tcp_wsock;

  int multi_addr, local_addr;
  short multi_port;
  CommData_IP 
    /// tcp/udp address and port used for new sockets
    *startCD,
    /// local ip address and port
    *localCD,
    /// remote ip address and port
    *remoteCD;

  /// Last message received 
  DCDT_Msg *msg;
  int fd_tcp, fd_udp;
  fd_set set;
};

class CommData_IP : public CommData {
public:
  Data_IP* data;

  CommData_IP();
  CommData_IP(int t_a, int t_p, int u_a, int u_p);
  CommData_IP(Data_IP* d);
  CommData_IP(CommData_IP *cd);
  virtual ~CommData_IP();

  CommData_IP& operator=(const CommData_IP& cd) {
    data->tcp_addr = cd.data->tcp_addr;
    data->tcp_port = cd.data->tcp_port;
    data->udp_addr = cd.data->udp_addr;
    data->udp_port = cd.data->udp_port;
    return (*this);
  }

  DCDT_Channel* CreateChannel(int AgoraID) {
    return (new DCDT_Channel_IP(AgoraID));
  }

};

/*! \brief Prepare the channel's data structure before opening the link.
 *
 *  Saves ip address ( tcp and udp ) and port of the local and remote end of the channel.
 */
inline void DCDT_Channel_IP::Prepare(CommData *local, CommData *remote)
{
  TRC_PRINT ( DCDT_TRC_COMM, TRC1, ("Prepare channel IP") );
  if (local) {
    if (localCD)
      delete localCD;
    localCD = new CommData_IP((CommData_IP*)local);

    TRC_PRINT ( DCDT_TRC_COMM, TRC1, ("localCD: tcp_addr=%x tcp_port=%hx udp_addr=%x udp_port=%hx", localCD->data->tcp_addr, localCD->data->tcp_port, localCD->data->udp_addr, localCD->data->udp_port));
  } else
    localCD = new CommData_IP();// (???)

  if (remote) {
    if (remoteCD)
      delete remoteCD;
    remoteCD = new CommData_IP((CommData_IP*)remote);

    TRC_PRINT ( DCDT_TRC_COMM, TRC1, ("remoteCD: tcp_addr=%x tcp_port=%hx udp_addr=%x udp_port=%hx", remoteCD->data->tcp_addr, remoteCD->data->tcp_port, remoteCD->data->udp_addr, remoteCD->data->udp_port));
  } else
    remoteCD = new CommData_IP();// (???)

}

/*! \brief Prepare the Finder's channel that opens dynamic links.
 *
 * Opens the UDP/multicast socket that the Finder uses to talk with others agora. The receiving mode for the socket is no blocking.
 */
inline void DCDT_Channel_IP::DynamicPrepare(int l_addr, int mc_addr, short mc_port)
{
  dynamic = true;
  profile = MPM_SYSTEM | MPM_LINK;

  if ( udp_sock )
    delete udp_sock;
  
  udp_sock = new UDPSocket();
  udp_sock->Open();
  udp_sock->MC_Set((multi_addr = mc_addr), local_addr = l_addr);

  udp_sock->Bind(0, (multi_port = mc_port));

  localCD = new CommData_IP();
  localCD->data->udp_addr = local_addr;
  localCD->data->udp_port = mc_port;

  TRC_PRINT ( DCDT_TRC_COMM, TRC1, ("UDP localCD: addr=%x port=%hx", localCD->data->udp_addr, localCD->data->udp_port));

  udp_sock->SetReceiveTimer(0);
  udp_sock->ReceiveTimerOn();
  udp_sock->SetLostTimer(IP_LOST_TIMER_VAL);
  udp_sock->LostTimerOn();
  udp_sock->SetConnTimer(0);
  udp_sock->ConnTimerOn();

  startCD = new CommData_IP();
  startCD->data->tcp_addr = startCD->data->udp_addr = local_addr;
  startCD->data->tcp_port = startCD->data->udp_port = 0;

}

/*! \brief Prepare the Finder's channel used to open static links.
 *
 * Open an UDP socket, bind it to the local address and port then allocate the structures used to create a static link. The receiving mode for the socket is no blocking. 
 */
inline void DCDT_Channel_IP::StaticPrepare(unsigned int prof, char *l_addr, short l_port, char *r_addr, short r_port)
{
  dynamic = false;
  profile = prof;

  // Local port
  localCD = new CommData_IP();
  inet_pton(AF_INET, l_addr, &localCD->data->udp_addr);
  local_addr = localCD->data->udp_addr;
  localCD->data->udp_port = l_port;

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("UDP localCD: addr=%x port=%hx", localCD->data->udp_addr, localCD->data->udp_port));

  // Remote port
  remoteCD = new CommData_IP();
  inet_pton(AF_INET, r_addr, &remoteCD->data->udp_addr);
  remoteCD->data->udp_port = r_port;

  udp_sock = new UDPSocket();
  udp_sock->Open();
  udp_sock->Bind(localCD->data->udp_addr, (short)localCD->data->udp_port);

  udp_sock->SetReceiveTimer(0);
  udp_sock->ReceiveTimerOn();
  udp_sock->SetLostTimer(IP_LOST_TIMER_VAL);
  udp_sock->LostTimerOn();
  udp_sock->SetConnTimer(0);
  udp_sock->ConnTimerOn();

  startCD = new CommData_IP();
  startCD->data->tcp_addr = startCD->data->udp_addr = localCD->data->udp_addr;
  startCD->data->tcp_port = startCD->data->udp_port = 0;
}

/*! \brief Open a new tcp and udp socket.\n
 *
 *  Open a tcp and an udp socket, send a test message on the udp socket.\n
 */
inline void DCDT_Channel_IP::Open(int link_status )
{
  short port;
  DCDT_Msg m(1,1);

  tcp_sock1 = new TCPSocket();
  tcp_sock1->Open();
  cout << "Bind to: " << localCD->data->tcp_addr << " port: " << localCD->data->tcp_port;
  try {
    tcp_sock1->Bind(localCD->data->tcp_addr, (short)localCD->data->tcp_port);
  } catch ( ChannelError e ) {
    cout << "Bind: ChannelError " << e.errval << "\n";
  }
  tcp_sock1->SetBufs(0, (int)SNDBUF_VAL/2, 0, SNDLOWAT_VAL);
  //sock1->SetSegSize(MSS_VAL);

  if (link_status == L_WAITING) {
    try {
      tcp_sock1->Listen();
    } catch ( ChannelError e ) {
      cout << "Listen: ChannelError " << e.errval << "\n";
    }
  }
  tcp_sock1->GetCommData(localCD->data->tcp_addr, port);
  localCD->data->tcp_port = (int)port;

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("TCP socket bound to: addr=%x port=%hx", localCD->data->tcp_addr, localCD->data->tcp_port));
  tcp_wsock = tcp_sock1;

  udp_sock = new UDPSocket();
  udp_sock->Open();
  udp_sock->Bind(localCD->data->udp_addr, (short)localCD->data->udp_port);
  try {
    udp_sock->Send(&m, "127.0.0.1", 1234);
  } catch ( ChannelError e ) {
    cout << "Bind: ChannelError " << e.errval << "\n";
  }
  udp_sock->GetCommData(localCD->data->udp_addr, port);
  localCD->data->udp_port = (int)port;
    /* } catch ( ChannelError e ) {
    cout << "Bind: ChannelError " << e.errval << "\n";
    }*/

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("UDP socket bound to: addr=%x port=%hx", localCD->data->udp_addr, localCD->data->udp_port));
}

/*! \brief Reopen a channel, used after a call to Stop()
 */
inline void DCDT_Channel_IP::Reopen(int link_status)
{
  short port;
  DCDT_Msg m(1,1);

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Reopen channel status=%d",link_status));
  tcp_sock1->Open();
  tcp_sock1->Bind(localCD->data->tcp_addr, (short)localCD->data->tcp_port);
  tcp_sock1->SetBufs(0, (int)SNDBUF_VAL/2, 0, SNDLOWAT_VAL);
  //sock1->SetSegSize(MSS_VAL);

  if (link_status == L_WAITING)
    tcp_sock1->Listen();
  tcp_sock1->GetCommData(localCD->data->tcp_addr, port);
  localCD->data->tcp_port = (int)port;
  tcp_wsock = tcp_sock1;

  udp_sock->Open();
  udp_sock->Bind(localCD->data->udp_addr, (short)localCD->data->udp_port);
  udp_sock->Send(&m, "127.0.0.1", 1234);
  udp_sock->GetCommData(localCD->data->udp_addr, port);
  localCD->data->udp_port = (int)port;

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("UDP socket bound to: addr=%x port=%hx", localCD->data->udp_addr, localCD->data->udp_port));
}

/*! Return the local communication channel informations */
inline CommData* DCDT_Channel_IP::GetCommData()
{
  CommData_IP *ret;
  ret = new CommData_IP ( localCD );
  return ret;
}

/*! Return the informations needed to really open an IP channel, this informations are the local address and port for the new channel */
inline CommData* DCDT_Channel_IP::GetStartingCD()
{
  CommData_IP *ret;
  // Communications data ( ip address / port ) 
  ret = new CommData_IP( startCD );
  return ret;
}

/*! Return the profile for this channel */
inline unsigned int DCDT_Channel_IP::ReadStartingProfile()
{
  return profile;
}

/*! \brief Method called when the link is in WAITING state.
 *
 *  Accepts a new connection on the tcp socket, then receive a message and if the message is MT_CONNECT save the connection informations and reconnect the udp socket to the just connected agora.
 */
inline void DCDT_Channel_IP::WaitConn()
{
  DCDT_Msg *m;
  Data_IP *d;

  // Accept a new connection and receive a message
  tcp_sock2 = tcp_wsock->Accept();
  tcp_wsock = tcp_sock2;
  tcp_wsock->ReceiveTimerOn();
  tcp_wsock->SetReceiveTimer(IP_CONN_TIMER_VAL);
  m = tcp_wsock->Receive();
  tcp_wsock->ReceiveTimerOff();

  // If the message received is MT_CONNECT 
  // reconnect the udp socket to the just connected agora
  if (m->ReadType() == MT_CONNECT) {
    d = (Data_IP*)m->GetPayload();

    if (remoteCD) {
      remoteCD->data->udp_addr = d->udp_addr;
      remoteCD->data->udp_port = d->udp_port;
    }
    else
      remoteCD = new CommData_IP(d->tcp_addr, d->tcp_port, d->udp_addr, d->udp_port);

    udp_sock->Connect(d->udp_addr, d->udp_port);
  }

  delete m;
  tcp_wsock->SetReceiveTimer(IP_RECV_TIMER_VAL);
  fd_tcp = tcp_wsock->GetFd();
  fd_udp = udp_sock->GetFd();
}

/*! \brief Method called when the link is in CONNECTING state.
 *
 *  Connect the tcp_wsock and udp_socket sockets to the remote end of the channel then send a MT_CONNECT message throught the tcp channel.\n
 */
inline void DCDT_Channel_IP::StartConn()
{
  // tcp
  tcp_wsock->Connect(remoteCD->data->tcp_addr, remoteCD->data->tcp_port);

  // Alloc a new message, at the end of the function the message will be
  // deleted; the payload has to be dynamically allocated
  DCDT_Msg m(MT_CONNECT, 1);
  m.AllocPayload(sizeof(Data_IP), localCD->data);

  // udp
  udp_sock->Connect(remoteCD->data->udp_addr, remoteCD->data->udp_port);

  tcp_wsock->Send(&m);
  fd_tcp = tcp_wsock->GetFd();
  fd_udp = udp_sock->GetFd();

}

/** \brief Receive a message ( blocking )
 *
 *  If an UDP message was already received return it, otherwise tries to receive a messages via TCP or UDP. If at the same time we receive two messages, one TCP and the other UDP we have a problem because we can return only a message at a time so we set the udp_ready flag and return the TCP message so the next time we can return the UDP message.\n
 * \exception TimeOut Timeout during operation
 * \exception ConnError Communication error see errval for details
 * \returns NULL - No message received\n
 * != NULL - Pointer to the message received
 */
inline DCDT_Msg* DCDT_Channel_IP::Receive()
{
  msg = NULL;


  TRC_PRINT ( DCDT_TRC_COMM, TRC1, ("startCD %x:%x",startCD?startCD->data->tcp_addr:0, startCD?startCD->data->tcp_port:0));

  if (udp_ready) {
    udp_ready = false;
    msg = udp_sock->Receive();
  } else {
    FD_ZERO(&set);
    FD_SET(fd_tcp, &set);
    FD_SET(fd_udp, &set);
    if (fd_tcp > fd_udp) {
      switch (select(fd_tcp+1, &set, NULL, NULL, NULL))
      {
        case 0: throw TimeOut();
        case -1: throw ConnError(errno);
      }
    } else {
      switch (select(fd_udp+1, &set, NULL, NULL, NULL))
      {
        case 0: throw TimeOut();
        case -1: throw ConnError(errno);
      }
    }
    if (FD_ISSET(fd_tcp, &set)) {
      msg = tcp_wsock->Receive();
      if (FD_ISSET(fd_udp, &set))
        udp_ready = true;
    } else if (FD_ISSET(fd_udp, &set))
      msg = udp_sock->Receive();
  }

  return msg;
}

/*! \brief Receive an hand-shake message.
 */
inline void DCDT_Channel_IP::ReceiveHS(HSMsgHeader*& header, CommData*& remote)
{
  void* data;
  TRC_PRINT ( DCDT_TRC_COMM, TRC1, ("startCD %x:%x",startCD->data->tcp_addr, startCD->data->tcp_port));
  udp_sock->ReceiveHS(header, data);

  remote = new CommData_IP((Data_IP*)data);
}

/*! \brief Send a message, if the message has TCP_WARRANTY set sent it throught the tcp socket otherwise use the udp one.
 *
 *  \exception ChannelError TCP/UDP socket not opened
 */
inline void DCDT_Channel_IP::Send(const DCDT_Msg *msg)
{
  TRC_PRINT ( DCDT_TRC_COMM, TRC1, ("startCD %x:%x",startCD?startCD->data->tcp_addr:0, startCD?startCD->data->tcp_port:0));
  if (msg->ReadDeliveryWarranty() == TCP_WARRANTY) {
    if (!tcp_wsock)
       throw ChannelError();
    tcp_wsock->Send(msg);
  }
  else {
    if (!udp_sock)
       throw ChannelError();
    udp_sock->Send(msg);
  }
}

inline void DCDT_Channel_IP::SendNotify()
{
  TRC_PRINT ( DCDT_TRC_COMM, TRC1, ("Sending HS_NOTIFY"));
  notify.channel = channel_type;
  if (dynamic)
    udp_sock->SendHS(multi_addr, multi_port, &notify, localCD->data);
  else udp_sock->SendHS(remoteCD->data->udp_addr, remoteCD->data->udp_port, &notify, localCD->data);
}

inline void DCDT_Channel_IP::SendAnswer(CommData* to, CommData* local)
{
  TRC_PRINT ( DCDT_TRC_COMM, TRC1, ("Sending HS_ANSWER"));
  answer.channel = channel_type;
  udp_sock->SendHS(((CommData_IP*)to)->data->udp_addr, ((CommData_IP*)to)->data->udp_port, &answer, ((CommData_IP*)local)->data);
}

inline void DCDT_Channel_IP::Dispose()
{
}

inline void DCDT_Channel_IP::Close()
{
  if (tcp_sock1)
    tcp_sock1->Close();
  if (tcp_sock2)
    tcp_sock2->Close();
  tcp_wsock = NULL;
  if (udp_sock)
    udp_sock->Close();
}

inline void DCDT_Channel_IP::SetTimers()
{
  udp_sock->SetLostTimer(IP_LOST_TIMER_VAL);
  udp_sock->LostTimerOn();
  udp_sock->SetConnTimer(IP_CONN_TIMER_VAL);
  udp_sock->ConnTimerOn();
  udp_sock->SetSendTimer(IP_SEND_TIMER_VAL);
  udp_sock->SendTimerOn();
  udp_sock->SetReceiveTimer(IP_RECV_TIMER_VAL/2);
  udp_sock->ReceiveTimerOn();

  tcp_wsock->SetLostTimer(IP_LOST_TIMER_VAL);
  tcp_wsock->LostTimerOn();
  tcp_wsock->SetConnTimer(IP_CONN_TIMER_VAL);
  tcp_wsock->ConnTimerOn();
  tcp_wsock->SetSendTimer(IP_SEND_TIMER_VAL);
  tcp_wsock->SendTimerOn();
  tcp_wsock->SetReceiveTimer(IP_RECV_TIMER_VAL/2);
  tcp_wsock->ReceiveTimerOn();
}

inline void DCDT_Channel_IP::SetLostTimer(int usec)
{
  udp_sock->SetLostTimer(usec);
  if (tcp_wsock)
    tcp_wsock->SetLostTimer(usec);
}

inline void DCDT_Channel_IP::LostTimerOn()
{
  udp_sock->LostTimerOn();
  if (tcp_wsock)
    tcp_wsock->LostTimerOn();
}

inline void DCDT_Channel_IP::LostTimerOff()
{
  udp_sock->LostTimerOff();
  if (tcp_wsock)
    tcp_wsock->LostTimerOff();
}

inline void DCDT_Channel_IP::SetConnTimer(int usec)
{
  udp_sock->SetConnTimer(usec);
  if (tcp_wsock)
    tcp_wsock->SetConnTimer(usec);
}

inline void DCDT_Channel_IP::ConnTimerOn()
{
  udp_sock->ConnTimerOn();
  if (tcp_wsock)
    tcp_wsock->ConnTimerOn();
}

inline void DCDT_Channel_IP::ConnTimerOff()
{
  udp_sock->ConnTimerOff();
  if (tcp_wsock)
    tcp_wsock->ConnTimerOff();
}

inline void DCDT_Channel_IP::SetSendTimer(int usec)
{
  udp_sock->SetSendTimer(usec);
  if (tcp_wsock)
    tcp_wsock->SetSendTimer(usec);
}

inline void DCDT_Channel_IP::SendTimerOn()
{
  udp_sock->SendTimerOn();
  if (tcp_wsock)
    tcp_wsock->SendTimerOn();
}

inline void DCDT_Channel_IP::SendTimerOff()
{
  udp_sock->SendTimerOff();
  if (tcp_wsock)
    tcp_wsock->SendTimerOff();
}

inline void DCDT_Channel_IP::UnblockSend()
{
  udp_sock->UnblockSend();
  if (tcp_wsock)
    tcp_wsock->UnblockSend();
}

inline void DCDT_Channel_IP::SetReceiveTimer(int usec)
{
  udp_sock->SetReceiveTimer(usec);
  if (tcp_wsock)
    tcp_wsock->SetReceiveTimer(usec);
}

inline void DCDT_Channel_IP::ReceiveTimerOn()
{
  udp_sock->ReceiveTimerOn();
  if (tcp_wsock)
    tcp_wsock->ReceiveTimerOn();
}

inline void DCDT_Channel_IP::ReceiveTimerOff()
{
  udp_sock->ReceiveTimerOff();
  if (tcp_wsock)
    tcp_wsock->ReceiveTimerOff();
}

inline void DCDT_Channel_IP::UnblockReceive()
{
  udp_sock->UnblockReceive();
  if (tcp_wsock)
    tcp_wsock->UnblockReceive();
}

inline void DCDT_Channel_IP::Restart(int link_status)
{
  Open(link_status);
}

inline void DCDT_Channel_IP::Stop()
{
  if (tcp_sock1)
    tcp_sock1->ForcedClose();
  if (tcp_sock2)
    tcp_sock2->ForcedClose();
  tcp_wsock = NULL;
  if (udp_sock)
    udp_sock->ForcedClose();
}


#endif

