/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_Defs.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                    Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_DEFS_H
#define DCDT_DEFS_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <iostream>
#include <errno.h>

// Use Boost for thread implementation
//#define USE_BOOST

// Use timeout when checking for new messages
//#define USE_TIMEOUT

#define NANOSLEEP_VERSION
//#define SEQUENCER_VERSION

// definition required only in nanosleep version
// to correct the timing bias introduced by the
// nanosleep() call

//in case HZ=1280
//#define NANOBIAS 781
//in case HZ=100
//#define NANOBIAS 10000
//In case of kernel 2.6
#define NANOBIAS 0
 
//Variables and costants used during the test's part

#define READCYCLE        10
#define WRITECYCLE       10 
#define READWRITECYCLE   10

#define NUMMEMBERFORMAIN 3
// Note that  NUMMEMBERFORMAIN = NUMREADER + NUMWRITER + NUMREADERWRITER 
#define NUMREADER       10
#define NUMWRITER       10
#define NUMREADERWRITER 10 


#define READINIT       11 
#define READEND        12
#define READDONEEND    15
#define READLOCKINIT   13
#define READLOCKEND    14

#define WRITEINIT      21
#define WRITEEND       22
#define WRITELOCKINIT  23
#define WRITELOCKEND   24

#define DELETEINIT     31
#define DELETEEND      32
#define DELETELOCKINIT 33
#define DELETELOCKEND  34

#define BRIDGEINIT     41
#define BRIDGEEND      42
#define BRIDGELOCKINIT 43
#define BRIDGELOCKEND  44

#define WAITINIT     51
#define WAITEND      52
 
// General definitions
#define SUCCESS    0
#define FAILURE   -1

// RunLevel definitions
#define DCDT_ZERO              0
#define DCDT_INIT              1
#define DCDT_LETSWORK          2
#define DCDT_SHUTDOWN          3  
#define DCDT_TERMINATING       4
#define DCDT_DOWN              5

// Member states
#define EMPTYSLOT     0
#define CREATED       1
#define ACTIVE        2
#define INITIALIZING  3
#define READY         4
#define RUNNING       5
#define SUSPENDED     6
#define DEACTIVATED   7
#define TERMINATING   8


#define UCHAR unsigned char
#define UINT  unsigned int
#define LLONG long long
#define DCDT_TIME unsigned int

// InnerLink first agora socket name
#define INNER_STARTNAME "/tmp/dcdt_startsock"
#define INNER_NAME "/tmp/dcdt_sock"
#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX 108
#endif


// link status types
#define L_WAITING      1
#define L_CONNECTING   2
#define L_RESETTOCONN  3
#define L_WORKING      4
#define L_LOST         5
#define L_RESTARTING   6
#define L_STOPPING     7
#define L_KILL         8

// timer di attesa del link sul canale
#define LINK_RESTART_NUM      3
#define LINK_LOST_NUM         2 // Was 5
#define LINK_LOST_TIMER       2000000

// usec di intervallo tra un notify e l'altro
#define FINDER_PERIOD   500000
#define NOTIFY_PERIOD   5000000

#define MAXLINKS 500
#define MAXCHANNELS 10

// Member Profile Masks
#define ProfileType unsigned int

#define MPM_SYSTEM        1
#define MPM_USER          2
#define MPM_LINK          4
#define MPM_BRIDGE        8
#define MPM_STATIC       16

// ***** Messages section *****

// number of seconds to wait before removing unread msgs from list
#define TIMEOUT_MSG 10


class DCDT_Channel;

/** Encapsulate all the date that belongs to a communication endpoint */
class CommData {
 public:
  CommData();
  virtual ~CommData();
  virtual DCDT_Channel* CreateChannel(int AgoraID) = 0;
};


#define SIZEOF_INTEGER 32

// Defines for the message types:
// remember that the max number of messages types must be:
// MAX_MSG_TYPE = 32 * NUMINT_MSGTYPE
// We use a bit for every message type ( ex. for subscription mask )
#define NUMINT_MSGTYPE  4
#define MAX_MSG_TYPE    (NUMINT_MSGTYPE*SIZEOF_INTEGER)

// Defines for the Agorà members:
// remember that the max number of members must be:
// MAX_MEMBERS = 32 * NUMINT_MEMBERS
// We use a bit for every member ( ex. for readers mask )
#define NUMINT_MEMBERS  6
#define MAX_MEMBERS     (NUMINT_MEMBERS*SIZEOF_INTEGER)

#ifdef USE_BOOST

#include <boost/utility.hpp>
#include <boost/thread/condition.hpp>
#include <boost/thread/thread.hpp>

class DCDT_Condition{
 public:
  DCDT_Condition() {
    lk = new boost::mutex::scoped_lock(myMutexCond,false);
  };
  
  ~DCDT_Condition() {
    delete lk;
  };

  inline void wait() {
    myCond.wait(*lk);
  };

  inline void signal() {
    myCond.notify_one();
  };

  inline void lock() {
    lk->lock();
  };

  inline void unlock() {
    lk->unlock();
  };

 private:
  boost::condition myCond;
  boost::mutex myMutexCond;
  boost::mutex::scoped_lock *lk;
};

class DCDT_Mutex{
 public:
  DCDT_Mutex() {
    lk = new boost::mutex::scoped_lock(myMutexCond,false);
  };
  
  ~DCDT_Mutex() {
    //delete lk;
  };

  inline void lock() { lk->lock(); };

  inline void unlock() { lk->unlock(); };

 private:
  boost::mutex myMutexCond;
  boost::mutex::scoped_lock *lk;
};
#else

#include <pthread.h>

class DCDT_Condition{
  public:
  DCDT_Condition() {
    pthread_cond_init( &myCond, NULL);
    pthread_mutex_init ( &myMutexCond, NULL );
  };
  ~DCDT_Condition() {
    pthread_cond_destroy( &myCond);
  };

  inline void wait() {
    pthread_cond_wait(&myCond, &myMutexCond);

  };
  inline void signal(){
    pthread_cond_signal( &myCond);
  };
  inline void lock() {
    pthread_mutex_lock( &myMutexCond);

  };
  inline void unlock(){
    pthread_mutex_unlock( &myMutexCond);
  };

  private:
  pthread_cond_t myCond;
  pthread_mutex_t myMutexCond;
  };

class DCDT_Mutex {
  public:
  DCDT_Mutex() {
    pthread_mutexattr_init( &attr );
    pthread_mutexattr_settype( &attr, PTHREAD_MUTEX_ERRORCHECK_NP );
    pthread_mutex_init( &myMutex, &attr );
  };
  ~DCDT_Mutex() {
    pthread_mutex_destroy( &myMutex );
  };

  inline void lock() {
    if ( pthread_mutex_lock( &myMutex ) == EDEADLK ) {
       char *mio=NULL;
       *mio=0x00;
    }

  };
  inline void unlock(){
    pthread_mutex_unlock( &myMutex );
  };

  private:
  pthread_mutex_t myMutex;
  pthread_mutexattr_t attr;

  };
#endif

// CTrace definitions
#define UNIT_MAX            100
#define DCDT_TRC_AGORA        1
#define DCDT_TRC_POSTOFFICE   2
#define DCDT_TRC_INNERLINK    3
#define DCDT_TRC_MEMBER       4
#define DCDT_TRC_COMM         5
#define DCDT_TRC_LINKRX       6

// Exceptions
/** Generic error */
class Exception {};

class ChannelError : public Exception {
public:
  int errval;
  ChannelError() { errval = 0; };
  ChannelError(int err) { errval = err; };
};

class ConnError : public ChannelError {
public:
  ConnError() : ChannelError() {};
  ConnError(int err) : ChannelError(err) {};
};

class TimeOut : public Exception {};

typedef enum { DCDT_LOCAL_MSG=1, DCDT_REMOTE_MSG=2, DCDT_ALL_MSG=3 } DCDT_RequestType;

#endif //define
