/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_InnerLinkManager.cpp

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef INNERLINKMAN_CPP
#define INNERLINKMAN_CPP

using namespace std;

#include <DCDT_InnerLinkManager.h>
#include <DCDT_Agora.h>
#include <DCDT_ctrace.h>

DCDT_InnerLinkManager::DCDT_InnerLinkManager(DCDT_Agora* agora) : DCDT_Member(agora)
{
  InnerChannel = new DCDT_InnerChannel(myAgoraID);
  InnerStartCD = new InnerCommData(INNER_STARTNAME);
  InnerCD = new InnerCommData(INNER_NAME);

  hs_header = NULL;
  hs_cd = NULL;
  innerlink_counter = 0;
  localCD  = remoteCD = NULL;
  name = NULL;
  name_cd = NULL;
}

DCDT_InnerLinkManager::~DCDT_InnerLinkManager()
{
  /* ??? Usare un metodo diverso da delete per eliminare l'InnerLinkChannel
         da' un warning in compilazione se dichiaro il distruttore
  */
  delete InnerChannel;
  delete InnerStartCD;
  //delete InnerCD;
  delete hs_header;
  delete hs_cd;

  //localCD->Dispose();
  //delete localCD;

  //remoteCD->Dispose();
  //delete remoteCD;
  delete name;

  delete name_cd;
}

/*! \brief Tries to find others Agorae on the local system.\n
 *
 *  \return 0 - This is the first Agor, starting an inner listening channel\n
 *  1 - This isn't the first Agor, star an inner link to the first Agor
 */
int DCDT_InnerLinkManager::LookForExistingAgora(int &agoraID)
{
  try {
    if ( remove(InnerStartCD->name) != 0 ) {
      if ( errno != ENOENT )
	cout << "Unable to remove '" << InnerStartCD->name << "' - " << errno <<"\n";
    }

    try {
      InnerChannel->StartingPrepare(InnerStartCD, INNER_NAME);
      InnerChannel->Open();
    } catch ( ConnError e ) {
      cout << "Error preparing InnerSocket\n";
    }
    
    try {
      InnerChannel->SendNotify();
    } catch (ConnError e) {
      // no socket on this machine or nobody linked to it: this is the first Agor
      remove(INNER_NAME);
      return 0;
    }
    InnerChannel->SetReceiveTimer(ILM_RECV_TIMER);
    InnerChannel->ReceiveTimerOn();
    try {
      InnerChannel->ReceiveHS(hs_header, hs_cd);
    } catch (TimeOut t) {
      // nobody answered: this becomes the first Agor
      remove(INNER_NAME);
      return 0;
    }

    // this is not the first Agor, setting up the inner link
    InnerChannel->Close();

    agoraID = hs_header->AgoraID;
    delete localCD;
    localCD = (InnerCommData*)hs_cd; // name assigned to the couple of sockets
    cout << "\n" << localCD->name;
    remoteCD = new InnerCommData(localCD);
    cout << "\n" << remoteCD->name;
    ++innerlink_counter;
    Delay(500000); // necessary ???
    CreateLink(L_CONNECTING);
    return 1;
  } catch (ChannelError e) {
    cout << "\nException! errno=" << e.errval << "\n";
    // problem with the inner channel, starting without it
    InnerChannel->Close();
    return 1;
  }
}

/*! \brief Preparing the listening inner channel
 */
void DCDT_InnerLinkManager::Init()
{
 try {
  InnerChannel->Close();
  InnerChannel->Prepare(InnerCD, NULL);
  InnerChannel->Open();
  InnerChannel->ReceiveTimerOff();
 } catch (ChannelError e) {
  cout << "\nInit InnerLinkManager exception! The InnerChannel is broken";
  cout << "\nerrno=" << e.errval << "\n";
  myAgora->RemoveMember(this);
 }
}


/*! \brief Listen to internal msgs and reacts opening inner links.
 *
 * The InnerChannel waits until a new message was received then opens a new link.
 */
void DCDT_InnerLinkManager::DoYourJob(int par)
{
  try {
    InnerChannel->ReceiveHS(hs_header, hs_cd);

    if (hs_header) {
      switch (hs_header->type) {

        case HS_NOTIFY:
	  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("InnerNotify received from Agor n.%i", ++innerlink_counter));

          name_cd = CreateName(innerlink_counter);

	  // Create a new LinkRx
          CreateLink(L_WAITING);
          InnerChannel->SendAnswer(hs_cd, name_cd);
          delete name_cd;
          name_cd = NULL;
          break;

        case HS_ANSWER:
        // error!

        default:
          cout << "\nInnerLinkManager received unknown or wrong msg!";
      }
      delete hs_header;
      delete hs_cd;
      hs_header = NULL;
      hs_cd = NULL;
    }
  } catch (Exception e) {
    cout << "\nInnerLinkManager exception! The InnerChannel is broken";
    myAgora->RemoveMember(this);
  }
}

/*! \brief Add a new LinkRx to our Agora.
 *
 * Link ID = 1000 + innerlink_counter
 */
void DCDT_InnerLinkManager::CreateLink(int status)
{
  //ProfileType profile = MPM_SYSTEM | MPM_BRIDGE;
  ProfileType profile = MPM_SYSTEM | MPM_LINK;

  SetCDs(status);

  DCDT_LinkRx *newLink;
  newLink = new DCDT_LinkRx(myAgora, 1000+innerlink_counter, 0);
  myAgora->AddSysMember(newLink, profile);
  newLink->Prepare(status, localCD, remoteCD);

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Link created!"));
  myAgora->ActivateMember(newLink);
}

/*! \brief Create the socket name for the InnerLink.
 *
 *  The name is of this form: INNER_NAME+innerlink_counter.
 *  \return Pointer to the new communication data structure
 */
InnerCommData* DCDT_InnerLinkManager::CreateName(int count)
{
  InnerCommData *cd;
  char *name, c=0;
  
  switch (count) {
  case 1 :
    c = '1';
    break;
  case 2 :
    c = '2';
    break;
  case 3 :
    c = '3';
    break;
  case 4 :
    c = '4';
    break;
  case 5 :
    c = '5';
    break;
  case 6 :
    c = '6';
    break;
  case 7 :
    c = '7';
    break;
  case 8 :
    c = '8';
    break;
  case 9 :
    c = '9';
    break;
  }
  
  name = new char[strlen(INNER_NAME)+2];
  strcpy(name, INNER_NAME);
  name[strlen(INNER_NAME)] = c;
  name[strlen(INNER_NAME)+1] = '\0';
  
  cd = new InnerCommData(name);
  
  localCD = new InnerCommData(cd);
  remoteCD = new InnerCommData(cd);
  
  return cd;
}

/*! \brief Set the socket's names ( must be different because stand on the same machine )
 */
void DCDT_InnerLinkManager::SetCDs(int status)
{
  if (status == L_WAITING) {
    strcat(localCD->name, "s");
    strcat(remoteCD->name, "c");
  } else {
    strcat(localCD->name, "c");
    strcat(remoteCD->name, "s");
  }
}

#endif
