/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_LinkTx.cpp

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef LINKTX_CPP
#define LINKTX_CPP

using namespace std;

#include <DCDT_LinkTx.h>
#include <DCDT_LinkRx.h>
#include <DCDT_Finder.h>
#define LINKTX_PERIOD 100000


static void InstallSignalHandler ( void )
{
   struct sigaction sa;
   __sigset_t sigset;
                                                                                
   // Ignore SIGPIPE from sockets
   sa.sa_handler = SIG_IGN;
   memset(&sigset,0,sizeof(sigset));;
   sa.sa_mask = sigset;
   sa.sa_flags = 0;
   sa.sa_restorer = NULL;
   if ( sigaction(SIGPIPE,&sa,NULL) != 0 ) {
      printf ( "Error: Unable to ignore SIGPIPE\n" );
      exit ( 1 );
   }
}

DCDT_LinkTx::DCDT_LinkTx(DCDT_Agora *agora, int id, int r_id, DCDT_LinkRx *ptr) : DCDT_Member(agora)
{
  linkID = id;
  remoteID = r_id;
  status = restart_counter = 0;
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    mymask[i] = 0;
  msgrcv = NULL;
  msgsub = CreateMsg(MT_SUBSUPD);
  msgsub->SetDeliveryWarranty(TCP_WARRANTY);
  msgsub->SetPayload(subs_mask, NUMINT_MSGTYPE*sizeof(unsigned int));
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    subs_mask[i] = 0;
  LinkRxPtr = ptr;
  InstallSignalHandler();
  force_mask=true;
};


void DCDT_LinkTx::Init()
{
}

void DCDT_LinkTx::Close()
{
   // ???
   // Don't free the payload it wasn't dynamically allocated ( see LinkTx.h:68 )
   // We can't "delete msgsub" because it will free payload; we have to change
   // the payload pointer the delete msgsub but payload is private
   //msgsub->header->payload_len=0;
   //msgsub->payload=NULL;
   //delete msgsub;
}

void DCDT_LinkTx::Prepare(DCDT_Channel *ch)
{
  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Preparing LINKTX"));
  channel = ch;
  status = L_WORKING;
}


/*! \brief FSM machine for the LinkTx member
 *
 *  Initial state is always L_WORKING.
 *  - L_WORKING:
 *     - Check subscription mask, if was changed send an update message
 *     - Check if there are new messages to be sent
 *     - Periodically send KeepAlive messages
 *  - L_RESTARTING: Our LinkRx said to close and reopen the link
 *  - L_LOST: The communication is lost wait until our LinkRx does something
 *  
 *  In any state if an exception occours ( TimeOut, ConnError or ChannelError ) the state is always set to L_LOST
 */
void DCDT_LinkTx::DoYourJob(int par)
{
 try {
  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkTx %i status %i", linkID, status));
  switch (status) {

    case L_WORKING:

      // KeepAlive must be sent periodically
      channel->KeepAlive();
      TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LINKTX %i WORKING", linkID));

      // Read the new suscribers mask, and store it in subs_mask
      // if the suscribers mask was changed send the new values 
      if ( force_mask == true ) {
         TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Forced MsgSub Member: %d", linkID));
         force_mask=false;
         channel->Send(msgsub);
      }

      if (myPostOffice->ReadSubscribedMsgMask(subs_mask,ReadID())) {
        TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Sending MsgSub Member: %d", linkID));
        channel->Send(msgsub);
      }

      // Sending messages 
      while ((msgrcv = ReceiveMsg( false ))) {
	TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LINKTX %i read local msg type %i", linkID, msgrcv->ReadType()));
        channel->Send(msgrcv);
	// ANCHE QUI VA GESTITA UNA PERIODICIT ( ??? )

      }
      Delay(LINKTX_PERIOD);
      break;

    case L_LOST:
      Delay(LINK_LOST_TIMER);
      break;

    case L_RESTARTING:
      channel = restarting_ch;
      status = L_WORKING;
  }
 } catch(TimeOut e) {
   cout << "\nTimeout on linktx " << linkID << "\nConnection lost";
   TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkTx %i TimeOut", linkID));
   LinkRxPtr->ChangeStatus(L_STOPPING);
   UnSubscribeAll();
   status = L_LOST;
   force_mask=true;
 } catch(ConnError e) {
   cout << "\nError on linktx " << linkID << "\nerrno=" << e.errval << "\nConnection lost";
   TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkTx %i ConnError", linkID));
   LinkRxPtr->ChangeStatus(L_STOPPING);
   UnSubscribeAll();
   status = L_LOST;
   force_mask=true;
 } catch(ChannelError e) {
   cout << "\nError on linktx " << linkID << "\nerrno=" << e.errval << "\nConnection lost";
   TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkTx %i ChannelError", linkID));
   LinkRxPtr->ChangeStatus(L_STOPPING);
   UnSubscribeAll();
   status = L_LOST;
   force_mask=true;
 } catch(Exception e) {
   cout << "\nException! Closing linktx " << linkID << "\n";
   TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkTx %i Exception", linkID));
   LinkRxPtr->ChangeStatus(L_STOPPING);
   UnSubscribeAll();
   status = L_LOST;
   force_mask=true;
   Close();
 }
}

/*! \brief Subscribe this member to the new mask messages */
void DCDT_LinkTx::Subscribe(unsigned int *mask)
{
  int i,j,dim = 8*sizeof(unsigned int), filter;
  for (i=0; i<NUMINT_MSGTYPE; i++) {
    filter = 1;
    for (j=1; j<=dim; j++) {
      if ((mask[i] & filter) > 0) {
	TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Subscribed to type %i", j+i*dim));
        SubscribeMsgType(j+i*dim);
      }
      filter <<= 1;
    }
  }
}

void DCDT_LinkTx::Restart(DCDT_Channel *ch)
{
  restarting_ch = ch;
  status = L_RESTARTING;
}

#endif
