/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_LinkTx.h

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef LINKTX_H
#define LINKTX_H

#include <DCDT_Member.h>
#include <DCDT_Channel.h>

class DCDT_LinkRx;

/*! \brief Member that can transmit on a channel.
 *
 *  This member implements only channel policies for the message transmission, the real transmission was implements in the DCDT_Channel specified by \a channel.This member has three main goals:
 *  - Send local messages to the remote Agora
 *  - Send the subscription mask for the local Agora to the remote one
 *  - Send KeepAlive messages 
 */
class DCDT_LinkTx : public DCDT_Member {
  friend class DCDT_LinkRx;
 public:
  DCDT_LinkTx(DCDT_Agora *agora, int id, int r_id, DCDT_LinkRx *ptr);
  ~DCDT_LinkTx() {};

  void Prepare(DCDT_Channel* ch);
  void Init();
  void DoYourJob(int par = 0);
  void Close();
  void Subscribe(unsigned int *mask);
  void Restart(DCDT_Channel* ch);
  void ChangeStatus(int newstat) { status = newstat; };
  
 private:
  int linkID, remoteID;
  int creation_status, status, restart_counter, lost_counter;

  /// Class that implements the communication functions
  DCDT_Channel *channel,
    /// Channel to be used after a restart
    *restarting_ch;
  unsigned int mymask[NUMINT_MSGTYPE];
  int i;
  const DCDT_Msg *msgrcv;
  DCDT_Msg *msgsub;

  /// Subscriber mask for this Agora, when changes send a message to the remote one
  unsigned int subs_mask[NUMINT_MSGTYPE];
  /// When true subscription mask is sent to the remote end; needed when restarting a channel
  bool force_mask;
  DCDT_LinkRx *LinkRxPtr;

};


#endif //define
