/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

TCPSocket.h

COPYRIGHT (C) 2002    Paolo Meriggi (meriggi@ing.unibs.it) 
                      Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                      Cristian Giussani (cgiussani@fastflow.it)


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_MEMBER_H
#define DCDT_MEMBER_H

#ifdef USE_BOOST
// Using Boost Library
#include <boost/thread/detail/config.hpp>
#include <boost/thread/thread.hpp>

#include <string.h>

struct BoostThread;
#else
//Using pthread
#include <pthread.h>
#endif

using namespace std;

#include <iostream>
#include <string.h>
#include <DCDT_Time.h>
#include <DCDT_Defs.h>
#include <iostream>
#include <DCDT_Msg.h>
#include <DCDT_PostOffice.h>
#include <DCDT_ctrace.h>

class DCDT_Agora;
class DCDT_Finder;

/*! \brief Virual class that defines an Agora member
 */
class DCDT_Member {
  friend class DCDT_Agora;

 public:
  DCDT_Member( DCDT_Agora *, DCDT_TIME period); // Periodical members prototype
  DCDT_Member( DCDT_Agora *);  // Cyclic members prototype
  
  virtual ~DCDT_Member();
  
  inline DCDT_Member*  GetMember(int myID);
  
  inline void *LifeCycle();

  inline void SetID( int ID);
  inline int ReadID();
  inline bool ReadExcededFlag();
  inline void ClearExcededFlag();

  inline bool CheckProfile(unsigned int mask);

  inline void Shutdown( int num_sec);
  
  unsigned int Profile;

  // Virtual functions that has to be defined for each new member
  virtual void Init(){};
  virtual void Close(){};
  virtual void DoYourJob(int par = 0) = 0;  //pure virtual

  /// Create a new thread that execute this member
  int Activate( const char *thdName=NULL );

  // Inner Methods
  inline void Run();
  inline void Suspend();
  inline void WakeUp();
  inline void Exit() { Status = TERMINATING; };

  inline void SubscribeMsgType(int type, DCDT_RequestType ReqType=DCDT_ALL_MSG);
  inline void UnSubscribeMsgType(int type, DCDT_RequestType ReqType=DCDT_ALL_MSG);
  inline void UnSubscribeAll();

  inline const DCDT_Msg* ReceiveMsg( bool wait );
  inline const DCDT_Msg* ReceiveMsgType( int type, bool wait );
  inline const DCDT_Msg* ReceiveLastMsg( int type, bool wait );
  inline DCDT_Msg* CreateMsg(int type, int delivery_warranty); // method to set the warranty
  
  // TODO: remove the following outdated method, now we have PostOffice
  inline DCDT_Msg* CreateMsg(int type);

  inline void ShareMsg(DCDT_Msg *);
  inline DCDT_Msg * DupMsg(DCDT_Msg *);
  inline void Wait( DCDT_TIME sleep_time);
  inline void Wait( struct timeval deadline );

 protected:

  DCDT_PostOffice* myPostOffice;
  char myIPAddr[INET_ADDRSTRLEN];

  /// Agora that the member belongs to
  DCDT_Agora* myAgora;

  /// Our Finder, valid only for system members
  DCDT_Finder* myFinder;

  /// Member ID
  int myID;

  /// Agora ID
  int myAgoraID;

  bool EndOfSession;
  DCDT_TIME 
    /// Execution period, if == 0 -> cyclic member
    Period, 
    /// Last cycle duration of DoYourJob method
    ExecTime,
    /// Last cycle duration
    CycleTime,
    /// When this member was activated
    StartTime;

  /// Flag: true - cycle duration > execution period
  bool CycleTimeExceded;
  struct timeval deadline;

  int Status;

#ifdef SEQUENCER_VERSION
  DCDT_Sequencer * mySequencer;
#endif
 
  /// Message subscription mask
  int SubsMsgMask[NUMINT_MSGTYPE];
  
#ifdef USE_BOOST
  struct BoostThread *myBoostThread;
  boost::thread *BoostThreadRef;
  void *myBoostThd;
  char filler[8];
#else
  // Thread management code
  pthread_t      myThread;
  pthread_attr_t myThreadAttr;
#endif
};

inline void DCDT_Member::SetID( int ID){
  myID = ID;
};

inline int DCDT_Member::ReadID(){
  return myID;
};

inline bool DCDT_Member::ReadExcededFlag() {
  return CycleTimeExceded;
}

inline void DCDT_Member::ClearExcededFlag() {
  CycleTimeExceded=false;
}

inline bool DCDT_Member::CheckProfile(unsigned int mask){
  return ((Profile & mask) == mask);
};

inline void DCDT_Member::SubscribeMsgType(int type, DCDT_RequestType ReqType) {
  myPostOffice->SubscribeMsgTypeID( myID, type, ReqType );
}

inline void DCDT_Member::UnSubscribeMsgType(int type, DCDT_RequestType ReqType) {
  myPostOffice->UnSubscribeMsgTypeID( myID, type, ReqType);
}

inline void DCDT_Member::UnSubscribeAll() {
  myPostOffice->UnSubscribeAll( myID );
}

inline const DCDT_Msg * DCDT_Member::ReceiveMsg( bool wait ) {
  return( myPostOffice->ReadNextMsg( myID, wait ));
}

inline const DCDT_Msg * DCDT_Member::ReceiveMsgType( int type, bool wait ) {
  return( myPostOffice->ReadNextMsg( myID, type, wait ) );
}

inline const DCDT_Msg * DCDT_Member::ReceiveLastMsg( int type, bool wait ) {
  return( myPostOffice->ReadLastMsg( myID, type, wait ) );
}

inline DCDT_Msg* DCDT_Member::CreateMsg(int type, int delivery_warranty){

  TRC_PRINT( DCDT_TRC_MEMBER, TRC1, ("CreateMsg()"));

  DCDT_Msg * newMsg = new DCDT_Msg(type) ;
  newMsg->SetMemberID( myID );
  newMsg->SetAgoraID( myAgoraID );
  newMsg->SetDeliveryWarranty( delivery_warranty ); 

  return newMsg;
};

inline DCDT_Msg* DCDT_Member::CreateMsg(int type ){

  TRC_PRINT( DCDT_TRC_MEMBER, TRC1, ("CreateMsg()"));

  DCDT_Msg * newMsg = new DCDT_Msg(type) ;
  newMsg->SetMemberID( myID );
  newMsg->SetAgoraID( myAgoraID );
  newMsg->SetDeliveryWarranty( 0 ); 

  return newMsg;
};

inline void DCDT_Member::ShareMsg(DCDT_Msg * Msg) {

  TRC_PRINT( DCDT_TRC_MEMBER, TRC1, ("ShareMsg()"));

  // If we are an MPM_LINK and MPM_BRIDGE all the messages that we share
  // are LOCAL messages; if we are only an MPM_LINK the messages are REMOTE
  // OLD BEHAVIOR
  if ( myAgora->CheckMemberProfile( myID, MPM_LINK ) ) {
    if ( myAgora->CheckMemberProfile( myID, MPM_BRIDGE ) ) {
        myPostOffice->AddMsg( Msg, myID, DCDT_LOCAL_MSG );
    }
     else {
        myPostOffice->AddMsg( Msg, myID, DCDT_REMOTE_MSG );
     }
  }
  else {
     myPostOffice->AddMsg( Msg, myID, DCDT_LOCAL_MSG );
  }
};

inline DCDT_Msg *DCDT_Member::DupMsg(DCDT_Msg * Msg) {
  int payloadlen;

  TRC_PRINT( DCDT_TRC_MEMBER, TRC1, ("DupMsg()"));

  DCDT_Msg * newMsg = new DCDT_Msg( Msg->ReadType(), Msg->ReadPriority());

  if( (payloadlen = Msg->ReadPayloadLen()) ) {
    void * newpayload = malloc( payloadlen );
    if ( newpayload != NULL ) {
      memmove(newpayload, Msg->GetPayload(), payloadlen);
      newMsg->SetPayload( newpayload, payloadlen);
    }
    else {
      delete newMsg;
      return ( NULL );
    }
  }
  return ( newMsg );
};

inline void DCDT_Member::Shutdown(int Sec_Delay) {
  
  // terminate the session within Sec_Delay seconds
  if( Sec_Delay)
    Delay ( Sec_Delay*1000000 );
  
  myAgora->SetStatus ( DCDT_TERMINATING );
};

#endif  //define
