/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_PostOffice.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                    Cristian Giussani (cgiussani@fastflow.it)


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#include <DCDT_Agora.h>
#include <DCDT_PostOffice.h>
#include <DCDT_PostOffice_SLWBU.h>
#include <DCDT_PostOffice_SLWSM.h>
#include <DCDT_PostOffice_SLWDC.h>

DCDT_PostOffice::DCDT_PostOffice (DCDT_Agora *agora, int type)
{
  myAgora = agora;
  PO_Type = type;
  
  PostOffice_SLWB=NULL;
  PostOffice_SLWDC=NULL;
  PostOffice_SLWSM=NULL;
  
  switch ( type ) {
  case POSTOFFICE_SLWB:
    PostOffice_SLWB = new DCDT_PostOffice_SLWBU( agora );
    printf( "PostOffice: Single Lock With Buffer\n" );
    break;
    
  case POSTOFFICE_SLWDC:
    PostOffice_SLWDC = new DCDT_PostOffice_SLWDC( agora );
    printf( "PostOffice: Single Lock With Dispatcher With Conditional Variables" );
    break;	
    
  case POSTOFFICE_SLWSM:
    PostOffice_SLWSM = new DCDT_PostOffice_SLWSM( agora );
    printf( "PostOffice: Single Lock With Single Message" );
    break;
    
  default:
    printf ( "Unknown PostOffice type %d\n", type );
    exit ( 1 );
    break;
  }
};


DCDT_PostOffice::~DCDT_PostOffice ()
{
  if ( PostOffice_SLWB )
    delete PostOffice_SLWB;
  
  if ( PostOffice_SLWDC )
    delete PostOffice_SLWDC;
  
  if ( PostOffice_SLWSM )
    delete PostOffice_SLWSM;
};

/*! \brief Log the message pointed by *msg
 *
 */
inline void DCDT_PostOffice::LogMsg( DCDT_Msg *msg ) {
  char buffer[512];
  DCDT_MsgHeader *header;
  char *payload;
  int idmy=0,idx=0,payloadlen=0;
  
  if ( msg != NULL ) {
    header = msg->GetHeader();
    
    if ( (payload=(char *)msg->GetPayload()) != NULL ) {
      payloadlen=msg->ReadPayloadLen();
      
      while ( ( idx < payloadlen ) && ( idmy < (int)sizeof(buffer)-7 ) ){
	if ( isprint( payload[idx] ) )
	  idmy += sprintf ( &buffer[idmy], "%c", payload[idx] );
	else
	  idmy += sprintf ( &buffer[idmy], "<0x%.2hhX>", 
			    (unsigned char)payload[idx] );
	idx++;
      }
    }
    else
      sprintf ( buffer, "NULL" );
    
    TRC_PRINT( DCDT_TRC_POSTOFFICE, TRC0, ("Type: %d Payload <%s>", msg->ReadType(), buffer) );
  }
};

/*! \brief Adds a new user member to the PostOffice 
 *
 *  Enable an Agora' user member, idenfied by \a Mem_ID, to send and receive messages througth the PostOffice.
 *  \param Mem_ID Identifier of the member to be actived  
 */
void DCDT_PostOffice::AddMember(int Mem_ID) {
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    PostOffice_SLWB->AddMember( Mem_ID );
    break;

  case POSTOFFICE_SLWDC:
    PostOffice_SLWDC->AddMember( Mem_ID );
    break;

  case POSTOFFICE_SLWSM:
    PostOffice_SLWSM->AddMember( Mem_ID );
    break;

  default:
    printf ( "AddMember not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }
};

/*! \brief Adds a new system member to the PostOffice 
 *
 *  Enable an Agora' system member, idenfied by \a Mem_ID, to send and receive messages througth the PostOffice.
 *  \param Mem_ID Identifier of the member to be actived  
 */
void DCDT_PostOffice::AddSysMember(int Mem_ID) {
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    PostOffice_SLWB->AddSysMember( Mem_ID );
    break;

  case POSTOFFICE_SLWDC:
    PostOffice_SLWDC->AddSysMember( Mem_ID );
    break;

  case POSTOFFICE_SLWSM:
    PostOffice_SLWSM->AddSysMember( Mem_ID );
    break;

  default:
    printf ( "AddSysMember not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }
};

/*! \brief Delete a member from the PostOffice 
 *
 *  Disable an Agora' member, idenfied by \a Mem_ID, from sending and receiving messages througth the PostOffice.
 *  @param Mem_ID Identifier of the member to be disabled
 */
void DCDT_PostOffice::DeleteMemberTableEntry(int Mem_ID){
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    PostOffice_SLWB->DeleteMemberTableEntry( Mem_ID );
    break;

  case POSTOFFICE_SLWDC:
    PostOffice_SLWDC->DeleteMemberTableEntry( Mem_ID );
    break;

  case POSTOFFICE_SLWSM:
    PostOffice_SLWSM->DeleteMemberTableEntry( Mem_ID );
    break;

  default:
    printf ( "DeleteMemberTableEntry not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }
  
};

/*! \brief Add a new message to the PostOffice queue.
 *
 *  When the member \a Mem_ID wants to insert a new message the same member checks the subscription table and when a member is subscribed to those messages copy the message into the member message queue; this function also updates the message with the number of members subscribed to this message type.
 *  \param MP pointer to the message to share
 *  \param Mem_ID member identifier of the publisher
 */
void DCDT_PostOffice::AddMsg (DCDT_Msg * MP, int Mem_ID, DCDT_RequestType Type)
{
  TRC_EXEC ( DCDT_TRC_POSTOFFICE, TRC0, LogMsg( MP ); );

  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    PostOffice_SLWB->AddMsg( MP, Mem_ID, Type );
    break;

  case POSTOFFICE_SLWDC:
    PostOffice_SLWDC->AddMsg( MP, Mem_ID );
    break;

  case POSTOFFICE_SLWSM:
    PostOffice_SLWSM->AddMsg( MP, Mem_ID, Type );
    break;

  default:
    printf ( "AddMsg not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
    
  }

}

/*! \brief Read next message for the member \a MemID.
 *
 *  Check if there is an unread message in the member queue; if we found one we have to mark it as deletable and return a pointer to it.
 *  If, during the queue slide, we find a message marked as deletable we have to delete it and go on.
 *  \param MemID Member identifier
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
const DCDT_Msg *DCDT_PostOffice::ReadNextMsg (int MemID, bool wait)
{
#ifdef DCDT_USE_DEBUG
  DCDT_Msg *msg;
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    msg = PostOffice_SLWB->ReadNextMsg( MemID, wait );
    break;

  case POSTOFFICE_SLWDC:
    msg = PostOffice_SLWDC->ReadNextMsg( MemID, wait );
    break;

  default:
    printf ( "ReadNextMsg( MemID ) not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }

  TRC_EXEC ( DCDT_TRC_POSTOFFICE, TRC0, LogMsg( msg ); );
  return ( msg );
#else
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    return ( PostOffice_SLWB->ReadNextMsg( MemID, wait ) );
    
  case POSTOFFICE_SLWDC:
    return ( PostOffice_SLWDC->ReadNextMsg( MemID, wait ) );
    
  default:
    printf ( "ReadNextMsg( MemID ) not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }
#endif /* DCDT_USE_DEBUG */

}

/*! \brief Read next message of type \a type for the member \a MemID.
 *
 *  Check if there is an unread message in the member queue; if we found one we have to mark it as deletable and return a pointer to it.
 *  If, during the queue slide, we find a message marked as deletable we have to delete it and go on.
 *  \param MemID Member identifier
 *  \param type Type of the message to be read
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
const DCDT_Msg *DCDT_PostOffice::ReadNextMsg ( int MemID, int type, bool wait )
{
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    return ( PostOffice_SLWB->ReadNextMsg( MemID, type, wait ) );

  case POSTOFFICE_SLWDC:
    return ( PostOffice_SLWDC->ReadNextMsg( MemID, type, wait ) );

  case POSTOFFICE_SLWSM:
    return ( PostOffice_SLWSM->ReadNextMsg ( MemID, type ) );

  default:
    printf ( "ReadNextMsg( MemID, type) not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }
}


/*! \brief Read the last message received of type \a type for the member \a MemID.
 *
 *  Check if there is an unread message in the member queue; if we found one we have to mark it as deletable and return a pointer to it.
 *  If, during the queue slide, we find a message marked as deletable or that isn't the last one we have to delete it and go on.
 *  \param MemID Member identifier
 *  \param type Type of the message to be read
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
const DCDT_Msg *DCDT_PostOffice::ReadLastMsg ( int MemID, int type, bool wait )
{
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    return ( PostOffice_SLWB->ReadLastMsg( MemID, type, wait ) );
    
  case POSTOFFICE_SLWDC:
    return ( PostOffice_SLWDC->ReadLastMsg( MemID, type, wait ) );
    
  case POSTOFFICE_SLWSM:
    return ( PostOffice_SLWSM->ReadNextMsg ( MemID, type ) );
    
  default:
    printf ( "ReadLastMsg not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }
}

/*! \brief Check if a message was read by all the member, if it is mark the message as deletable.
 */
inline void DCDT_PostOffice::SetActMsgDeletable (int Member_ID)
{
}

/*! \brief Called periodically by MsgManager
 */
void DCDT_PostOffice::CleanupList()
{
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    PostOffice_SLWB->CleanupList();
    break;

  case POSTOFFICE_SLWDC:
    PostOffice_SLWDC->CleanupList();
    break;

  case POSTOFFICE_SLWSM:
    PostOffice_SLWSM->CleanupList();
    break;

  default:
    printf ( "CleanupList not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }
}

/* !\brief Subscribe member \a Member_ID to the messages of type \a MsgType.
 *
 * \param Member_ID Member identifier
 * \param MsgType Type of messages that the member want to receive
 * \param ReqType LOCAL/REMOTE/both messages
 */
void DCDT_PostOffice::SubscribeMsgTypeID (int Member_ID, int MsgType, DCDT_RequestType ReqType )
{
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    PostOffice_SLWB->SubscribeMsgTypeID(Member_ID, MsgType, ReqType );
    break;

  case POSTOFFICE_SLWDC:
    PostOffice_SLWDC->SubscribeMsgTypeID(Member_ID, MsgType, ReqType );
    break; 

  case POSTOFFICE_SLWSM:
    PostOffice_SLWSM->SubscribeMsgTypeID(Member_ID, MsgType, ReqType );
    break;

  default:
    printf ( "SubscribeMsgTypeID not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }
}


/*! \brief Unsubscribe member \a Member_ID from the messages of type \a MsgType.
 *
 * \param Member_ID Member identifier
 * \param MsgType Messages type for with the member wants to be unsubscribed
 * \param ReqType 
 */
void DCDT_PostOffice::UnSubscribeMsgTypeID (int Member_ID, int MsgType, DCDT_RequestType ReqType)
{
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    PostOffice_SLWB->UnSubscribeMsgTypeID (Member_ID, MsgType, ReqType);
    break;

  case POSTOFFICE_SLWDC:
    PostOffice_SLWDC->UnSubscribeMsgTypeID (Member_ID, MsgType, ReqType);
    break;

  case POSTOFFICE_SLWSM:
    PostOffice_SLWSM->UnSubscribeMsgTypeID ( Member_ID, MsgType, ReqType );
    break;
  
  default:
    printf ( "UnSubscribeMsgTypeID not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }
}

/*! \brief Unsubscribe the member \a Member_ID from all the messages types
 *
 * \param Member_ID Member identifier
 */
void DCDT_PostOffice::UnSubscribeAll ( int Member_ID )
{
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    PostOffice_SLWB->UnSubscribeAll ( Member_ID );
    break;

  case POSTOFFICE_SLWDC:
    PostOffice_SLWDC->UnSubscribeAll ( Member_ID );
    break;

  case POSTOFFICE_SLWSM:
    PostOffice_SLWSM->UnSubscribeAll ( Member_ID );
    break;

  default:
    printf ( "UnSubscribeAll not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }
}

/*! \brief Check if SubscribedMsgMask ( used by LinkTx ) is changed.
 *
 *  The global mask is build upon the single message mask of each active member, the mask don't include the mask of the member \a MemID ( usually the calling member ); if \a MemID is a bridge all the members masks are used otherwise only the masks of MPM_USER and MPM_BRIDGE type members are used; so in a LinkTx member we can build the subscriber mask that needs to be passed to the other side of the link.
 *  \param mask Pointer to the subscribed message mask
 *  \param MemID Member identifier
 *  \return TRUE - SubscribedMsgMask was changed
 *  \return FALSE - SubscribedMsgMask wasn't changed
 */
bool DCDT_PostOffice::ReadSubscribedMsgMask(unsigned int *mask, int MemID) 
{
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    return ( PostOffice_SLWB->ReadSubscribedMsgMask(mask, MemID) );

  case POSTOFFICE_SLWDC:
    return ( PostOffice_SLWDC->ReadSubscribedMsgMask(mask, MemID) );
    
  case POSTOFFICE_SLWSM:
    return ( PostOffice_SLWSM->ReadSubscribedMsgMask(mask, MemID) );

  default:
    printf ( "ReadSubscribeMsgMask not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );
  }
}

/*! \brief Chech if the Suscribed mask for the member \aMemID was changed
 *
 *  \param MemID Membed identifier
 */
bool DCDT_PostOffice::UpdateSubs(int MemID) 
{
  switch ( PO_Type ) {
  case POSTOFFICE_SLWB:
    return ( PostOffice_SLWB->UpdateSubs(MemID) );

  case POSTOFFICE_SLWDC:
    return ( PostOffice_SLWDC->UpdateSubs(MemID) );

  case POSTOFFICE_SLWSM:
    return ( PostOffice_SLWSM->UpdateSubs(MemID) );
    
  default:
    printf ( "UpdateSubs not implemented for PostOffice type %d", PO_Type );
    exit ( 1 );    
  } 
}

#ifdef TEST_PART

void DCDT_PostOffice::InsertListTime (int MemID, int Type_Op)
{
  DCDT_TIME x = GetTime();
  TimeTesterElem *newElem;
  newElem = new TimeTesterElem;
  newElem->Data = x;
  newElem->Type_Op = Type_Op;
  newElem->prev = TestElem[MemID].prev;
  TestElem[MemID].prev = newElem; 
} 


void DCDT_PostOffice::VisualList()
{
  TimeTesterElem *newElem;
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    newElem = TestElem[i].prev;
    while (newElem) {
      cout<<"["<<i<<","<<newElem->Type_Op<<","<<newElem->Data-TestElem[i].Data<<"]";
      newElem=newElem->prev;
    }
  } 
}

void DCDT_PostOffice::SaveList(char* nomefile)  
{
  TimeTesterElem *newElem; 
  FILE *fout;
  time_t longtime;
  struct tm* time_experiment;
  
  char string[100] = "DATI_";

  strcat(string,nomefile); 
  
  fout = fopen (string,"wb");

  // Insert time and date of the operation
  time(&longtime);
  time_experiment = localtime (&longtime);

  putw(time_experiment->tm_hour,fout);
  putw(time_experiment->tm_min,fout); 
  putw(time_experiment->tm_mday,fout); 
  putw(time_experiment->tm_mon,fout);
  putw(1900 + time_experiment->tm_year,fout);

  // Software version and system parameters
  putw(VERSION,fout);
  putw(NUMMEMBERFORMAIN,fout);

  // Note that  NUMMEMBERFORMAIN = NUMREADER + NUMWRITER + NUMREADERWRITER 
  putw(NUMREADER,fout);      
  putw(NUMWRITER,fout);      
  putw(NUMREADERWRITER,fout); 
  
  // Number of cycles  
  putw(READCYCLE,fout); 
  putw(WRITECYCLE,fout);        
  putw(READWRITECYCLE,fout);   
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    
    newElem = TestElem[i].prev;
    
    while (newElem) {
      putw(newElem->Type_Op,fout);
      putw(newElem->Data-TestElem[i].Data,fout);
      
      newElem=newElem->prev;
    }
    putw(666,fout);
  }

  fclose(fout);
}

#endif
