/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_Time.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
          (C) 2003  Cristian Giussani ( cgiussani@fastflow.it )


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_TIME_H
#define DCDT_TIME_H

#include <sys/time.h>
#include <errno.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <DCDT_Defs.h>

/// delay_time is in microseconds
inline void Delay( DCDT_TIME delay_time) {
 struct timespec timeout;
  
  timeout.tv_sec = (delay_time / 1000000);
  timeout.tv_nsec = ((long)delay_time % 1000000) * 1000;

  while( nanosleep( &timeout, &timeout )) {
    if (errno != EINTR )
      break;
  };
  
};

inline int DeltaSeconds( struct timeval last, struct timeval now)
{
    return (now.tv_sec - last.tv_sec);
};

inline int DeltaSeconds( struct timespec last, struct timespec now)
{
    return (now.tv_sec - last.tv_sec);
};

inline int DeltaSeconds( DCDT_TIME last, DCDT_TIME now)
{
    return (now/1000000 - last/1000000);
};




inline DCDT_TIME GetTime() {
  struct timeval tv;
  //  struct timezone tz;
  
  // impostato a NULL il secondo termine per
  // velocizzare la routine
  
  gettimeofday(&tv, NULL);
  //  gettimeofday(&tv, &tz);
  
  return (tv.tv_sec * 1000000 + tv.tv_usec);

};

inline DCDT_TIME tv2DCDT_TIME( struct timeval tv) {

  return ((DCDT_TIME)(tv.tv_sec * 1000000 + tv.tv_usec));

};


inline bool operator < ( struct timeval t1, struct timeval t2 )
{
  return t1.tv_sec < t2.tv_sec ||
         (t1.tv_sec == t2.tv_sec && t1.tv_usec < t2.tv_usec);
}

inline bool operator > ( struct timeval t1, struct timeval t2 )
{
  return t1.tv_sec > t2.tv_sec ||
         (t1.tv_sec == t2.tv_sec && t1.tv_usec > t2.tv_usec);
}

inline bool operator >= ( struct timeval t1, struct timeval t2) {
  return ( (t1.tv_sec > t2.tv_sec ) ||
         ( (t1.tv_sec == t2.tv_sec) && (t1.tv_usec >= t2.tv_usec)));
};

inline bool operator == ( struct timeval t1, struct timeval t2 )
{
  return (t1.tv_sec == t2.tv_sec && t1.tv_usec == t2.tv_usec);
}

inline timeval operator + ( struct timeval t1, struct timeval t2 )
{
  struct timeval tmp;
  tmp.tv_sec = t1.tv_sec + t2.tv_sec;
  if ( (tmp.tv_usec = t1.tv_usec + t2.tv_usec) >= 1000000 ) {
    ++tmp.tv_sec;
    tmp.tv_usec -= 1000000;
  }
  return tmp;
}

inline struct timeval operator + ( struct timeval t1, long t2 )
{
  struct timeval tmp;
  tmp.tv_sec = t1.tv_sec;
  if ( (tmp.tv_usec = t1.tv_usec + t2) >= 1000000 ) {
    ++tmp.tv_sec;
    tmp.tv_usec -= 1000000;
  }
  return tmp;
}

inline struct timeval operator + ( struct timeval t1, DCDT_TIME t2 )
{
  struct timeval tmp;
  tmp.tv_sec = t1.tv_sec;
  if ( (tmp.tv_usec = t1.tv_usec + t2) >= 1000000 ) {
    ++tmp.tv_sec;
    tmp.tv_usec -= 1000000;
  }
  return tmp;
};

inline struct timespec operator + ( struct timespec t1, unsigned int t2 )
{
  struct timespec tmp;
  tmp.tv_sec = t1.tv_sec;
  if ( (tmp.tv_nsec = t1.tv_nsec + t2) >= 1000000000 ) {
    ++tmp.tv_sec;
    tmp.tv_nsec -= 1000000000;
  }
  return tmp;
};

inline struct timeval& operator += ( struct timeval &t1, struct timeval t2 )
{
  t1.tv_sec += t2.tv_sec;
  if ( (t1.tv_usec += t2.tv_usec) >= 1000000 ) {
    ++t1.tv_sec;
    t1.tv_usec -= 1000000;
  }
  return t1;
}



//-----------------------------------------------------------------------------//
/*
inline struct timeval operator - ( struct timeval t1, struct timeval t2 )
{
  struct timeval tmp;
  tmp.tv_sec = t1.tv_sec - t2.tv_sec;
  if ( (tmp.tv_usec = t1.tv_usec - t2.tv_usec) < 0 ) {
    --tmp.tv_sec;
    tmp.tv_usec += 1000000;
  }
  return tmp;
}
*/

inline long int operator - ( struct timeval t1, struct timeval t2 )
{
  struct timeval tmp;
  tmp.tv_sec = t1.tv_sec - t2.tv_sec;
  if ( (tmp.tv_usec = t1.tv_usec - t2.tv_usec) < 0 ){
    --tmp.tv_sec;
    tmp.tv_usec += 1000000;
  }

  printf("diff = %ld sec, %ld usec\n", tmp.tv_sec, tmp.tv_usec);
  fflush (stdout);

  return (tmp.tv_usec + (1000000 * tmp.tv_sec));
}

//-----------------------------------------------------------------------------//

inline struct timeval& operator -= ( struct timeval& t1, struct timeval t2 )
{
  t1.tv_sec -= t2.tv_sec;
  if ( (t1.tv_usec -= t2.tv_usec) < 0 ) {
    --t1.tv_sec;
    t1.tv_usec += 1000000;
  }
  return t1;
}

/*
  bool operator > (struct timeval t1, struct timeval t2);
  bool operator < (struct timeval t1, struct timeval t2);
  bool operator <= (struct timeval t1, struct timeval t2);
  bool operator >= (struct timeval t1, struct timeval t2);
  struct timeval operator + (struct timeval t1, struct timeval t2);
  struct timeval operator + (struct timeval t1, long );
  //struct timeval operator - (struct timeval t1, struct timeval t2 );
  long int operator - (struct timeval t1, struct timeval t2 );
*/

#ifdef SEQUENCER_VERSION
/*! \brief 
 *
 *  Detail
 */
 class SequencerElem  {
  public:
  SequencerElem(struct timeval new_deadline) {
    deadline = new_deadline;
    next = NULL;
    pthread_cond_init( &condition, NULL);
    pthread_mutex_init( &mutex, NULL );
  };

  ~SequencerElem() {
    next = NULL;
    pthread_mutex_destroy( &mutex );
    pthread_cond_destroy( &condition );
  };

  struct timeval deadline;
  pthread_cond_t condition;
  pthread_mutex_t mutex;
  SequencerElem *next;
};

/*! \brief
 *
 *  Detail
 */
class DCDT_Sequencer {
  friend class DCDT_MsgManager;
  friend class DCDT_Agora;
  friend class DCDT_Member;
  public:
    DCDT_Sequencer() { Head = NULL; Tail = NULL; NumOfDeadlines = 0;};
    ~DCDT_Sequencer() {
        SequencerElem *tmpElem;
        while( Head ) {
          tmpElem = Head;
          delete(Head);
          Head = tmpElem->next;
        }
     };

    inline void Wait( DCDT_TIME sleep_time);
    inline void Wait( struct timeval deadline );
    inline void CheckNextDeadline();

    inline void InsertElem( SequencerElem *newElem);

  private:

    int NumOfDeadlines;

    pthread_mutex_t SequencerMutex;
    SequencerElem *Head, *Tail;  //Start and end of the event list
};

/*! \brief Insert a new deadline element.
 */
inline void DCDT_Sequencer::InsertElem( SequencerElem * newElem ) {

  SequencerElem *nextElem, *lastElem;

  pthread_mutex_lock( &SequencerMutex );

  if( Head ) {
    // there is at least yet one element

    if ( newElem->deadline > Head->deadline ) {
      // the element is not the first
      // dbg_printf("Sequencer: not the first\n");

      lastElem = Head;
      nextElem = Head->next;

      // let's find the right place
      while( nextElem && (newElem->deadline > lastElem->deadline)) {
          lastElem = nextElem;
          nextElem = nextElem->next;
      };

      // dbg_printf("Sequencer: among the elements\n");

      lastElem->next = newElem;
      newElem->next = nextElem;
    }
    else {
      // we have to insert the new element on the top of the list
      // dbg_printf("Sequencer: first of a crowded list\n");

      newElem->next = Head;
      Head = newElem;
    }
  }
  else {
    // List is empty
    // dbg_printf("Sequencer: first of an empty list\n");
    Head = newElem;

  };

  NumOfDeadlines++;

/*
  printf("Insert               Sequencer = %d\n", NumOfDeadlines);
  fflush(stdout);
*/
  pthread_mutex_unlock( &SequencerMutex );
}

/*! \brief Wait for the specified amount of time then return.
 *
 *  Create ad insert a new deadline into the sequencer, after the insertion wait that the corresponding condition becomes true; this function returns only when \a sleep_time was passed.
 */
inline void DCDT_Sequencer::Wait( DCDT_TIME sleep_time ) {
  struct timeval now, deadline;

  gettimeofday( &now, NULL);

  deadline = now + sleep_time;
  SequencerElem * newElem = new SequencerElem( deadline );

  InsertElem( newElem );
  // now we wait for someone to wake us up

  pthread_mutex_lock( &(newElem->mutex) );
  pthread_cond_wait( &(newElem->condition), &(newElem->mutex) );

  pthread_mutex_unlock( &(newElem->mutex) );

  // we are again awake. let's remove the element

  delete( newElem );

//  NumOfDeadlines--;

/*
  printf("Del               Sequencer = %d\n", NumOfDeadlines);
  fflush(stdout);
*/

};

/*! \brief Wait for specified deadline then return.
 *
 *  Create ad insert a new deadline into the sequencer, after the insertion wait that the corresponding condition becomes true; this function returns only when \a sleep_time was passed.
 */
inline void DCDT_Sequencer::Wait( struct timeval deadline ) {

  struct timeval now;
  SequencerElem * newElem;

  gettimeofday( &now, NULL );

  if (deadline > now ) {
    newElem = new SequencerElem( deadline );

    InsertElem( newElem );
    // now we wait for someone to wake us up

    pthread_mutex_lock( &(newElem->mutex) );
    pthread_cond_wait( &(newElem->condition), &(newElem->mutex) );

    // unlock was automatically done by cond_wait
    // pthread_mutex_unlock( &(newElem->mutex) );

    // we are again awake. let's remove the element

    delete( newElem );

    //  NumOfDeadlines--;
    /*
    printf("Del               Sequencer = %d\n", NumOfDeadlines);
      fflush(stdout);
    */
  }
  else {
  //  dbg_printf("**** Missed deadline. No more delay. returning *****\n");

  }
};

/*! \brief Check if the first ( the nearest ) deadline was experied.
 *
 *  If so delete the deadline from the queue and signal it to the waiting thread via the corresponding condition. 
 */
inline void DCDT_Sequencer::CheckNextDeadline() {
  struct timeval now;
  SequencerElem * tmpElem;


  pthread_mutex_lock( &SequencerMutex );

  /*
  tmpElem = Head;

  int i = 1;
  while( tmpElem ) {
    printf("-|%d|-", i++);
    tmpElem = tmpElem->next;
    }
  printf("\n");
  fflush(stdout);

  */
  tmpElem = Head;

  if( tmpElem ) {

     gettimeofday(&now, NULL);

     if( now >= tmpElem->deadline ) {
      // time is elapsed: let's signal it

      Head = tmpElem->next;
      pthread_cond_signal( &(tmpElem->condition) );

      // and now let's remove the Synchro elem from
      // the list. We actually not delete it.
      // it's up to the thread who requested the service to
      // do it

      NumOfDeadlines--;
    }
  }
  pthread_mutex_unlock( &SequencerMutex );

};
#endif // SEQUENCER_VERSION

#endif
 
