/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_PostOffice.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                    Cristian Giussani (cgiussani@fastflow.it)


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_POSTOFFICE_H
#define DCDT_POSTOFFICE_H



#ifdef TEST_PART

typedef struct _TimeTesterElem
{
  // Time of the measure
  DCDT_TIME Data;
  
  // Type of the measure
  int Type_Op;
    
  _TimeTesterElem *prev;
}
TimeTesterElem;

#endif

#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <iostream>
#include <DCDT_Agora.h>
#include <DCDT_Defs.h>
#include <sys/time.h>
#include <DCDT_Time.h>
#include <DCDT_Msg.h>

class DCDT_Member;

/// Structure needed to keep track of various member's subscriptions
typedef struct _SubsMsgElem
{
  DCDT_Mutex mtx;
  // Local and Remote are referred to the msg, i.e. to the its publisher

  /// Suscribers for local generated messages
  unsigned int SubscribersIfLocal[NUMINT_MEMBERS];

  /// Subscribers for remote generated messages
  unsigned int SubscribersIfRemote[NUMINT_MEMBERS];
}
SubsMsgTypeElem;


/*! \brief Data structures and method to subscribe, insert, delete and read messages.
 *
 *  NOTE: This class isn't a member of the Agor, the real member that uses this class is DCDT_MsgManager.
 */
class DCDT_PostOffice
{
  friend class DCDT_MsgManager;
  friend class DCDT_Agora;
  friend class DCDT_Member;

 private:
  int Type;

  DCDT_PostOffice_TL *PostOffice_TL;
  DCDT_PostOffice_TLWRC *PostOffice_TLWRC;
  DCDT_PostOffice_SLWD *PostOffice_SLWD;
  DCDT_PostOffice_SLWBR *PostOffice_SLWBR;
  DCDT_PostOffice_SLWBU *PostOffice_SLWBU;
  DCDT_PostOffice_SLWDC *PostOffice_SLWDC;

 public:
  DCDT_PostOffice ( int type, DCDT_Agora *agora )
    {
      Type = type;

      switch ( Type ) {

      case PO_TL:
	PostOffice_TL = new DCDT_PostOffice_TL( agora );
	break;

      case PO_TLWRC:
	PostOffice_TLWRC = new DCDT_PostOffice_TLWRC( agora );
	break;

      case PO_SLWD:
	PostOffice_SLWD = new DCDT_PostOffice_SLWD( agora );
	break;

      case PO_SLWBR:
	PostOffice_SLWBR = new DCDT_PostOffice_SLWBR( agora );
	break;

      case PO_SLWBU:
	PostOffice_SLWBU = new DCDT_PostOffice_SLWBU( agora );
	break;

      case PO_SLWDC:
	PostOffice_SLWDC = new DCDT_PostOffice_SLWDC( agora );
	break;

      default:
	count << "Unknowk PostOffice type " << type << EOL;
	exit ( 1 );
      }
    }

  ~DCDT_PostOffice ()
    {
      switch ( Type ) {

      case PO_TL:
	delete PostOffice_TL;
	PostOffice_TL = NULL;
	break;

      case PO_TLWRC:
	delete PostOffice_TLWRC;
	PostOffice_TLWRC = NULL;
	break;

      case PO_SLWD:
	delete PostOffice_SLWD;
	PostOffice_SLWD = NULL;
	break;

      case PO_SLWBR:
	delete PostOffice_SLWBR;
	PostOffice_SLWBR = NULL;
	break;

      case PO_SLWBU:
	delete PostOffice_SLWBU;
	PostOffice_SLWBU = NULL;
	break;

      case PO_SLWDC:
	delete PostOffice_SLWDC;
	PostOffice_SLWDC = NULL;
	break;
	
      default:
	break;
      }
    };

  inline void AddMember(int Mem_ID);
  inline void AddSysMember(int Mem_ID);
  inline void DeleteMemberTableEntry(int Mem_ID );
  inline void AddMsg (DCDT_Msg * MP, int Mem_ID);
  inline DCDT_Msg *ReadNextMsg (int MemID); // TODO: Check if return "const DCDT_MSG *" is better
  inline void CleanupList ();
  inline void SubscribeMsgTypeID (int Member_ID, int MsgType);
  inline void UnSubscribeMsgTypeID (int Member_ID, int MsgType);
  inline void UnSubscribeAll( int Member_ID );
  inline void UpdateSubscribersMask (int MemID);
  inline bool ReadSubscribedMsgMask(unsigned int *mask, int MemID);
  inline bool UpdateSubs(int MemID);
    
};

/*! \brief Activate the PostOffice for the specified member
 */
inline void DCDT_PostOffice::AddMember(int MemID) {
  switch ( Type ) {

  case PO_TL:
    PostOffice_TL->AddMember( MemID );
    break;

  case PO_TLWRC:
    PostOffice_TLWRC->AddMember( MemID );
    break;

  case PO_SLWD:
    PostOffice_SLWD->AddMember( MemID );
    break;

  case PO_SLWBR:
    PostOffice_SLWBR->AddMember( MemID );
    break;

  case PO_SLWBU:
    PostOffice_SLWBU->AddMember( MemID );
    break;

  case PO_SLWDC:
    PostOffice_SLWDC->AddMember( MemID );

  default:
    break;
  }
};

/*! \brief Activate the PostOffice for the specified System Member
 */
inline void DCDT_PostOffice::AddSysMember(int Mem_ID) {
  switch ( Type ) {
    
  case PO_TL:
    PostOffice_TL->AddSysMember( MemID );
    break;
    
  case PO_TLWRC:
    PostOffice_TLWRC->AddSysMember( MemID );
    break;
    
  case PO_SLWD:
    PostOffice_SLWD->AddSysMember( MemID );
    break;

  case PO_SLWBR:
    PostOffice_SLWBR->AddSysMember( MemID );
    break;

  case PO_SLWBU:
    PostOffice_SLWBU->AddSysMember( MemID );
    break;

  case PO_SLWDC:
    PostOffice_SLWDC->AddSysMember( MemID );

  default:
    break;
  }
};

/*! \brief Stop processing messages for the specified member
 */
inline void DCDT_PostOffice::DeleteMemberTableEntry(int Mem_ID){
  switch ( Type ) {

  case PO_TL:
    PostOffice_TL->AddMember( MemID );
    break;

  case PO_TLWRC:
    PostOffice_TLWRC->AddMember( MemID );
    break;

  case PO_SLWD:
    PostOffice_SLWD->AddMember( MemID );
    break;

  case PO_SLWBR:
    PostOffice_SLWBR->AddMember( MemID );
    break;

  case PO_SLWBU:
    PostOffice_SLWBU->AddMember( MemID );
    break;

  case PO_SLWDC:
    PostOffice_SLWDC->AddMember( MemID );

  default:
    break;
  }
};

/*! \brief Add a new message to the PostOffice queue.
 */
inline void DCDT_PostOffice::AddMsg (DCDT_Msg * MP, int Mem_ID)
{
  switch ( Type ) {
    
  case PO_TL:
    PostOffice_TL->AddMsg( MP, MemID );
    break;

  case PO_TLWRC:
    PostOffice_TLWRC->AddMsg( MP, MemID );
    break;

  case PO_SLWD:
    PostOffice_SLWD->AddMsg( MP, MemID );
    break;

  case PO_SLWBR:
    PostOffice_SLWBR->AddMsg( MP, MemID );
    break;

  case PO_SLWBU:
    PostOffice_SLWBU->AddMsg( MP, MemID );
    break;

  case PO_SLWDC:
    PostOffice_SLWDC->AddMsg( MP, MemID );

  default:
    break;
  }
}


/*! \brief Read next message for the member \a MemID.
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg *DCDT_PostOffice::ReadNextMsg (int MemID)
{
  switch ( Type ) {
    
  case PO_TL:
    return ( PostOffice_TL->ReadNextMsg( MP, MemID ) );

  case PO_TLWRC:
    return ( PostOffice_TLWRC->ReadNextMsg( MP, MemID ) );

  case PO_SLWD:
    return ( PostOffice_SLWD->ReadNextMsg( MP, MemID ) );

  case PO_SLWBR:
    return ( PostOffice_SLWBR->ReadNextMsg( MP, MemID ) );

  case PO_SLWBU:
    return ( PostOffice_SLWBU->ReadNextMsg( MP, MemID );

  case PO_SLWDC:
    return ( PostOffice_SLWDC->ReadNextMsg( MP, MemID ) );

  default:
    return ( (DCDT_Msg *) NULL );
  }
}

/*! \brief Check if a message was read by all the member, if it is mark the message as deletable.
 */
inline void DCDT_PostOffice::SetActMsgDeletable (int MemID)
{
  switch ( Type ) {
    
  case PO_TL:
    PostOffice_TL->SetActMsgDeletable( MemID );
    break;

  case PO_TLWRC:
    PostOffice_TLWRC->SetActMsgDeletable( MemID );
    break;

  case PO_SLWD:
    PostOffice_SLWD->SetActMsgDeletable( MemID );
    break;

  case PO_SLWBR:
    PostOffice_SLWBR->SetActMsgDeletable( MemID );
    break;

  case PO_SLWBU:
    PostOffice_SLWBU->SetActMsgDeletable( MemID );
    break;

  case PO_SLWDC:
    PostOffice_SLWDC->SetActMsgDeletable( MemID );

  default:
    break;
  }
}

/** \brief Cleanup the message list.
 */
inline void DCDT_PostOffice::CleanupList ()
{
  switch ( Type ) {
    
  case PO_TL:
    PostOffice_TL->CleanupList();
    break;

  case PO_TLWRC:
    PostOffice_TLWRC->CleanupList();
    break;

  case PO_SLWD:
    PostOffice_SLWD->CleanupList();
    break;

  case PO_SLWBR:
    PostOffice_SLWBR->CleanupList();
    break;

  case PO_SLWBU:
    PostOffice_SLWBU->CleanupList();
    break;

  case PO_SLWDC:
    PostOffice_SLWDC->CleanupList();

  default:
    break;
  }
}

/* !\brief Subscribe member \a Member_ID to the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice::SubscribeMsgTypeID (int Member_ID, int MsgType)
{
  switch ( Type ) {
    
  case PO_TL:
    PostOffice_TL->SubscribeMsgTypeID( Member_ID, MsgType );
    break;

  case PO_TLWRC:
    PostOffice_TLWRC->SubscribeMsgTypeID( Member_ID, MsgType );
    break;

  case PO_SLWD:
    PostOffice_SLWD->SubscribeMsgTypeID( Member_ID, MsgType );
    break;

  case PO_SLWBR:
    PostOffice_SLWBR->SubscribeMsgTypeID( Member_ID, MsgType );
    break;

  case PO_SLWBU:
    PostOffice_SLWBU->SubscribeMsgTypeID( Member_ID, MsgType );
    break;

  case PO_SLWDC:
    PostOffice_SLWDC->SubscribeMsgTypeID( Member_ID, MsgType );

  default:
    break;
  }
}

/*! \brief Unsubscribe member \a Member_ID from the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice::UnSubscribeMsgTypeID (int Member_ID, int MsgType)
{
  switch ( Type ) {
    
  case PO_TL:
    PostOffice_TL->UnSubscribeMsgTypeID( Member_ID, MsgType );
    break;

  case PO_TLWRC:
    PostOffice_TLWRC->UnSubscribeMsgTypeID( Member_ID, MsgType );
    break;

  case PO_SLWD:
    PostOffice_SLWD->UnSubscribeMsgTypeID( Member_ID, MsgType );
    break;

  case PO_SLWBR:
    PostOffice_SLWBR->UnSubscribeMsgTypeID( Member_ID, MsgType );
    break;

  case PO_SLWBU:
    PostOffice_SLWBU->UnSubscribeMsgTypeID( Member_ID, MsgType );
    break;

  case PO_SLWDC:
    PostOffice_SLWDC->UnSubscribeMsgTypeID( Member_ID, MsgType );

  default:
    break;
  }
}

/*! \brief Unsubscribe the member \a Member_ID from all the messages types
 *
 */
inline void DCDT_PostOffice::UnSubscribeAll ( int Member_ID )
{
  switch ( Type ) {
    
  case PO_TL:
    PostOffice_TL->UnSubscribeAll( Member_ID );
    break;

  case PO_TLWRC:
    PostOffice_TLWRC->UnSubscribeAll( Member_ID );
    break;

  case PO_SLWD:
    PostOffice_SLWD->UnSubscribeAll( Member_ID );
    break;

  case PO_SLWBR:
    PostOffice_SLWBR->UnSubscribeAll( Member_ID );
    break;

  case PO_SLWBU:
    PostOffice_SLWBU->UnSubscribeAll( Member_ID );
    break;

  case PO_SLWDC:
    PostOffice_SLWDC->UnSubscribeAll( Member_ID );

  default:
    break;
  }
}

/*! \brief Check if SubscribedMsgMask is changed.
 *
 *  \return TRUE - SubscribedMsgMask was changed
 *  \return FALSE - SubscribedMsgMask wasn't changed  
 */
inline bool DCDT_PostOffice::ReadSubscribedMsgMask(unsigned int *mask, int MemID) 
{
  switch ( Type ) {
    
  case PO_TL:
    return ( PostOffice_TL->ReadSubscribedMsgMask( mask, MemID ) );

  case PO_TLWRC:
    return ( PostOffice_TLWRC->ReadSubscribedMsgMask( mask, MemID ) );

  case PO_SLWD:
    return ( PostOffice_SLWD->ReadSubscribedMsgMask( mask, MemID ) );

  case PO_SLWBR:
    return ( PostOffice_SLWBR->ReadSubscribedMsgMask( mask, MemID ) );

  case PO_SLWBU:
    return ( PostOffice_SLWBU->ReadSubscribedMsgMask( mask, MemID ) );

  case PO_SLWDC:
    return ( PostOffice_SLWDC->ReadSubscribedMsgMask( mask, MemID ) );

  default:
    return ( false );
  }
}

inline bool DCDT_PostOffice::UpdateSubs(int MemID) 
{
  switch ( Type ) {
    
  case PO_TL:
    return ( PostOffice_TL->UpdateSubs( MemID ) );

  case PO_TLWRC:
    return ( PostOffice_TLWRC->UpdateSubs( MemID ));

  case PO_SLWD:
    return ( PostOffice_SLWD->UpdateSubs( MemID ) );

  case PO_SLWBR:
    return ( PostOffice_SLWBR->UpdateSubs( MemID ) );

  case PO_SLWBU:
    return ( PostOffice_SLWBU->UpdateSubs( MemID ) );

  case PO_SLWDC:
    return ( PostOffice_SLWDC->UpdateSubs( MemID ) );

  default:
    return ( false );
  }
}

#endif
