/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_PostOffice.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                    Cristian Giussani (cgiussani@fastflow.it)


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_POSTOFFICE_H
#define DCDT_POSTOFFICE_H



#ifdef TEST_PART

typedef struct _TimeTesterElem
{
  // Time of the measure
  DCDT_TIME Data;
  
  // Type of the measure
  int Type_Op;
    
  _TimeTesterElem *prev;
}
TimeTesterElem;

#endif

#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <iostream>
#include <DCDT_Agora.h>
#include <DCDT_Defs.h>
#include <sys/time.h>
#include <DCDT_Time.h>
#include <DCDT_Msg.h>

class DCDT_Member;
class DCDT_Httpd;

/// Structure needed to keep track of various member's subscriptions
typedef struct _SubsMsgElem
{
  DCDT_Mutex mtx;
  // Local and Remote are referred to the msg, i.e. to the its publisher

  /// Suscribers for local generated messages
  unsigned int SubscribersIfLocal[NUMINT_MEMBERS];

  /// Subscribers for remote generated messages
  unsigned int SubscribersIfRemote[NUMINT_MEMBERS];
}
SubsMsgTypeElem;

/*! \brief Informations about the single message stored.
 *
 *  When an elemet was deleted it checks if the message was read by all the members, if this is the last member the message is deleted, otherwise the readers counter is decremented.
 */
class DCDT_MsgListElem
{
 public:
  DCDT_MsgListElem ()
    {
#ifdef TOTAL_LOCK_WITH_READER_CLEANER
#else
      ToBeDeleted = false;
#endif
      MsgPtr = NULL;
      
      for(int i = 0; i < NUMINT_MEMBERS; i++)
	{
	  readers[i] = 0;
	};
      
#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL 
      numMemberPointing=0;
#endif
      
#ifdef TOTAL_LOCK
      numMemberPointing=0;
#endif
      
#ifdef TOTAL_LOCK_WITH_READER_CLEANER
      numMemberPointing=0;
#endif
    };
  
  ~DCDT_MsgListElem ()
    {
      next = NULL;
      prev = NULL;
      free (next);
      free (prev);
#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL 
      nextToBeDeleted = NULL;
#endif
      
      // If this is the last reader we have to delete the message.
      // Oterwise we have to decrement the readers counter
      if (MsgPtr->Get_Readers_Counter() == 1)  
	delete MsgPtr;
      else
	MsgPtr->Decrement_Readers_Counter(); 
    };
  
  struct timeval insert_time;  
  
#ifdef TOTAL_LOCK_WITH_READER_CLEANER
#else
  bool ToBeDeleted;
#endif
  
#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL 
  DCDT_Mutex mtx;
#endif
  
#ifdef SINGLE_LOCK_WITH_DISPATCHER
  DCDT_Mutex mtx;
#endif

#ifdef SINGLE_LOCK_WITH_DISPATCHER_WITH_CONDITION_VARIABLES
  DCDT_Mutex mtx;
#endif

  /// Flag: true -> message requested by a local user member ( member profile = MPM_USER )  
  bool local;
  
  int LocalPublisherID;
  
  DCDT_MsgListElem *next;
  DCDT_MsgListElem *prev;
  
  
#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL 
  // Pointer that allow the cleaner to follow a different path during the deleting stage; this permits to build a logically separated list built only of useless elements
  DCDT_MsgListElem *nextToBeDeleted;
#endif
  
  /// Array of bit reporting members subscribed to this message type
  unsigned int readers[NUMINT_MEMBERS];
  
  /// Type of the message
  unsigned int type_mask[NUMINT_MSGTYPE];

  DCDT_Msg *MsgPtr;
  
  // Counter for the members pointing at this element.
  // In the BRIDGE version the couter must be syncronized with a mutex
#ifdef TOTAL_LOCK
  unsigned int numMemberPointing;
#endif
  
#ifdef TOTAL_LOCK_WITH_READER_CLEANER
  unsigned int numMemberPointing;
#endif
  
#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL
  unsigned int numMemberPointing;
  DCDT_Mutex mtxnumMemberPointing;
#endif
  
  /** Mutex used when modifing the mask */
  DCDT_Mutex mtxmask;
  
#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL
  /** Mutex used to lock access at the previous and next elements */
  DCDT_Mutex mtxPrevPointer;
  DCDT_Mutex mtxNextPointer;
#endif 
};

// Structure for the registered members
typedef struct _MemberTableElem
{
  DCDT_Mutex mtx;
  
  // Flag: reporting whether or not this slot is hosting a member
  bool alive;
  
  // subcribed msg list 
  unsigned int SubscribedMsg[NUMINT_MSGTYPE];

  // Flag: reports that the susbscription mask was updated
  bool update;
  
#ifdef TOTAL_LOCK
  /// Pointer to the actual visited element
  DCDT_MsgListElem *ActReadMsgElem;

  /// Pointer to the last visited element
  DCDT_MsgListElem *LastVisitedMsgElem;
#endif
  
#ifdef TOTAL_LOCK_WITH_READER_CLEANER
  /// Pointer to the actual visited element
  DCDT_MsgListElem *ActReadMsgElem;

  // Pointer to the last visited element
  DCDT_MsgListElem *LastVisitedMsgElem;
#endif
  
#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL
  /// Pointer to the last visited element
  DCDT_MsgListElem *LastVisitedMsgElem;
#endif

  // Pointers to the MemberTable when every single member has a buffer
  // Allow the buffer slide
#ifdef SINGLE_LOCK_WITH_DISPATCHER
  DCDT_MsgListElem *HeadBuffer;
  DCDT_MsgListElem *TailBuffer;
#endif  
  
#ifdef SINGLE_LOCK_WITH_DISPATCHER_WITH_CONDITION_VARIABLES
  DCDT_MsgListElem *HeadBuffer;
  DCDT_MsgListElem *TailBuffer;
  DCDT_Condition condition_var;
#endif  

#ifdef SINGLE_LOCK_WITH_BUFFER
  DCDT_MsgListElem *HeadBuffer;
  DCDT_MsgListElem *TailBuffer;
#endif  
  
}
MemberTableElem;

/*! \brief Data structures and method to subscribe, insert, delete and read messages.
 *
 *  NOTE: This class isn't a member of the Agor, the real member that uses this class is DCDT_MsgManager.
 */
class DCDT_PostOffice
{
  friend class DCDT_MsgManager;
  friend class DCDT_Agora;
  friend class DCDT_Member;
  friend class DCDT_Httpd;

 public:
    
  DCDT_PostOffice (DCDT_Agora * agora)
    {
      myAgora = agora;
      NumOfMsgPresent = 0;

#ifdef _DEBUG_MSG
      myLog = new DCDT_Log ( "Postoffice.log" );
#endif

#ifdef SINGLE_LOCK_WITH_BUFFER
      
#else
      Head = new DCDT_MsgListElem;
#endif
      
#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL
      Tail = new DCDT_MsgListElem;
#endif
      
#ifdef SINGLE_LOCK_WITH_DISPATCHER 
      Tail = new DCDT_MsgListElem;
#endif
      
#ifdef SINGLE_LOCK_WITH_DISPATCHER_WITH_CONDITION_VARIABLES
      Tail = new DCDT_MsgListElem;
#endif
      
#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL
      
      Head->MsgPtr = NULL;
      Head->prev = NULL;
      
      Head->nextToBeDeleted=Tail;
      Head->ToBeDeleted=false; 
      Head->next = Tail;
      gettimeofday(&(Head->insert_time), NULL);   
      
      // MsgList Initialization
      Tail->MsgPtr = NULL;
      Tail->prev = Head;
      Tail->nextToBeDeleted=NULL;
      Tail->ToBeDeleted=false;
      Tail->next = NULL;
      Tail->numMemberPointing=1;
      gettimeofday(&(Tail->insert_time), NULL);
      
      IndiceHead;
      LastInsert=Head;
      prevDelete=Head;
      
#endif
      
#ifdef SINGLE_LOCK_WITH_BUFFER
      
      
      
      
#endif
      
#ifdef TOTAL_LOCK
      Tail = Head;
      Head->MsgPtr = NULL;
      Head->prev = NULL;
      Head->next = NULL;
      
#endif
      
#ifdef TOTAL_LOCK_WITH_READER_CLEANER
      
      Tail = Head;
      Head->MsgPtr = NULL;
      Head->prev = NULL;
      Head->next = NULL;
      
#endif

      
      
#ifdef SINGLE_LOCK_WITH_DISPATCHER
      
      // MsgList Initialization
      Head->MsgPtr = NULL;
      Head->prev = NULL;
      Head->ToBeDeleted=false; 
      Head->next = Tail;
      gettimeofday(&(Head->insert_time), NULL);   
      
      // MsgList Initialization
      Tail->MsgPtr = NULL;
      Tail->prev = Head;
      Tail->ToBeDeleted=false;
      Tail->next = NULL;
      gettimeofday(&(Tail->insert_time), NULL);
      
#endif
      
#ifdef SINGLE_LOCK_WITH_DISPATCHER_WITH_CONDITION_VARIABLES
      
      // MsgList Initialization
      
      Head->MsgPtr = NULL;
      Head->prev = NULL;
      Head->ToBeDeleted=false; 
      Head->next = Tail;
      gettimeofday(&(Head->insert_time), NULL);   
      
      // MsgList Initialization
      Tail->MsgPtr = NULL;
      Tail->prev = Head;
      Tail->ToBeDeleted=false;
      Tail->next = NULL;
      gettimeofday(&(Tail->insert_time), NULL);
      
#endif
      
#ifdef TEST_PART
      
      // Init list used during the test phase
      DCDT_TIME InitTime = GetTime();
      
      
      for (int j = 0; j < MAX_MEMBERS; j++)
	{
	  TestElem[j].prev = NULL;
	  TestElem[j].Data = InitTime;
	}
      
#endif


      // MemberTable initialization
      for (int i = 0; i < MAX_MEMBERS; i++)
	{
	  MemberTable[i].alive = false;
	  MemberTable[i].update = false;
	  
#ifdef TOTAL_LOCK 
	  MemberTable[i].ActReadMsgElem = Head;
	  MemberTable[i].LastVisitedMsgElem = Head; 
	  Head->numMemberPointing++;
#endif
	  
#ifdef TOTAL_LOCK_WITH_READER_CLEANER
	  MemberTable[i].ActReadMsgElem = Head;
	  MemberTable[i].LastVisitedMsgElem = Head; 
	  Head->numMemberPointing++;
#endif

#ifdef SINGLE_LOCK_WITH_BUFFER
	  MemberTable[i].TailBuffer = NULL;
	  MemberTable[i].HeadBuffer = NULL;
#endif

#ifdef SINGLE_LOCK_WITH_DISPATCHER
	  MemberTable[i].TailBuffer = NULL;
	  MemberTable[i].HeadBuffer = NULL;
#endif

#ifdef SINGLE_LOCK_WITH_DISPATCHER_WITH_CONDITION_VARIABLES
	  MemberTable[i].TailBuffer = NULL;
	  MemberTable[i].HeadBuffer = NULL;
#endif

#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL
	  MemberTable[i].LastVisitedMsgElem = Head; 
	  Head->numMemberPointing++;
#endif
	  
	  for (int j = 0; j < NUMINT_MSGTYPE; j++)
	    MemberTable[i].SubscribedMsg[j] = 0;
	}
      
      for (int i = 0; i < MAX_MSG_TYPE; i++)
	{
	  for (int j = 0; j < NUMINT_MEMBERS; j++) 
	    {
	      SubscribedTable[i].SubscribersIfLocal[j] = 0;
	      SubscribedTable[i].SubscribersIfRemote[j] = 0;
	    }
	}
      printf ("Postoffice created\n");
    };
  

  ~DCDT_PostOffice ()
    {
      // TODO: clean of MsgList

#ifdef _DEBUG_MSG
      delete myLog;
#endif

    };

#ifdef _DEBUG_MSG
  inline void LogMsg( DCDT_Msg *msg, int memb, char *prefix ) {
    char buffer[256];
    DCDT_MsgHeader *header;
    char *payload;
    int idmy=0,idx=0,payloadlen=0;

    if ( msg != NULL ) {
      header = msg->GetHeader();

      if ( (payload=(char *)msg->GetPayload()) != NULL ) {
	payloadlen=msg->ReadPayloadLen();
	
	while ( ( idx < payloadlen ) && ( idmy < sizeof(buffer)-7 ) ){
	  if ( isprint( payload[idx] ) )
	    idmy += sprintf ( &buffer[idmy], "%c", payload[idx] );
	  else
	    idmy += sprintf ( &buffer[idmy], "<0x%.2hhX>", 
			      (unsigned char)payload[idx] );
	  idx++;
	}
      }
    else
      sprintf ( buffer, "NULL" );
    
      if ( prefix != NULL )
	myLog->Log("%sMember: %d - Type: %d Agora:%u Payload: '%s' Len %d (%*.*s)\n", prefix, memb, header->type, header->AgoraID,
		   buffer, payloadlen, payloadlen, payloadlen, payload );
      else
	myLog->Log("Member: %d - Type: %d Agora:%u Payload: '%s' Len %d (%*.*s)\n", memb, header->type, header->AgoraID,
		   buffer, payloadlen, payloadlen, payloadlen, payload );
    }
    else {
      if ( prefix != NULL )
	myLog->Log("%sMember: %d - NULL\n", prefix, memb );
      else
	myLog->Log("Member: %d - NULL\n", memb );
    }
    
  };

  inline void LogMsg ( int memb, char *msg )
    {
      myLog->Log("Member: %d - %s\n", memb, msg );
    };
#endif  

  inline void AddMember(int Mem_ID);
  inline void AddSysMember(int Mem_ID);
  
  inline void DeleteMemberTableEntry(int Mem_ID );
  
  inline void AddMsg (DCDT_Msg * MP, int Mem_ID);
  
  inline DCDT_Msg * ReadNextMsg (int MemID); // TODO: Check if return "const DCDT_MSG *" is better

  inline void CleanupList ();
  
  inline void SubscribeMsgTypeID (int Member_ID, int MsgType);
  inline void UnSubscribeMsgTypeID (int Member_ID, int MsgType);
  inline void UnSubscribeAll( int Member_ID );
  
  inline DCDT_MsgListElem *SearchMsg (DCDT_Msg *);
  
  inline void UpdateSubscribersMask (int MemID);
  inline bool ReadSubscribedMsgMask(unsigned int *mask, int MemID);
  inline bool UpdateSubs(int MemID);
    
 private:
  
  inline void SetReaders (DCDT_MsgListElem * Elem, int num_type); 
  
#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL 
  
  inline int CheckIfMsgRead (DCDT_MsgListElem *); 

  inline void SetMsgDeletable (int Member_ID, DCDT_Msg *);
  
  inline DCDT_MsgListElem* GoToNextListMsg (DCDT_MsgListElem* Elem,int MemID);  
  
  inline void SetListElemDeletable (DCDT_MsgListElem *Elem);

  inline DCDT_MsgListElem* RemoveListElemDeletable (DCDT_MsgListElem *);

  inline void SetMsgRead (DCDT_MsgListElem * Elem, int Mem_ID);
  
  DCDT_MsgListElem *Head, *Tail;
  
  /*! \brief Last useful element found by the cleaner during the first step.
   *
   *  It's the base element to build a bridge around a useless element.
   */
  DCDT_MsgListElem *Indice;

  /// Last element of the massage list as seen by the cleaner at the beginning of a clean cycle; messages inserted during the clean cycle are inserted after this
  DCDT_MsgListElem *LastInsert;
  
  // Pointer used only during the second step of the clear task, it's used to store the nextToBeDeleted pointer of the last deleted message
  DCDT_MsgListElem *prevDelete;
  
#endif
  
  
#ifdef _DEBUG_MSG
  DCDT_Log *myLog;
#endif
  
#ifdef TOTAL_LOCK 
  
  inline int CheckIfMsgRead (DCDT_MsgListElem *);
  inline void SetActMsgDeletable (int Member_ID); 
  inline void RemoveListElement (DCDT_MsgListElem *);  
  
  DCDT_Mutex insert_mtx;
  DCDT_MsgListElem 
    /// Head of the message list
    *Head,
    /// Tail of the message list
    *Tail;
#endif
  
#ifdef TOTAL_LOCK_WITH_READER_CLEANER 
  inline int CheckIfMsgRead (DCDT_MsgListElem *);  
  
  inline void SetMsgRead (DCDT_MsgListElem * Elem, int Mem_ID);  
  inline void RemoveListElement (DCDT_MsgListElem *);  
  DCDT_Mutex insert_mtx;
  DCDT_MsgListElem *Head, *Tail;
#endif
  
  
#ifdef SINGLE_LOCK_WITH_DISPATCHER
  
  inline void RemoveBufferElement(int MemID);
  DCDT_MsgListElem *Head, *Tail;

  // Pointer to the list's last element.
  // If we set this pointer at the beginning of a clear cycle the cleaner can be stopped before touching elements inserted during the cycle.
  DCDT_MsgListElem *LastInsert;

#endif
  

#ifdef SINGLE_LOCK_WITH_DISPATCHER_WITH_CONDITION_VARIABLES
  
  inline void RemoveBufferElement(int MemID);
  DCDT_MsgListElem *Head, *Tail;

  // Pointer to the list's last element.
  // If we set this pointer at the beginning of a clear cycle the cleaner can be stopped before touching elements inserted during the cycle.
  DCDT_MsgListElem *LastInsert;

#endif
  
#ifdef SINGLE_LOCK_WITH_BUFFER
  inline void RemoveBufferElement(int MemID);
#endif
  
  inline int MatchMsgTypeMask (int Member_ID, DCDT_MsgListElem *);
  inline int MatchReadersMask (int Member_ID, DCDT_MsgListElem *);
  
  inline void SetType (DCDT_MsgListElem * Elem, int type);

  // Functions used during the test phase
#ifdef TEST_PART
  inline void InsertListTime (int MemID, int TypeOp);
  inline void VisualList ();
  inline void SaveList (char* nomefile);
#else
  inline void InsertListTime (int MemID, int TypeOp){}; 
  inline void VisualList (){};
  inline void SaveList (char *nomefile){};
  
#endif
    
  inline bool CheckTimeoutElem(DCDT_MsgListElem *); 
  
  inline int GetTypeFromMask (DCDT_MsgListElem * Elem);
  
  /*! \brief For each message type we have the corresponding member mask.
   *
   *  A member mask use a sequence of bit to repesent all the members, if a bit is set to 1 the corresponding member is subscribed to this type of messages.
   *  Ex: 00010001 -> Members 1 and 5 are subscribed to this message type.
   *  The member mask is composed of two parts local and remote, the local one was used when a message was locally originated or was received from a BRIDGE member; the remote one was used when the message was received from a STATIC or LINK member.
   */
  SubsMsgTypeElem SubscribedTable[MAX_MSG_TYPE];
  
  /// Mutex to access SubscribedMsgMask
  DCDT_Mutex subscribed_mtx;

  /// Message types requested by local members used by LinkTx members
  unsigned int SubscribedMsgMask[NUMINT_MSGTYPE];
  
  // Mutex to access UpdateMask_flag and MemberTable[i].update
  DCDT_Mutex UpdateMask_mtx;

  /// Flag: true when SubscribedMsgMask was updated
  bool UpdateMask_flag;

#ifdef TEST_PART 

  // Lists used during the test phase, each member has a private list to avoid
  // concurrent access
  TimeTesterElem TestElem[MAX_MEMBERS];
  
#endif
  
  MemberTableElem MemberTable[MAX_MEMBERS];
  
  int NumOfMsgPresent;

  DCDT_Agora *myAgora;  
};




inline void DCDT_PostOffice::AddMember(int Mem_ID) {
  MemberTable[Mem_ID].alive = true;
};

inline void DCDT_PostOffice::AddSysMember(int Mem_ID) {
  MemberTable[Mem_ID].alive = true;
};

inline void DCDT_PostOffice::DeleteMemberTableEntry(int Mem_ID){
  
  // TODO
  MemberTable[Mem_ID].alive = false;
  
};




#ifdef SINGLE_LOCK_WITH_BUFFER

/*! \brief Add a new message to the PostOffice queue.
 *
 *  When the member \aMem_ID wants to insert a new message the same member checks the subscription table and when a member is subscribed to those messages copy the message into the member message queue; this function also updates the message with the number of members subscribed to this message type.
 */
inline void DCDT_PostOffice::AddMsg (DCDT_Msg * MP, int Mem_ID)
{
  int msg_type;
  int templ,tempr,temp;
  int MemberID = 0;
  
  InsertListTime(Mem_ID,WRITEINIT);
  DCDT_MsgListElem *newMsgListElem;
  
  DBG_PO_PRINTF(("\nAddMsg %d\n",Mem_ID));
  
  msg_type = MP->ReadType ();
  if ( msg_type > MAX_MSG_TYPE ) {
    // Quando riprende un link interno da un'agora' fermata e fatta ripartire
    return;
  }

  // Check if someone is subscribed to this message
  for (int i=0; i<NUMINT_MEMBERS;i++) { 
    for (int t=1;t<=SIZEOF_INTEGER;t++) { 
      MemberID++;
      templ = SubscribedTable[msg_type].SubscribersIfLocal[i];
      templ = templ << (SIZEOF_INTEGER - t);
      templ = templ >> (SIZEOF_INTEGER-1);
      
      tempr = SubscribedTable[msg_type].SubscribersIfRemote[i];
      tempr = tempr << (SIZEOF_INTEGER - t);
      tempr = tempr >> (SIZEOF_INTEGER-1); 
      
      temp = (templ | tempr);
      
      if (temp && (MemberID != Mem_ID)) {

	// We have a member interested so we haveto allocate a new element for the local message queue and copy the message pointer
	DCDT_Msg *tmp;
	tmp = new DCDT_Msg(msg_type);
	
	newMsgListElem = new (DCDT_MsgListElem);
	newMsgListElem->MsgPtr = MP; 
	newMsgListElem->next = NULL; 

	MP->Increment_Readers_Counter();
	gettimeofday(&(newMsgListElem->insert_time), NULL);
	SetType (newMsgListElem, msg_type);
	InsertListTime(Mem_ID,WRITELOCKINIT);
	MemberTable[MemberID].mtx.lock();
	InsertListTime(Mem_ID,WRITELOCKEND);
	if (MemberTable[MemberID].HeadBuffer)
	  MemberTable[MemberID].TailBuffer->next = newMsgListElem;
	else
	  MemberTable[MemberID].HeadBuffer = newMsgListElem;

	// Check ???
	MemberTable[MemberID].TailBuffer = newMsgListElem;

	NumOfMsgPresent++;
	MemberTable[MemberID].mtx.unlock();
      }
    }
  }

  MemberID = 0;
  InsertListTime(Mem_ID,WRITEEND);
  
  DBG_PO_PRINTF (("(Messaggi in coda: %d)\n", NumOfMsgPresent));
  DBG_PO_PRINTF( ("AddMsg(%d): EXIT\n",Mem_ID));
};

#endif


#ifdef TOTAL_LOCK

/*! \brief Add a new message to the PostOffice queue.
 *
 * To add a new message to the PostOffice queue a member must have to acquire the write lock to the queue then he can add the message and release the lock.
 */
inline void DCDT_PostOffice::AddMsg (DCDT_Msg * MP, int Mem_ID)
{
  int msg_type;
  InsertListTime(Mem_ID,WRITEINIT); 
  InsertListTime(Mem_ID,WRITELOCKINIT);
  insert_mtx.lock ();	
  InsertListTime(Mem_ID,WRITELOCKEND);
  DBG_PO_PRINTF(("\nAddMsg %d\n",Mem_ID));
  DBG_PRINTF( ("Locked Insert\n"));

  DCDT_MsgListElem *newMsgListElem = new (DCDT_MsgListElem);
  newMsgListElem->MsgPtr = MP;
  newMsgListElem->LocalPublisherID = Mem_ID;
  newMsgListElem->prev = NULL;
  newMsgListElem->next = NULL;
  newMsgListElem->local = myAgora->CheckMemberProfile(Mem_ID, MPM_USER);
  newMsgListElem->numMemberPointing = 0; 
  gettimeofday(&(newMsgListElem->insert_time), NULL);
  
  msg_type = MP->ReadType ();
  
  SetType (newMsgListElem, msg_type);

  SetReaders (newMsgListElem, msg_type);

  newMsgListElem->prev = Tail;
  Tail->next = newMsgListElem;
  Tail = newMsgListElem;
  
  NumOfMsgPresent++;

  DBG_PO_PRINTF (("(Messages in queue: %d)\n", NumOfMsgPresent));
    
  insert_mtx.unlock ();
  InsertListTime(Mem_ID,WRITEEND);
  DBG_PRINTF( ("Unlocked Insert\n"));
  DBG_PO_PRINTF( ("AddMsg(): EXIT\n"));
};


#endif

#ifdef TOTAL_LOCK_WITH_READER_CLEANER

/*! \brief Add a new message to the PostOffice queue.
 *
 * To add a new message to the PostOffice queue a member must have to acquire the write lock to the queue then he can add the message and release the lock.
 */
inline void DCDT_PostOffice::AddMsg (DCDT_Msg * MP, int Mem_ID)
{
  int msg_type;
  
  InsertListTime(Mem_ID,WRITEINIT); 
  InsertListTime(Mem_ID,WRITELOCKINIT);
  insert_mtx.lock (); 
  InsertListTime(Mem_ID,WRITELOCKEND);
  DBG_PO_PRINTF(("\nAddMsg %d\n",Mem_ID));
  DBG_PRINTF( ("Locked Insert\n"));
  
  DCDT_MsgListElem *newMsgListElem = new (DCDT_MsgListElem);
  newMsgListElem->MsgPtr = MP;
  newMsgListElem->LocalPublisherID = Mem_ID;
  newMsgListElem->prev = NULL;
  newMsgListElem->next = NULL;
  newMsgListElem->local = myAgora->CheckMemberProfile(Mem_ID, MPM_USER);
  newMsgListElem->numMemberPointing = 0; 
  gettimeofday(&(newMsgListElem->insert_time), NULL);
  
  msg_type = MP->ReadType ();
  
  SetType (newMsgListElem, msg_type);
  
  DBG_PRINTF( ("Lock Subscribed\n"));
    
  SetReaders (newMsgListElem, msg_type);
 
  DBG_PRINTF( ("Unlocked Subscribed\n"));

  // Add the new message to the tail of the queue
  newMsgListElem->prev = Tail;
  Tail->next = newMsgListElem;
  Tail = newMsgListElem;
  
  NumOfMsgPresent++;

  DBG_PO_PRINTF (("(Messages in queue: %d)\n", NumOfMsgPresent));

  insert_mtx.unlock ();
  InsertListTime(Mem_ID,WRITEEND);
  DBG_PRINTF( ("Unlocked Insert\n"));
  DBG_PO_PRINTF( ("AddMsg(): EXIT\n"));
};
#endif






#ifdef SINGLE_LOCK_WITH_DISPATCHER

/*! \brief Add a message to the PostOffice.
 *
 *  Add the new message to the queue owned by the dispatcher, in the next dispatch cycle the message will be sent to each subscribed member; each member will receive a copy of the message in his local queue.
 */
inline void DCDT_PostOffice::AddMsg (DCDT_Msg * MP, int Mem_ID)
{
  int msg_type;
  
  DBG_PO_PRINTF(("\nAddMsg %d\n",Mem_ID));

  InsertListTime(Mem_ID,WRITEINIT);

  // Create the new element
  DCDT_MsgListElem *newMsgListElem = new (DCDT_MsgListElem);
  MP->Increment_Readers_Counter();
  newMsgListElem->MsgPtr = MP;
  
  newMsgListElem->LocalPublisherID = Mem_ID;
  newMsgListElem->prev = NULL;
  newMsgListElem->next = NULL;
  
  
  newMsgListElem->local = myAgora->CheckMemberProfile(Mem_ID, MPM_USER);
  
  gettimeofday(&(newMsgListElem->insert_time), NULL);
  
  msg_type = MP->ReadType ();
  
  SetType (newMsgListElem, msg_type);
  
  // Lock the dispatcher's queue
  InsertListTime(Mem_ID,WRITELOCKINIT);
  Tail->mtx.lock();
  InsertListTime(Mem_ID,WRITELOCKEND);

  // Insert the new message before the tail element
  newMsgListElem->next = Tail; 
  newMsgListElem->prev = Tail->prev;  
  Tail->prev->next = newMsgListElem;  
  Tail->prev = newMsgListElem;

  LastInsert = newMsgListElem; 
 
  // Now we can unlock the queue
  Tail->mtx.unlock(); 

  NumOfMsgPresent++;
  
  InsertListTime(Mem_ID,WRITEEND);
  DBG_PO_PRINTF (("(Messages in queue: %d)\n", NumOfMsgPresent));
  DBG_PO_PRINTF( ("AddMsg(%d): EXIT\n",Mem_ID));
};

#endif

#ifdef SINGLE_LOCK_WITH_DISPATCHER_WITH_CONDITION_VARIABLES

/*! \brief Add a message to the PostOffice.
 *
 *  Add the new message to the queue owned by the dispatcher, in the next dispatch cycle the message will be sent to each member subscribed; each member will receive a copy of the message in his local queue.
 */ 
inline void DCDT_PostOffice::AddMsg (DCDT_Msg * MP, int Mem_ID)
{
  int msg_type;
  
  DBG_PO_PRINTF(("\nAddMsg %d\n",Mem_ID));
  
  InsertListTime(Mem_ID,WRITEINIT);
  DCDT_MsgListElem *newMsgListElem = new (DCDT_MsgListElem);
  MP->Increment_Readers_Counter();
  newMsgListElem->MsgPtr = MP;
  
  newMsgListElem->LocalPublisherID = Mem_ID;
  newMsgListElem->prev = NULL;
  newMsgListElem->next = NULL;
  
  
  newMsgListElem->local = myAgora->CheckMemberProfile(Mem_ID, MPM_USER);
  
  gettimeofday(&(newMsgListElem->insert_time), NULL);
  
  msg_type = MP->ReadType ();
  
  SetType (newMsgListElem, msg_type);
  
  // Lock the dispatcher's queue and insert the new message
  InsertListTime(Mem_ID,WRITELOCKINIT);
  Tail->mtx.lock();
  InsertListTime(Mem_ID,WRITELOCKEND);

  // Add the message to the dispatcher's queue, just before the last element
  newMsgListElem->next = Tail; 
  newMsgListElem->prev = Tail->prev;
  Tail->prev->next = newMsgListElem;
  Tail->prev = newMsgListElem;

  LastInsert = newMsgListElem; 
 
  Tail->mtx.unlock(); 
  
  NumOfMsgPresent++;
  
  InsertListTime(Mem_ID,WRITEEND);

  DBG_PO_PRINTF (("(Messages in queue: %d)\n", NumOfMsgPresent));
  DBG_PO_PRINTF( ("AddMsg(%d): EXIT\n",Mem_ID));
};

#endif


#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL
/*! \brief Add a message to the PostOffice.
 *
 *  Add the new message to the Postoffice queue just before the tail element.
 *  Locks only the pointer that we have to change, with this method we allow other members to access all the other elements.
 */ 
inline void DCDT_PostOffice::AddMsg (DCDT_Msg * MP, int Mem_ID)
{
  int msg_type;
  
  DBG_PO_PRINTF(("\nAddMsg %d\n",Mem_ID));
  
  InsertListTime(Mem_ID,WRITEINIT);

  // Create a new message element ad set it up
  DCDT_MsgListElem *newMsgListElem = new (DCDT_MsgListElem);
 
  newMsgListElem->MsgPtr = MP;  
  newMsgListElem->LocalPublisherID = Mem_ID;
  newMsgListElem->prev = NULL;
  newMsgListElem->next = NULL;
  newMsgListElem->nextToBeDeleted = NULL; 
  
  newMsgListElem->local = myAgora->CheckMemberProfile(Mem_ID, MPM_USER);
  
  gettimeofday(&(newMsgListElem->insert_time), NULL);
  
  msg_type = MP->ReadType ();
  
  SetType (newMsgListElem, msg_type);
  
  InsertListTime(Mem_ID,WRITELOCKINIT);
  
  // Lock only the pointers that we need to change ( tail and new element )
  Tail->mtx.lock();
  Tail->prev->mtxNextPointer.lock();
  Tail->mtxPrevPointer.lock();
  newMsgListElem->mtxPrevPointer.lock(); 
  newMsgListElem->mtxNextPointer.lock();
  
  InsertListTime(Mem_ID,WRITELOCKEND);

  SetReaders (newMsgListElem, msg_type);
  
  // Add the new message before the tail element
  newMsgListElem->next = Tail; 
  newMsgListElem->prev = Tail->prev;   
  newMsgListElem->nextToBeDeleted = Tail; 
  
  Tail->prev->next = newMsgListElem;  
  Tail->prev->nextToBeDeleted = newMsgListElem;  
  Tail->prev = newMsgListElem;

  LastInsert = newMsgListElem; 

  // Unlock the pointers
  newMsgListElem->prev->mtxNextPointer.unlock();
  newMsgListElem->mtxPrevPointer.unlock(); 
  newMsgListElem->mtxNextPointer.unlock();
  Tail->mtxPrevPointer.unlock();
  Tail->mtx.unlock(); 

  NumOfMsgPresent++;
  InsertListTime(Mem_ID,WRITEEND);
  
  DBG_PO_PRINTF (("(Messages in queue: %d)\n", NumOfMsgPresent));
  
  DBG_PO_PRINTF( ("AddMsg(%d): EXIT\n",Mem_ID));
};
#endif

#ifdef TOTAL_LOCK

/*! \brief Check if all the members subscribed to a message have already read it.
    \return TRUE  - All the members have read the message ( ??? )
            FALSE - Not all the members have read the message 
 */
inline int DCDT_PostOffice::CheckIfMsgRead (DCDT_MsgListElem * elem) {
  int temp;
  
  DBG_PO_PRINTF( ("\nCheckIfMsgRead\n"));
  temp = 0;
  
  DBG_PRINTF( ("Readers:"));
  for (int i = 0; i < NUMINT_MEMBERS; i++) {
    
    DBG_PRINTF(("[%d]", elem->readers[i]));
    
    temp |= elem->readers[i];
  }
  
  DBG_PRINTF((" temp =%d, !temp=%d\n", temp, !temp));
  
  return !temp;
};

#endif

#ifdef TOTAL_LOCK_WITH_READER_CLEANER

/*! \brief Check if all the members subscribed to a message have already read it.
    \return TRUE  - All the members have read the message ( ??? )
            FALSE - Not all the members have read the message 
 */
inline int DCDT_PostOffice::CheckIfMsgRead (DCDT_MsgListElem * elem) 
{
  int temp;
  
  DBG_PO_PRINTF( ("\nCheckIfMsgRead\n"));
  temp = 0;
  
  DBG_PRINTF( ("Readers:"));
  for (int i = 0; i < NUMINT_MEMBERS; i++) 
    {
    
      DBG_PRINTF(("[%d]", elem->readers[i]));
      
      temp |= elem->readers[i];
      
    }
  
  DBG_PRINTF((" temp =%d, !temp=%d\n", temp, !temp));
  
  return !temp;
};
#endif



#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL

/*! \brief Check if all the members subscribed to a message have already read it.
    \return TRUE  - All the members have read the message ( ??? )
            FALSE - Not all the members have read the message 
 */
inline int DCDT_PostOffice::CheckIfMsgRead (DCDT_MsgListElem * Elem) 
{
  int temp;
  
  DBG_PO_PRINTF( ("\nCheckIfMsgRead\n"));
  temp = 0;
  
  DBG_PO_PRINTF( ("Readers:"));
  
  for (int i = 0; i < NUMINT_MEMBERS; i++) {
    
    DBG_PO_PRINTF(("[%d]", Elem->readers[i]));
    
    temp |= Elem->readers[i];
  }
  
  DBG_PRINTF((" temp =%d, !temp=%d\n", temp, !temp));
  return !temp;
};
#endif

#ifdef SINGLE_LOCK_WITH_BUFFER

/*! \brief Read next message for the member \a MemID.
 *
 *  Check if there is an unread message in the member queue; if we found one we have to mark it as deletable and return a pointer to it.
 *  If, during the queue slide, we find a message marked as deletable we have to delete it and go on.
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg *DCDT_PostOffice::ReadNextMsg (int MemID)
{
  DCDT_MsgListElem *Elem;
  DBG_PO_PRINTF( ("\nReadNextMsg %d\n",MemID));
  InsertListTime(MemID,READINIT);
  if (MemberTable[MemID].HeadBuffer) {
    InsertListTime(MemID,READLOCKINIT);
    MemberTable[MemID].mtx.lock(); 
    InsertListTime(MemID,READLOCKEND);
    Elem = MemberTable[MemID].HeadBuffer;

    // This message has to be deleted ?
    if (Elem->ToBeDeleted || CheckTimeoutElem(Elem)) {

      // Yes, deleted it and check the next message
      InsertListTime(MemID,DELETEINIT);
      RemoveBufferElement(MemID);
      InsertListTime(MemID,DELETEEND);
      Elem = MemberTable[MemID].HeadBuffer;
      if (Elem) { 
	Elem->ToBeDeleted = true;
	MemberTable[MemID].mtx.unlock();
	InsertListTime(MemID,READDONEEND);
        DBG_PO_PRINTF( ("\nReadNextMsg %d: EXIT with Msg (1)\n",MemID));
	return (Elem->MsgPtr);
      }

      // No, new message available
      MemberTable[MemID].mtx.unlock();
      InsertListTime(MemID,READEND);
      DBG_PO_PRINTF( ("\nReadNextMsg %d: EXIT Null (1)\n",MemID));
      return ((DCDT_Msg *) NULL);
    }
    
    // Mark the message as deletable and return a pointer to it
    Elem->ToBeDeleted = true;
    MemberTable[MemID].mtx.unlock(); 
    
    InsertListTime(MemID,READDONEEND);
    DBG_PO_PRINTF( ("\nReadNextMsg %d: EXIT with Msg (2)\n",MemID));
    return (Elem->MsgPtr);
  }

  // No new messages available
  DBG_PO_PRINTF( ("\nReadNextMsg %d: EXIT Null (2)\n",MemID));
  InsertListTime(MemID,READEND);
  return ((DCDT_Msg *) NULL);
};
#endif

#ifdef TOTAL_LOCK

/*! \brief Read next message for the member \a MemID.
 *
 *  Check if there is an unread message in the message queue; if we found one we have to return a pointer to it.
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg *
DCDT_PostOffice::ReadNextMsg (int MemID)
{
  DCDT_MsgListElem *nextElem;

  InsertListTime(MemID,READINIT); 
  InsertListTime(MemID,READLOCKINIT);
  insert_mtx.lock ();
  InsertListTime(MemID,READLOCKEND);
  DBG_PO_PRINTF( ("\nReadNextMsg %d\n",MemID));
 
  DBG_PO_PRINTF ( ("ReadNextMsg: insert_mtx LOCKED\n"));
 
  // Slide the message list until we find an interesting message
  nextElem = MemberTable[MemID].LastVisitedMsgElem->next;
 
  while ((nextElem != NULL)) {

    MemberTable[MemID].LastVisitedMsgElem->numMemberPointing--;
    MemberTable[MemID].LastVisitedMsgElem = nextElem;
    nextElem->numMemberPointing++;
    
    if (MatchMsgTypeMask (MemID, nextElem) && MatchReadersMask (MemID, nextElem)) {
      // We are subscribed to this type of messages, we have to return it
      MemberTable[MemID].ActReadMsgElem = nextElem;
      insert_mtx.unlock ();
      DBG_PO_PRINTF ( ("ReadNextMsg: insert_mtx UNLOCKED\n")); 
      InsertListTime(MemID,READDONEEND);
      return (nextElem->MsgPtr);
    }
     
    nextElem = MemberTable[MemID].LastVisitedMsgElem->next;
  }

  // No new messages found
  insert_mtx.unlock ();
  DBG_PO_PRINTF ( ("ReadNextMsg: insert_mtx UNLOCKED\n"));

  DBG_PO_PRINTF( ("\nReadNextMsg: EXIT\n"));
  InsertListTime(MemID,READEND);
  return ((DCDT_Msg *) NULL);
};
#endif

#ifdef TOTAL_LOCK_WITH_READER_CLEANER

/*! \brief Read next message for the member \a MemID.
 *
 *  Check if there is an unread message in the message queue; if we found one we have to mark it as deletable and return a pointer to it.
 *  If we find a message read by all the members subscribed to it and we are the only member using it we have to delete it and go to the next message.
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg * DCDT_PostOffice::ReadNextMsg (int MemID)
{
  DCDT_MsgListElem *nextElem;
  DCDT_MsgListElem *tempElem;

  InsertListTime(MemID,READINIT); 
  InsertListTime(MemID,READLOCKINIT);
  insert_mtx.lock (); 
  InsertListTime(MemID,READLOCKEND);
  DBG_PO_PRINTF( ("\nReadNextMsg %d\n",MemID));
  
  DBG_PO_PRINTF ( ("ReadNextMsg: insert_mtx LOCKED\n"));

  nextElem = MemberTable[MemID].ActReadMsgElem;

  if (nextElem != Head)
    if  ((CheckIfMsgRead (nextElem)) && 
	 (nextElem->numMemberPointing == 1)) {

      // The first message has to be deleted because all the member have read it and only us are pointing to it
      tempElem=nextElem;
      nextElem=MemberTable[MemID].LastVisitedMsgElem->prev;

      MemberTable[MemID].ActReadMsgElem = nextElem;
      MemberTable[MemID].LastVisitedMsgElem = nextElem;
      
      nextElem->numMemberPointing++;
      InsertListTime(MemID,DELETEINIT);
      RemoveListElement (tempElem);
      InsertListTime(MemID,DELETEEND);
    }
    else
      nextElem = MemberTable[MemID].LastVisitedMsgElem->next;
  else
    nextElem = MemberTable[MemID].LastVisitedMsgElem->next;

  // Search into the message list an that is useful for us
  while ((nextElem != NULL)) {
    MemberTable[MemID].LastVisitedMsgElem->numMemberPointing--;
    MemberTable[MemID].LastVisitedMsgElem = nextElem;	
    MemberTable[MemID].ActReadMsgElem = nextElem;
    nextElem->numMemberPointing++;

    // We are subscribed to this type of messages ?
    if (MatchMsgTypeMask (MemID, nextElem) && MatchReadersMask (MemID, nextElem)) {
      // Yes, we have to mark the message as read and return a pointer to it
      int offset,pos,mask;

      offset = (MemID - 1) / SIZEOF_INTEGER;
      pos = MemID - (offset * SIZEOF_INTEGER);
      mask = ~(1 << (pos - 1));

      MemberTable[MemID].ActReadMsgElem->readers[offset] &= mask;
      
      insert_mtx.unlock ();	
      DBG_PO_PRINTF ( ("ReadNextMsg: insert_mtx UNLOCKED\n"));
      InsertListTime(MemID,READDONEEND);
      return (nextElem->MsgPtr);
    }
    
    // We aren't interest in this type of messages, so go to the next
    nextElem = MemberTable[MemID].LastVisitedMsgElem->next;
  }

  // We haven't found a message
  DBG_PO_PRINTF ( ("ReadNextMsg: MemberTable[%d].mtk UNLOCKED\n", MemID));
  insert_mtx.unlock ();

  DBG_PO_PRINTF( ("\nReadNextMsg: EXIT\n"));
  InsertListTime(MemID,READEND);
  return ((DCDT_Msg *) NULL);
};
#endif



#ifdef SINGLE_LOCK_WITH_DISPATCHER

/*! \brief Read next message for the member \a MemID.
 *
 *  Check if there is an unread message in the member queue; if we found one we have to mark it as deletable and return a pointer to it.
 *  If we find a message marked as deletable we have to delete it and go to the next message.
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg * DCDT_PostOffice::ReadNextMsg (int MemID)
{
  
  DCDT_MsgListElem *Elem;
  DCDT_MsgListElem *tmpElem;
  DBG_PO_PRINTF( ("\nReadNextMsg %d\n",MemID));

  InsertListTime(MemID,READINIT);

  // If there is a message to be read ...
  if (MemberTable[MemID].HeadBuffer) { 
    InsertListTime(MemID,READLOCKINIT);
    MemberTable[MemID].mtx.lock();
    InsertListTime(MemID,READLOCKEND);
    Elem = MemberTable[MemID].HeadBuffer;
    
    // Check if the message has to be deleted
    if (Elem->ToBeDeleted || CheckTimeoutElem(Elem) ) { 
      InsertListTime(MemID,DELETEINIT); 
      RemoveBufferElement(MemID);
      InsertListTime(MemID,DELETEEND);
      Elem = MemberTable[MemID].HeadBuffer;

      // we have deleted the message, but now there is another message ?
      if (Elem) {

	// Yes, mark it as deletable and return it
	Elem->ToBeDeleted = true;
	MemberTable[MemID].mtx.unlock();
	InsertListTime(MemID,READDONEEND);
	return (Elem->MsgPtr);
      }
      
      // No, no new messages available
      MemberTable[MemID].mtx.unlock();
      InsertListTime(MemID,READEND);
      return ((DCDT_Msg *) NULL);
    }
    
    // Mark the message as deletable and return it
    Elem->ToBeDeleted = true;
    MemberTable[MemID].mtx.unlock(); 
    InsertListTime(MemID,READDONEEND);
    return (Elem->MsgPtr);
  }

  // No messages available.
  DBG_PO_PRINTF( ("\nReadNextMsg%d: EXIT\n",MemID));
  InsertListTime(MemID,READEND);
  return ((DCDT_Msg *) NULL);
};
#endif

#ifdef SINGLE_LOCK_WITH_DISPATCHER_WITH_CONDITION_VARIABLES

/*! \brief Delete read messages then read next message for the member \a MemID.
 *
 *  Check if there is an unread message in the member queue; if we found one we have to mark it as deletable and return a pointer to it.
 *  If the queue is empty we have to wait until we receive a signal from our dispatcher; the signal means that we have a new message in our queue. Before returning the message we have to check if the current message available has to be deleted or his timeout was experied.
 *  \return NULL - The only message available was deleted so no message is available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg *DCDT_PostOffice::ReadNextMsg (int MemID)
{
  
  DCDT_MsgListElem *Elem;
  DBG_PO_PRINTF( ("\nReadNextMsg %d\n",MemID));

  InsertListTime(MemID,READLOCKINIT);
  MemberTable[MemID].condition_var.lock(); 
  InsertListTime(MemID,READLOCKEND); 

  // If we don't have a message wait a signal from our dispatcher
  while ( MemberTable[MemID].HeadBuffer == NULL ) {
    InsertListTime(MemID,WAITINIT);
    MemberTable[MemID].condition_var.wait();
    InsertListTime(MemID,WAITEND);
  }

  MemberTable[MemID].condition_var.unlock();
  InsertListTime(MemID,READINIT);
  InsertListTime(MemID,READLOCKINIT);
  MemberTable[MemID].mtx.lock();
  InsertListTime(MemID,READLOCKEND);
  Elem = MemberTable[MemID].HeadBuffer;
 
  if (Elem->ToBeDeleted || CheckTimeoutElem(Elem) ) {

    // If the first queued message has to be deleted delete it ....
    InsertListTime(MemID,DELETEINIT); 
    RemoveBufferElement(MemID);
    InsertListTime(MemID,DELETEEND);

    // .. then read the next message
    // If we don't have a message wait a signal from our dispatcher
    while ( MemberTable[MemID].HeadBuffer == NULL ) {
      MemberTable[MemID].mtx.unlock();
      InsertListTime(MemID,READDONEEND);
      InsertListTime(MemID,READLOCKINIT);
      MemberTable[MemID].condition_var.lock();
      InsertListTime(MemID,READLOCKEND);
      InsertListTime(MemID,WAITINIT);
      MemberTable[MemID].condition_var.wait();
      InsertListTime(MemID,WAITEND);
      MemberTable[MemID].mtx.lock();
      InsertListTime(MemID,READLOCKEND);
      MemberTable[MemID].condition_var.unlock();
    }

    Elem = MemberTable[MemID].HeadBuffer;
    if (Elem) {	      
      Elem->ToBeDeleted = true;
      MemberTable[MemID].mtx.unlock();
      InsertListTime(MemID,READDONEEND);

      DBG_PO_PRINTF( ("\nDelete&ReadNextMsg %d: EXIT\n",MemID));
      return (Elem->MsgPtr);
    }
    
    // The only message available was deleted
    MemberTable[MemID].mtx.unlock();
    InsertListTime(MemID,READEND);
    return ((DCDT_Msg *) NULL);
  }
  
  Elem->ToBeDeleted = true;
  MemberTable[MemID].mtx.unlock(); 
  InsertListTime(MemID,READDONEEND);
  return (Elem->MsgPtr);
  DBG_PO_PRINTF( ("\nReadNextMsg %d: EXIT\n",MemID));
};
#endif



#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL
//
//ANDREAS: QUESTO METODO DA' IN USCITA IL PRIMO ELEMENTO DELLA LISTA CHE IL MEMBRO MEMID
//PUO' LEGGERE. SE NON CI SONO ELEMENTI PER QUEL DETERMINATO MEMBRO RITORNA NULL.
//
// **********************************************
// ReadNextMsg
/*! \brief Read next message for the member \a MemID.
 *
 *  Check if there is an unread message in the message queue.
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg *DCDT_PostOffice::ReadNextMsg (int MemID)
     
{
  DCDT_MsgListElem *nextElem;
  DCDT_MsgListElem *tmpElem;

  DBG_PO_PRINTF( ("\nReadNextMsg %d\n",MemID));
  
  InsertListTime(MemID,READINIT);

  // Start the queue slide from the last visited element
  nextElem=MemberTable[MemID].LastVisitedMsgElem;

  while ((nextElem != Tail->prev)) {
    // Safe method to go to the next element 
    nextElem=GoToNextListMsg(nextElem,MemID);
    
    // We are subscribed to this type of messages ?
    if (MatchMsgTypeMask (MemID, nextElem) && MatchReadersMask (MemID, nextElem)) {
      // Yes, mark the message as read and return a pointer to it
      SetMsgRead(nextElem,MemID);
      InsertListTime(MemID,READDONEEND);
      return (nextElem->MsgPtr);
    }
    
  }
  
  // No new message available
  DBG_PO_PRINTF( ("\nReadNextMsg%d: EXIT\n",MemID));
  InsertListTime(MemID,READEND);
  return ((DCDT_Msg *) NULL);
};

#endif




#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL
/** \brief Check and mark the element \a Elem as deletable.
 *
 *  Check if the element \a Elem is deletable and if it can be deleted we build a bridge around it so the element is logically put put into a separated queue of useless elements.
 *  Note: this method was used only by the cleaner.
 *
 * Ex:
 * Indice = "Msg 1" ( Last useful message )
 * Elem = "Msg 2"
 *
 *  |p| Msg 1 | |<----|p| Msg 2 | |<----|p| Msg 3 | |   
 *  | |       |n|---->| |       |n|---->| |       |n|
 *
 * Indice = "Msg 1"
 * Elem = "Msg 2"
 *
 *  |p| Msg 1 | |<----|p| Msg 3 | |   NULL<----|p| Msg 2 | |     |p| Msg 3 | |
 *  | |       |n|---->| |       |n|            | |       |n|---->| |       | |
 */
inline void DCDT_PostOffice:: SetListElemDeletable (DCDT_MsgListElem *Elem)
{  
  if ( CheckIfMsgRead(Elem) || CheckTimeoutElem( Elem))
    {
      InsertListTime(0,BRIDGELOCKINIT);
      
      // Lock all the pointers that we have to change
      Indice->mtxNextPointer.lock();
      Elem->mtxNextPointer.lock();
      Elem->next->mtxPrevPointer.lock();
      
      InsertListTime(0,BRIDGELOCKEND);
      
      // Build a bridge around the element
      Indice->next=Elem->next;
      Elem->next->prev=Indice;
      Elem->prev=NULL;
      
      // Mark the elemet as deletable
      Elem->ToBeDeleted = true;
      
      // Unlock all the pointers
      Indice->mtxNextPointer.unlock();
      Elem->mtxNextPointer.unlock();
      Elem->next->mtxPrevPointer.unlock();
    }
  else
    // If the element can't be deleted this is the last useful element
    Indice=Elem;                         
}
#endif


#ifdef TOTAL_LOCK

/*! \brief Check if a message was read by all the member, if it is mark the message as deletable.
 */
inline void DCDT_PostOffice::SetActMsgDeletable (int Member_ID)
{
  int offset, pos;
  unsigned int mask;
  
  DBG_PO_PRINTF( ("\nSetActMsgDeletable\n"));

  // Lock the message queue
  insert_mtx.lock();

  offset = (Member_ID - 1) / SIZEOF_INTEGER;
  pos = Member_ID - (offset * SIZEOF_INTEGER);
  mask = ~(1 << (pos - 1));
  
  MemberTable[Member_ID].ActReadMsgElem->readers[offset] &= mask;
  
  
  // TODO: If we have changed the readers mask change the pointer to the
  //       read message, so we can guarantee the consistency
  
  if (CheckIfMsgRead (MemberTable[Member_ID].ActReadMsgElem))
    {
      DBG_PRINTF(("Deleting an element\n"));    
      MemberTable[Member_ID].ActReadMsgElem->ToBeDeleted = true;
    }

  // Unlock the message queue
  insert_mtx.unlock();

};

#endif



#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL

/*! \brief Delete the message \a Elem if it was previously mareked as deletable.
 *
 *  This method is called during the second step of the cleaner when a message has the flag ToBeDeleted = TRUE.
 *  The field NextToBeDeleted points to the next element marked for deletion, so we have a logically separated list of deletable elements.
 *  \return Pointer to the next element
 */
inline DCDT_MsgListElem*  DCDT_PostOffice::RemoveListElemDeletable (DCDT_MsgListElem * Elem)
{
  DCDT_MsgListElem* ElemDeleted;
  
  ElemDeleted=Elem;
  
  // Connect the previous element to the next element
  // Note: next and previous are referred to the direction of the cleaner.
  //       The cleaner goes from head to tail 
  prevDelete->nextToBeDeleted=Elem->nextToBeDeleted;
  
  // Go to the next element
  Elem=Elem->nextToBeDeleted;
  
  // 1 - Cut the element from the queue
  ElemDeleted->prev=NULL;
  ElemDeleted->next=NULL;
  ElemDeleted->nextToBeDeleted=NULL;
  
  // 2 - delete the element
  delete ElemDeleted;
  
  //NumOfMsgPresent--;

  // 3 - return the next message
  return Elem;
}
#endif

#ifdef SINGLE_LOCK_WITH_BUFFER

/*| \brief Delete the head element from \aMemId's local queue 
 *
 *  NB The message is deleted when needed with the destructor of ElemDeletable
 */
inline void DCDT_PostOffice::RemoveBufferElement(int MemID)
     
{
  DCDT_MsgListElem * ElemDeletable;
 
  ElemDeletable = MemberTable[MemID].HeadBuffer;
  if (MemberTable[MemID].HeadBuffer->next)
    {
      
      MemberTable[MemID].HeadBuffer = MemberTable[MemID].HeadBuffer->next;
      // Decrements MsgPtr->Readers_Counter and if == 0 delete the message
      delete ElemDeletable;
    }
  else
    {
      
      MemberTable[MemID].HeadBuffer = NULL;
      MemberTable[MemID].TailBuffer = NULL;
      // Decrements MsgPtr->Readers_Counter and if == 0 delete the message
      delete ElemDeletable;
      
    }
  NumOfMsgPresent--;
}

#endif

#ifdef TOTAL_LOCK

/*! \brief Delete the element \a Elem locking the global message queue.
 */
inline void DCDT_PostOffice::RemoveListElement (DCDT_MsgListElem * Elem)
{
  
  int PointingMemID;
  int counter=0;
  DBG_PO_PRINTF( ("\nRemoveListElem\n"));

  if ( Elem ) {
    if (Elem->next) {
      if (Elem->prev) {
	DBG_PRINTF( ("The element is in the middle. Deleting it.\n"));
	
	Elem->prev->next = Elem->next;
	Elem->next->prev = Elem->prev;
	
	delete Elem;
      }
      else {
	DBG_PRINTF( ("Deleting the first element\n"));
	Head->next = Elem->next;
	Elem->next->prev = NULL;
	
	delete Elem;
      }
    }
    else {  
      DBG_PRINTF(("Deleting the last element\n"));
      Tail = Elem->prev;
      Elem->prev->next = NULL;
      
      delete Elem;
    }
    
    NumOfMsgPresent--;
  }
  
  DBG_PO_PRINTF(("RemoveListElement: EXIT\n"));
};

#endif



#ifdef TOTAL_LOCK_WITH_READER_CLEANER

/*! \brief Delete the element \a Elem locking the global message queue.
 */
inline void DCDT_PostOffice::RemoveListElement (DCDT_MsgListElem * Elem)
{
  int PointingMemID;
  int counter=0;
  DBG_PO_PRINTF( ("\nRemoveListElem\n"));
  
  if ( Elem ) {
    if (Elem->next) {
      if (Elem->prev) {
	DBG_PRINTF( ("Deleting an element in the middle\n") );
	
	Elem->prev->next = Elem->next;
	Elem->next->prev = Elem->prev;
	delete Elem;
      }
      else {  
	  DBG_PRINTF( ("Deleting the first element\n"));
	  Head->next = Elem->next;
	  Elem->next->prev = NULL;
	  delete Elem;
      }
    }
    else {  
      DBG_PRINTF(("Deleting the last element\n"));
      Tail = Elem->prev;
      Elem->prev->next = NULL;
      
      delete Elem; 
	  
    }
    
    NumOfMsgPresent--;
  }
  
  DBG_PO_PRINTF(("RemoveListElement: EXIT\n"));
};
#endif


#ifdef SINGLE_LOCK_WITH_DISPATCHER

/*| \brief Delete the head element from \aMemId's local queue 
 *
 *  NB The message is deleted when needed with the destructor of ElemDeletable
 */
inline void DCDT_PostOffice::RemoveBufferElement(int MemID)
     
{
  DCDT_MsgListElem * ElemDeletable;
  
  ElemDeletable = MemberTable[MemID].HeadBuffer;

  if (MemberTable[MemID].HeadBuffer->next) {
    MemberTable[MemID].HeadBuffer = MemberTable[MemID].HeadBuffer->next;
    delete ElemDeletable;
  }
  else {
    MemberTable[MemID].HeadBuffer = NULL;
    MemberTable[MemID].TailBuffer = NULL;  
    delete ElemDeletable;
  }
}

#endif


#ifdef SINGLE_LOCK_WITH_DISPATCHER_WITH_CONDITION_VARIABLES

/*| \brief Delete the head element from \aMemId's local queue 
 *
 *  NB The message is deleted when needed with the destructor of ElemDeletable
 */
inline void DCDT_PostOffice::RemoveBufferElement(int MemID)    
{
  DCDT_MsgListElem * ElemDeletable;
  
  ElemDeletable = MemberTable[MemID].HeadBuffer;
  if (MemberTable[MemID].HeadBuffer->next) {
    MemberTable[MemID].HeadBuffer = MemberTable[MemID].HeadBuffer->next;
    delete ElemDeletable;
  }
  else {
    MemberTable[MemID].HeadBuffer = NULL;
    MemberTable[MemID].TailBuffer = NULL;  
    delete ElemDeletable;
  }
}

#endif

/*! \brief Check if the message pointed by  \a elem was expired
 */ 
inline bool DCDT_PostOffice::CheckTimeoutElem( DCDT_MsgListElem * elem)
{
  struct timeval now; 
  DCDT_MsgListElem * elemtmp;

#ifdef SINGLE_LOCK_WITH_BUFFER
  elemtmp = NULL;  
#else

#ifdef SINGLE_LOCK_WITH_DISPATCHER
  elemtmp = NULL;
#else
#ifdef SINGLE_LOCK_WITH_DISPATCHER_WITH_CONDITION_VARIABLES
  elemtmp = NULL;
#else
  elemtmp = Head;
#endif
#endif
#endif

  if ( elem != elemtmp  ) 
    {
      gettimeofday(&now, NULL) ;     
      return ( DeltaSeconds( elem->insert_time, now) > TIMEOUT_MSG);
    }
  else
    return false;
};



#ifdef SINGLE_LOCK_WITH_BUFFER
inline void DCDT_PostOffice::CleanupList ()
{

}
#endif

#ifdef TOTAL_LOCK

/** \brief Cleanup the message list.
 *
 *  Lock and slide the message list from head to tail.
 *  If we find a message that was marked for deletion ( means that all subscribers have already read the message ) delete it.
 */

inline void DCDT_PostOffice::CleanupList ()
{
  DCDT_MsgListElem *actElem, *tempElem;
  InsertListTime(0,DELETELOCKINIT);
  insert_mtx.lock ();
  InsertListTime(0,DELETELOCKEND);
  DBG_PO_PRINTF( ("Cleanup Routine\n"));

  if( NumOfMsgPresent ) {
    actElem = Head->next;
    
    //  DBG_PRINTF( ("lock Insert\n"));
    
    while (actElem) {
      tempElem = actElem;	
      
      actElem = actElem->next;
      
      if ( tempElem->ToBeDeleted || CheckTimeoutElem( tempElem)) {
	
	if (tempElem->numMemberPointing == 0) {
	  DBG_PO_PRINTF(("Deleting an element (ToBeDeleted = %d, Timeout = %d)\n", 
			 tempElem->ToBeDeleted ,CheckTimeoutElem( tempElem)));    
	  InsertListTime(0,DELETEINIT);
	  RemoveListElement (tempElem);
	  InsertListTime(0,DELETEEND);
	}
      } 
    }
  }
  
  insert_mtx.unlock (); 
  DBG_PRINTF( ("Unlocked Insert\n"));  
  DBG_PO_PRINTF( ("Cleanup Routine: EXIT\n"));
};

#endif // TOTAL_LOCK



#ifdef TOTAL_LOCK_WITH_READER_CLEANER

/*! \brief Cleanup the message list.
 *
 *  Lock and slide the message list, check each message timeout, if it's 
 *  expired the message is deleted.
 *  NB: We don't have to delete other messages because each member deletes
 *  a message when he's the last reader.
 */
inline void DCDT_PostOffice::CleanupList ()
{
  DCDT_MsgListElem *actElem, *tempElem;
  InsertListTime(0,DELETELOCKINIT);
  insert_mtx.lock ();
  InsertListTime(0,DELETELOCKEND);
  DBG_PO_PRINTF( ("Cleanup Routine\n"));
  
  if( NumOfMsgPresent ) 
    {
      actElem = Head->next;
      
      while (actElem)
	{
	  
	  tempElem = actElem;	
	  actElem = actElem->next;
	  
	  // TODO: Add a check to see if the message was in queue for too time
	  if ( CheckTimeoutElem( tempElem) ) {
	    if (tempElem->numMemberPointing == 0) {
	      DBG_PO_PRINTF(("Deleting an element (ToBeDeleted = %d, Timeout = %d)\n", 
			     tempElem->ToBeDeleted ,CheckTimeoutElem( tempElem)));    
	      InsertListTime(0,DELETEINIT);
	      RemoveListElement (tempElem);
	      InsertListTime(0,DELETEEND);
	    }
	  }  
	}
    }
  
  insert_mtx.unlock (); 
  DBG_PRINTF( ("Unlocked Insert\n"));  
  DBG_PO_PRINTF( ("Cleanup Routine: EXIT\n"));
  
};

#endif



#ifdef SINGLE_LOCK_WITH_DISPATCHER

/*! \brief Slides the dispatcher's queue and manages the messages.
 *
 *  Slides the dispatcher's queue from the head and deliver the messages to the member according to the subscription informations.
 *  Then clean the dispatcher's queue; if there is only a single message to be processed nothing is done.
 */
inline void DCDT_PostOffice::CleanupList ()
{
  DCDT_MsgListElem *actElem;
  DCDT_MsgListElem *tmpLastInsert;
  DCDT_MsgListElem *tmpElem;
  int MemberID = 0;
  int msg_type;
  int temp,templ,tempr;
  
  actElem=Head->next;

  // We use a new temporary variable to prevent the deletion of a message inserted but not processed.
  // The new message will be processed in the next dispatcher cycle if the message isn't the last in the queue  
  tmpLastInsert = LastInsert;
  
  
  
  DBG_PO_PRINTF( ("Cleanup Routine\n"));
  int z = 0;
  if( NumOfMsgPresent>0 ) 
    {
      
      while (actElem!=tmpLastInsert)
	{
	  InsertListTime(0,WRITEINIT);
	  msg_type = actElem->MsgPtr->ReadType ();
	  MemberID = 0;
	  for (int i=0; i<NUMINT_MEMBERS;i++)
	    {
	      
	      for (int t=1;t<=SIZEOF_INTEGER;t++)
		{ 
		  
		  MemberID++;
		  templ = SubscribedTable[msg_type].SubscribersIfLocal[i];
		  templ = templ << (SIZEOF_INTEGER - t);
		  templ = templ >> (SIZEOF_INTEGER-1);
		  
		  tempr = SubscribedTable[msg_type].SubscribersIfRemote[i];
		  tempr = tempr << (SIZEOF_INTEGER - t);
		  tempr = tempr >> (SIZEOF_INTEGER-1); 
		  
		  temp = (templ | tempr);
		  
		  if (temp && (MemberID != actElem->LocalPublisherID))
		    {
		      // We have found a member subscribed to this message so
                      // we have to insert the message into the member's queue

		      DCDT_MsgListElem * newMsgListElem;
		      newMsgListElem = new (DCDT_MsgListElem);
		      
		      actElem->MsgPtr->Increment_Readers_Counter();
		      newMsgListElem = new (DCDT_MsgListElem);
		      newMsgListElem->MsgPtr = actElem->MsgPtr; 
		      newMsgListElem->next = NULL; 
		      
		      gettimeofday(&(newMsgListElem->insert_time), NULL);
		      SetType (newMsgListElem, msg_type);
		      InsertListTime(0,WRITELOCKINIT);
		      MemberTable[MemberID].mtx.lock();
		      InsertListTime(0,WRITELOCKEND);
		     
		      if (MemberTable[MemberID].HeadBuffer)
			MemberTable[MemberID].TailBuffer->next = newMsgListElem;
		      else
			MemberTable[MemberID].HeadBuffer = newMsgListElem;
		      MemberTable[MemberID].TailBuffer = newMsgListElem;
		      NumOfMsgPresent++;
		      MemberTable[MemberID].mtx.unlock();
		    }
		}
	    }
	  InsertListTime(0,DELETEINIT);
	  
	  // Cut the actual element from the queue and delete it.
	  // Then we can go to the next message
	  actElem->prev->next = actElem->next;
	  actElem->next->prev = actElem->prev;
	  
	  tmpElem = actElem;
	  
	  actElem = actElem->next;
	  delete tmpElem;
	  InsertListTime(0,DELETEEND);
	  InsertListTime(0,WRITEEND); 
	}
    }
    
  DBG_PO_PRINTF( ("Cleanup Routine: EXIT\n"));
};
#endif

#ifdef SINGLE_LOCK_WITH_DISPATCHER_WITH_CONDITION_VARIABLES

/*! \brief Slides the dispatcher's queue and manages the messages.
 *
 *  Slides the dispatcher's queue from the head and deliver the messages to the member according to the subscription informations.
 *  Then clean the dispatcher's queue; if there is only a single message to be processed nothing is done.
 */
inline void DCDT_PostOffice::CleanupList ()
{
  DCDT_MsgListElem *actElem;
  DCDT_MsgListElem *tmpLastInsert;
  DCDT_MsgListElem *tmpElem;
  int MemberID = 0;
  int msg_type;
  int temp,templ,tempr;

  actElem=Head->next;

  // We use a new temporary variable to prevent the deletion of a message inserted but not processed.
  // The new message will be processed in the next dispatcher cycle if the message isn't the last in the queue
  tmpLastInsert = LastInsert;
  
  DBG_PO_PRINTF( ("Cleanup Routine\n"));

  if( NumOfMsgPresent>0 ) {
    DBG_PO_PRINTF( ("NumOfMsgPresent %d\n", NumOfMsgPresent ));

    //while (actElem!=tmpLastInsert) {
    while ( true ) {
      DBG_PO_PRINTF (( "actElem %p ( MsgPtr %p)\n", actElem, actElem->MsgPtr));
      if ( actElem->MsgPtr == NULL ) {
	DBG_PO_PRINTF( ( "ERRORE sono alla fine della lista!!!!\n" ));
	break;
      }
      InsertListTime(0,WRITEINIT);
      msg_type = actElem->MsgPtr->ReadType ();
      MemberID = 0;
      for (int i=0; i<NUMINT_MEMBERS;i++) {
	for (int t=1;t<=SIZEOF_INTEGER;t++) { 
		
	  MemberID++;
	  templ = SubscribedTable[msg_type].SubscribersIfLocal[i];
	  templ = templ << (SIZEOF_INTEGER - t);
	  templ = templ >> (SIZEOF_INTEGER-1);
	  
	  tempr = SubscribedTable[msg_type].SubscribersIfRemote[i];
	  tempr = tempr << (SIZEOF_INTEGER - t);
	  tempr = tempr >> (SIZEOF_INTEGER-1); 
	  
	  temp = (templ | tempr);
	  
	  if (temp && (MemberID != actElem->LocalPublisherID)) {

	    // Copy the message into the member's queue
	    DCDT_MsgListElem * newMsgListElem;
	    newMsgListElem = new (DCDT_MsgListElem);
	    
	    actElem->MsgPtr->Increment_Readers_Counter();

	    newMsgListElem = new (DCDT_MsgListElem);

	    newMsgListElem->MsgPtr = actElem->MsgPtr; 
	    
	    newMsgListElem->next = NULL; 
	         
	    gettimeofday(&(newMsgListElem->insert_time), NULL);
	    SetType (newMsgListElem, msg_type);

	    InsertListTime(0,WRITELOCKINIT);
	    MemberTable[MemberID].mtx.lock();
	    MemberTable[MemberID].condition_var.lock();
	    InsertListTime(0,WRITELOCKEND);
	   
	    // If the member's queue is empty we have to send a signal
	    /*if ( (MemberTable[MemberID].HeadBuffer) &&
                 !(MemberTable[MemberID].HeadBuffer->ToBeDeleted) ) {*/
            if ( (MemberTable[MemberID].HeadBuffer) ) {
	      MemberTable[MemberID].TailBuffer->next = newMsgListElem;
            }
	    else {
	      MemberTable[MemberID].HeadBuffer = newMsgListElem;
	      MemberTable[MemberID].condition_var.signal(); 
	    }

	    MemberTable[MemberID].TailBuffer = newMsgListElem;
	    // NumOfMsgPresent++;
	    MemberTable[MemberID].condition_var.unlock();
	    MemberTable[MemberID].mtx.unlock();
	  }
	}
      }

      InsertListTime(0,DELETEINIT);

      // Cut the actual element from the queue and delete it
      // Then we can go to the next message
      actElem->prev->next = actElem->next;
      actElem->next->prev = actElem->prev;
      
      tmpElem = actElem;
      
      actElem = actElem->next;
      delete tmpElem;
      NumOfMsgPresent--;

      if ( tmpLastInsert == tmpElem )
	break;

      InsertListTime(0,DELETEEND);
      InsertListTime(0,WRITEEND); 
    }
  }
  
  DBG_PO_PRINTF( ("Cleanup Routine: EXIT\n"));
};
#endif


#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL

/*! \brief Slides the dispatcher's queue and manages the messages.
 *
 *  Slides the dispatcher's queue from the head to the tail 2 times.
 *  During the first cycle we build a new "path" that contains only messages
 *  that has to be read; we use the already readmessages to build another 
 *  queue.
 *  In the second run we slide the useless queue and we delete each element.
 *
 *  SetListElemDeletable - Build a bridge around an element and put the
                           element into the useless queue.
 *  RemoveListElemDeletable - Delete and element.
 *
 *  ( ??? ) Verificare, se c'e' un membro che punta all'elemento da cancellare
 *          l'elemento non viene cancellato e non viene rimesso in coda quindi
 *          si perdono tutti i riferimenti ma non lo si cancella.
 */
inline void DCDT_PostOffice::CleanupList ()
{
  DCDT_MsgListElem *actElem;
  DCDT_MsgListElem *tmpLastInsert;
  
  
  Indice=Head;
   
  actElem=Head->next;
  
  // We use a new temporary variable to prevent the deletion of a message inserted but not processed.
  // The new message will be processed in the next dispatcher cycle if the message isn't the last in the queue
  
  tmpLastInsert = LastInsert;
  prevDelete = Head;
  
  DBG_PO_PRINTF( ("Cleanup Routine\n"));
  Indice=Head;
  
  if( NumOfMsgPresent>0 ) 
    {
      // First run: Build bridges around useless elements and put them into 
      //            the useless queue
      while (actElem!=tmpLastInsert)
	{
	  InsertListTime(0,BRIDGEINIT);
	  
	  SetListElemDeletable(actElem); 
	  InsertListTime(0,BRIDGEEND);
	  actElem = actElem->next;  
	}
      
      // Second run: delete the useless queue; a useless element has to be:
      // 1 - Marked for deletion
      // 2 - Not pointed by any member
      
      Indice=Head;
      actElem=Indice->nextToBeDeleted;

      while (actElem!=tmpLastInsert)
	{
	  // Is the element deletable ?
	  if ((actElem->ToBeDeleted))
	    {
	      InsertListTime(0,DELETELOCKINIT);
	      actElem->mtxNextPointer.lock();
	      InsertListTime(0,DELETELOCKEND);
	      
	      actElem->next=Indice->next;
	      actElem->mtxNextPointer.unlock();
	      
	      // There isn't a member pointing to it ?
	      if ( actElem->numMemberPointing == 0 )
		{  
		  // Delete the current element
		  InsertListTime(0,DELETEINIT);
		  actElem = RemoveListElemDeletable (actElem); 
		  InsertListTime(0,DELETEEND);
		  NumOfMsgPresent--;
		}
	      else
		{
		  prevDelete = actElem;
		  actElem = actElem->nextToBeDeleted;  
		}
	    }
	  else
	    {
	      Indice=actElem;
	      prevDelete = actElem;
	      actElem = actElem->nextToBeDeleted; 
	    }
	}
    };

  actElem = NULL;
  Indice = NULL;
  prevDelete = NULL;
  tmpLastInsert = NULL;

  DBG_PO_PRINTF( ("Cleanup Routine: EXIT\n"));
};




#endif



#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL

/*! \brief Mark the message pointed by \a Elem as read from member \a Mem_ID .
 *
 *  Update the readers mask of the message \a *Elem.
 */
inline void DCDT_PostOffice::SetMsgRead (DCDT_MsgListElem * Elem, int Mem_ID)
{
  int i;
  int offset, pos;
  unsigned int mask;

  DBG_PRINTF(("\nSetReaders\n"));
  
  offset = (Mem_ID - 1) / SIZEOF_INTEGER;
  pos = Mem_ID - offset * SIZEOF_INTEGER;
  mask = ~(1 << (pos - 1));
  
  // Use a lock to avoid concurrent modifications
  Elem->mtxmask.lock ();
  Elem->readers[offset] &= mask;
  Elem->mtxmask.unlock ();
  
};

#endif // SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL

/*! \brief Update the readers mask for the element \a Elem and message type \a num_type.
 */
inline void DCDT_PostOffice::SetReaders (DCDT_MsgListElem * Elem, int num_type)
{  
  int i;
  int PublisherID, offset, pos;
  unsigned int mask;

#ifdef TOTAL_LOCK  
  unsigned int control_value=0; // needed to control if there are readers
#endif

  DBG_PRINTF(("\nSetReaders\n"));
  
  if (Elem && ((num_type > 0) && (num_type <= MAX_MSG_TYPE))) {
    PublisherID = Elem->LocalPublisherID;
    DBG_PRINTF(("PublisherID=%d\n", PublisherID));
    offset = (PublisherID - 1) / SIZEOF_INTEGER;
    pos = PublisherID - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));

    // Fill the readers mask 
    for (i = 0; i < NUMINT_MEMBERS; i++) {
      if (Elem->local || myAgora->CheckMemberProfile(PublisherID, MPM_BRIDGE))
	Elem->readers[i] = SubscribedTable[num_type].SubscribersIfLocal[i];
      else
	Elem->readers[i] = SubscribedTable[num_type].SubscribersIfRemote[i];
      
#ifdef TOTAL_LOCK 
      control_value |= !Elem->readers[i];
#endif
      
    };
#ifdef TOTAL_LOCK 
    if (control_value)
      Elem->ToBeDeleted = true;
#endif
    
    // Avoid self-reading
    Elem->readers[offset] &= mask;
  }
};

/* !\brief Subscribe member \a Member_ID to the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice::SubscribeMsgTypeID (int Member_ID, int MsgType)
{
  int offset, pos;

  // Check if the message type is correct
  if ((MsgType > 0) && (MsgType <= MAX_MSG_TYPE)) {
    
    DBG_PRINTF(("Subscription request from Member %d to Msg Type %d\n", Member_ID, MsgType));
    
    // Update the member subscription table
    // Don't need a lock because only the same member can update his subscription mask
    offset = (MsgType - 1) / SIZEOF_INTEGER;
    pos = MsgType - (offset * SIZEOF_INTEGER);
    MemberTable[Member_ID].SubscribedMsg[offset] |= (1 << (pos - 1));
    
#ifdef _DEBUG_DCDT
    printf ("MemberTable=");
    
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].SubscribedMsg[i]);
    printf ("\n");
#endif
    
    // Update the message subscription table
    offset = (Member_ID - 1) / SIZEOF_INTEGER;
    pos = Member_ID - (offset * SIZEOF_INTEGER);
    
    SubscribedTable[MsgType].mtx.lock ();
    
    // Reading politics: local members and bridges read all msgs;
    // remote members read local msgs
    
    SubscribedTable[MsgType].SubscribersIfLocal[offset] |= (1 << (pos - 1));
    if (myAgora->CheckMemberProfile(Member_ID, MPM_USER) || 
	myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE))
      SubscribedTable[MsgType].SubscribersIfRemote[offset] |= (1 << (pos - 1));
    
    SubscribedTable[MsgType].mtx.unlock ();
    
    
#ifdef _DEBUG_DCDT
    printf ("SubscribedTable: Local=");
    for (int i = (NUMINT_MEMBERS - 1); i >= 0; i--)
      printf ("[%x]", SubscribedTable[MsgType].SubscribersIfLocal[i]);
    printf(" Remote=");
    for (int i = (NUMINT_MEMBERS - 1); i>= 0; i--)
      printf ("[%x]", SubscribedTable[MsgType].SubscribersIfRemote[i]);
    printf ("\n\n");
#endif
    
    // Update the update flags
    UpdateMask_mtx.lock();
    for (int i=0; i<Member_ID; i++)
      MemberTable[i].update = true;
    for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
      MemberTable[i].update = true;
    
    UpdateMask_flag = true;
    UpdateMask_mtx.unlock();
    
  } //end if valid msg type
};

/*! \brief Unsubscribe member \a Member_ID from the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice::UnSubscribeMsgTypeID (int Member_ID, int MsgType)
{
  int  offset, pos;
  unsigned int mask;
  
  // Check if the message type is correct
  if ((MsgType > 0) && (MsgType <= MAX_MSG_TYPE)) {
    
    // Update member subscription table
    offset = (MsgType - 1) / SIZEOF_INTEGER;
    pos = MsgType - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));

    MemberTable[Member_ID].SubscribedMsg[offset] &= mask;
    
    // DEBUG
#ifdef _DEBUG_DCDT
    printf ("MemberTable=");
    
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].SubscribedMsg[i]);
    printf ("\n");
#endif
    
    // Update message subscription table
    offset = (Member_ID - 1) / SIZEOF_INTEGER;
    pos = Member_ID - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));
    
    SubscribedTable[MsgType].mtx.lock ();
    
    SubscribedTable[MsgType].SubscribersIfLocal[offset] &= mask;
    if (myAgora->CheckMemberProfile(Member_ID, MPM_USER) || 
	myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE))
      SubscribedTable[MsgType].SubscribersIfRemote[offset] &= mask;
    
    SubscribedTable[MsgType].mtx.unlock ();

    // Update the update flags    
    UpdateMask_mtx.lock();
    for (int i=0; i<Member_ID; i++)
      MemberTable[i].update = true;
    for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
      MemberTable[i].update = true;
    UpdateMask_flag = true;
    UpdateMask_mtx.unlock();
    
  }

  // TODO Check if there isn't messages with the same Member_ID to be deleted
};

/*! \brief Unsubscribe the member \a Member_ID from all the messages types
 *
 */
inline void DCDT_PostOffice::UnSubscribeAll ( int Member_ID )
{
  int i, j, offset, pos, MsgType;
  unsigned int msg_mask, mask;
  
  offset = (Member_ID - 1) / SIZEOF_INTEGER;
  pos = Member_ID - (offset * SIZEOF_INTEGER);
  mask = ~(1 << (pos - 1));
  
  // For all the messages:
  // - If we are susbscribed unsubscribe and update the message subscription mask
  for (i = 0; i < NUMINT_MSGTYPE; i++) {
    for (j = 0; j < SIZEOF_INTEGER; j++) {
      msg_mask = 1;
      msg_mask <<= j;
      if (msg_mask & MemberTable[Member_ID].SubscribedMsg[i]) {
	MsgType = SIZEOF_INTEGER * i + j + 1;
	SubscribedTable[MsgType].mtx.lock ();
	
	SubscribedTable[MsgType].SubscribersIfLocal[offset] &= mask;
	if ( myAgora->CheckMemberProfile(Member_ID, MPM_USER) || 
	     myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ))
	  SubscribedTable[MsgType].SubscribersIfRemote[offset] &= mask;
	
	SubscribedTable[MsgType].mtx.unlock ();
      }
    }
    
    MemberTable[Member_ID].SubscribedMsg[i] = 0;
  }
  
  UpdateMask_mtx.lock();
  for (int i=0; i<Member_ID; i++)
    MemberTable[i].update = true;
  for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
    MemberTable[i].update = true;
  UpdateMask_flag = true;
  UpdateMask_mtx.unlock();
  
}

/*! \brief Check if member \a Member_ID is interested in the message pointed by \a Elem.
 *
 *  We don't change the values so no lock is needed.
 *  \return TRUE - The member is subscribed to this message.
 *  \return FALSE - The meber isn't subscribed to this message. 
 */
inline int DCDT_PostOffice::MatchMsgTypeMask (int Member_ID, DCDT_MsgListElem * Elem)
{
  int i;
  unsigned int match = 0;
  DBG_PRINTF( ("\nMatchMask\n"));
  for (i = (NUMINT_MSGTYPE - 1); i >= 0; i--) {
    match |= (Elem->type_mask[i] & MemberTable[Member_ID].SubscribedMsg[i]);
  }
  
  return match;
};

/*! \brief Check if \a Member_ID is interested in the message pointed by \a Elem.
 *
 *  We don't change so no lock is needed.
 *  \return TRUE - The member is subscribed to this message.
 *  \return FALSE - The member isn't subscribed to this message.
 */
inline int DCDT_PostOffice::MatchReadersMask (int Member_ID, DCDT_MsgListElem * Elem)
{
  int offset, pos;
  unsigned int mask;
  int match = 0;
  
  DBG_PRINTF( ("\nMatchReadersMask\n"));
  offset = (Member_ID - 1) / SIZEOF_INTEGER;
  pos = Member_ID - (offset * SIZEOF_INTEGER);
  mask = 1 << (pos - 1);
  match = (Elem->readers[offset] & mask);
  DBG_PRINTF(("readers[%d]=%d, mask = %d, match=%d\n", 
	      offset, Elem->readers[offset], mask, match));
  
  return match;
};

/*! \brief Set the type of the message point by \a Elem to \a num_type.
 *
 *  This method is used only during the insert phase so no lock is required because there aren't concurrency problems.
 */
inline void DCDT_PostOffice::SetType (DCDT_MsgListElem * Elem, int num_type)
{
  int offset, pos;
  DBG_PRINTF( ("\nSetType\n"));
  
  if (Elem && ((num_type > 0) && (num_type <= MAX_MSG_TYPE))) {
    offset = (num_type - 1) / SIZEOF_INTEGER;
    pos = num_type - (offset * SIZEOF_INTEGER);
    
    Elem->type_mask[offset] |= 1 << (pos -1);
    
  }
  else {
    DBG_PRINTF( ("SetType ERROR\n"));
  }
}


#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL

/*! \brief Safe method to go to the next message.
 */
inline DCDT_MsgListElem * DCDT_PostOffice::GoToNextListMsg (DCDT_MsgListElem * Elem,int MemID)
{
  DCDT_MsgListElem* Elemtmp ;

  // Lock those mutex in sequence:
  // 1 - Counter of members pointing to this element
  // 2 - Pointer to the next element
  // 3 - Pointer to the previous element of the next element
  // 4 - Counter of members pointing to the next element
  InsertListTime(MemID,READLOCKINIT);

  Elem->mtxnumMemberPointing.lock(); 
  Elem->mtxNextPointer.lock();
  Elem->next->mtxPrevPointer.lock();
  Elem->next->mtxnumMemberPointing.lock();
  
  InsertListTime(MemID,READLOCKEND);

  // Go to the next element
  Elemtmp=Elem;
  Elem=Elem->next;

  Elem->numMemberPointing++;
  Elemtmp->numMemberPointing--;
     
  MemberTable[MemID].LastVisitedMsgElem=Elem;

  // Unlock all the mutex
  Elemtmp->mtxnumMemberPointing.unlock();
  Elemtmp->mtxNextPointer.unlock();
  Elem->mtxPrevPointer.unlock();
  Elem->mtxnumMemberPointing.unlock();
  Elemtmp=NULL;
  
  // We are on the next element
  return (Elem);
  
  
};

#endif

/*! \brief Update the SubscribedMsgMask used by LinkTx when talking with another Agora.
 *
 *  Rebuild the global subscribers mask; the global mask is build upon the single message mask of each active member, don't include the mask of the member \a MemID; if \a MemID is a bridge all the members masks are used otherwise only the masks of MPM_USER and MPM_BRIDGE type members are used; so in a LinkTx member we can build the subscriber mask that needs to be passed to the other side of the link.
 */
inline void DCDT_PostOffice::UpdateSubscribersMask (int MemID)
{
  int i, j;
  unsigned int temp[NUMINT_MSGTYPE];
  
  for (j = 0; j < NUMINT_MSGTYPE; j++)
    temp[j] = 0;
  
  // Build a new temporary mask
  for (i = 0; i < MAX_MEMBERS; i++)
    {
      if (MemberTable[i].alive && (i != MemID)) {
	// MemID == MPM_BRIDGE -> request any message needed by local members
        // MemID != MPM_BRIDGE -> request messages only for MPM_USER and MPM_BRIDGE members ( not dynamic links, MPM_LINK )
	if(myAgora->CheckMemberProfile(i, MPM_USER) ||
	   myAgora->CheckMemberProfile(i, MPM_BRIDGE) || 
	   myAgora->CheckMemberProfile(MemID, MPM_BRIDGE)) {
	  for (j = 0; j < NUMINT_MSGTYPE; j++)
	    temp[j] |= MemberTable[i].SubscribedMsg[j];
	}
      }
    }
  
  subscribed_mtx.lock ();
  
  // Now update the mask
  for (j = 0; j < NUMINT_MSGTYPE; j++)
    SubscribedMsgMask[j] = temp[j];
  
  subscribed_mtx.unlock ();  
};

/*! \brief Returns the decimal value corresponding to the mask signature of the element pointed by \a Elem.
 */
inline int DCDT_PostOffice::GetTypeFromMask (DCDT_MsgListElem * Elem)
{
  int i, j, temp;
  
  for (i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
    {
      if (Elem->type_mask[i])
    	{
  	  temp = Elem->type_mask[i];
  	  j = 0;
  	  while (temp)
	    {
	      temp = temp >> 1;
	      j++;
	    };
	  return (i * SIZEOF_INTEGER + j);
	};
    };
};


/*! \brief Check if SubscribedMsgMask is changed.
 *
 *  \return TRUE - SubscribedMsgMask was changed
 *  \return FALSE - SubscribedMsgMask wasn't changed  
 */
inline bool DCDT_PostOffice::ReadSubscribedMsgMask(unsigned int *mask, int MemID) 
{
  bool match=true;
  
  UpdateMask_mtx.lock();
  if(MemberTable[MemID].update) 
    {
      UpdateSubscribersMask(MemID);
      MemberTable[MemID].update = false;
    };
  UpdateMask_mtx.unlock();
  
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    match = match && (mask[i] == SubscribedMsgMask[i]);
  
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    mask[i] = SubscribedMsgMask[i];

  // The old mask matches ?
  if (!match)
    return true;
  else
    return false;
}

inline bool DCDT_PostOffice::UpdateSubs(int MemID) 
{
  return MemberTable[MemID].update;
}

#ifdef TEST_PART

inline void DCDT_PostOffice::InsertListTime (int MemID, int Type_Op)
{
  DCDT_TIME x = GetTime();
  TimeTesterElem *newElem;
  newElem = new TimeTesterElem;
  newElem->Data = x;
  newElem->Type_Op = Type_Op;
  newElem->prev = TestElem[MemID].prev;
  TestElem[MemID].prev = newElem; 
} 


inline void DCDT_PostOffice::VisualList()
{
  TimeTesterElem *newElem;
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    newElem = TestElem[i].prev;
    while (newElem) {
      cout<<"["<<i<<","<<newElem->Type_Op<<","<<newElem->Data-TestElem[i].Data<<"]";
      newElem=newElem->prev;
    }
  } 
}

inline void DCDT_PostOffice::SaveList(char* nomefile)  
{
  TimeTesterElem *newElem; 
  FILE *fout;
  time_t longtime;
  struct tm* time_experiment;
  
  char string[100] = "DATI_";

  strcat(string,nomefile); 
  
  fout = fopen (string,"wb");

  // Insert time and date of the operation
  time(&longtime);
  time_experiment = localtime (&longtime);

  putw(time_experiment->tm_hour,fout);
  putw(time_experiment->tm_min,fout); 
  putw(time_experiment->tm_mday,fout); 
  putw(time_experiment->tm_mon,fout);
  putw(1900 + time_experiment->tm_year,fout);

  // Software version and system parameters
  putw(VERSION,fout);
  putw(NUMMEMBERFORMAIN,fout);

  // Note that  NUMMEMBERFORMAIN = NUMREADER + NUMWRITER + NUMREADERWRITER 
  putw(NUMREADER,fout);      
  putw(NUMWRITER,fout);      
  putw(NUMREADERWRITER,fout); 
  
  // Number of cycles  
  putw(READCYCLE,fout); 
  putw(WRITECYCLE,fout);        
  putw(READWRITECYCLE,fout);   
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    
    newElem = TestElem[i].prev;
    
    while (newElem) {
      putw(newElem->Type_Op,fout);
      putw(newElem->Data-TestElem[i].Data,fout);
      
      newElem=newElem->prev;
    }
    putw(666,fout);
  }

  fclose(fout);
}

#endif

#endif
