#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <sched.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <signal.h>
#include "gd.h"
#include "gdfonts.h"

#include "latency-graph.h"

void determine_cpu_hz(void);

  double cpu_hz;

  int lat_numsamples;
  int lat_numlatencies;
  float *lat_time_arr[MAXLATENCIES];
  float *lat_latency_arr[MAXLATENCIES];
  float lat_max_timediffs[MAXLATENCIES];
  double lat_deadline[MAXLATENCIES];
  double lat_max_latency[MAXLATENCIES];
  int lat_overruns;
  int lat_actual_sample[MAXLATENCIES];
  double lat_time_start[MAXLATENCIES];

  unsigned long lat_flags[MAXLATENCIES];

  double lat_range1low[MAXLATENCIES];
  double lat_range1high[MAXLATENCIES];
  double lat_range2low[MAXLATENCIES];
  double lat_range2high[MAXLATENCIES];

  double lat_range1frac[MAXLATENCIES];
  double lat_range2frac[MAXLATENCIES];

  char lat_color_main_R[MAXLATENCIES];
  char lat_color_main_G[MAXLATENCIES];
  char lat_color_main_B[MAXLATENCIES];

  char lat_color_deadline_R[MAXLATENCIES];
  char lat_color_deadline_G[MAXLATENCIES];
  char lat_color_deadline_B[MAXLATENCIES];
 
  char lat_color_range1_R[MAXLATENCIES];
  char lat_color_range1_G[MAXLATENCIES];
  char lat_color_range1_B[MAXLATENCIES];

  char lat_color_range2_R[MAXLATENCIES];
  char lat_color_range2_G[MAXLATENCIES];
  char lat_color_range2_B[MAXLATENCIES];



// initializes the API, malloc() of arrays

int lat_init(int numlatencies,int numsamples)
{
  int i,u;
  float *ptr;

  if(numlatencies>MAXLATENCIES) return(-1);

  determine_cpu_hz();

  lat_numlatencies=numlatencies;
  lat_numsamples=numsamples;

  lat_overruns=0;

// zero all arrays , faulting the pages because we need
// "hard-realtime" behaviour when writing to the elements of the
// arrays

  for( i=0; i<numlatencies; i++ )
  {
     ptr=(float *)malloc((numsamples+1)*sizeof(float));
     for(u=0;u<numsamples;u++)
     {
       ptr[u]=0.0;
     }
     lat_time_arr[i]=ptr;
     lat_actual_sample[i]=0;
  }

  for(i=0;i<numlatencies;i++)
  {
    
    ptr=(float *)malloc((numsamples+1)*sizeof(float));
    for(u=0;u<numsamples;u++)
    {
      ptr[u]=0.0;
    }
    lat_latency_arr[i]=ptr;
    lat_range1low[i]=0;
    lat_range1high[i]=0;
    lat_range2low[i]=0;
    lat_range2high[i]=0;
    lat_range1frac[i]=0.0;
    lat_range2frac[i]=0.0;

    lat_flags[i]=0;

    lat_deadline[i]=0.0;

    lat_color_main_R[i]=255;
    lat_color_main_G[i]=255;
    lat_color_main_B[i]=255;

    lat_color_deadline_R[i]=255;
    lat_color_deadline_G[i]=0;
    lat_color_deadline_B[i]=0;

    lat_color_range1_R[i]=255;
    lat_color_range1_G[i]=255;
    lat_color_range1_B[i]=0;

    lat_color_range2_R[i]=0;
    lat_color_range2_G[i]=255;
    lat_color_range2_B[i]=0;
  }

  return(0);

}


void lat_set_flags(int numlatency,unsigned long flags)
{
  lat_flags[numlatency]=flags;
}

void lat_set_deadline(int numlatency,double deadline)
{
  lat_deadline[numlatency]=deadline;
}

void lat_set_range1(int numlatency,double low,double high)
{
  lat_range1low[numlatency]=low;
  lat_range1high[numlatency]=high;
}

void lat_set_range2(int numlatency,double low,double high)
{
  lat_range2low[numlatency]=low;
  lat_range2high[numlatency]=high;
}


void lat_set_color_main(int numlatency,char R,char G,char B)
{
  lat_color_main_R[numlatency]=R;
  lat_color_main_G[numlatency]=G;
  lat_color_main_B[numlatency]=B;
}

void lat_set_color_deadline(int numlatency,char R,char G,char B)
{
  lat_color_deadline_R[numlatency]=R;
  lat_color_deadline_G[numlatency]=G;
  lat_color_deadline_B[numlatency]=B;
}

void lat_set_color_range1(int numlatency,char R,char G,char B)
{
  lat_color_range1_R[numlatency]=R;
  lat_color_range1_G[numlatency]=G;
  lat_color_range1_B[numlatency]=B;
}

void lat_set_color_range2(int numlatency,char R,char G,char B)
{
  lat_color_range2_R[numlatency]=R;
  lat_color_range2_G[numlatency]=G;
  lat_color_range2_B[numlatency]=B;
}



/* common routines */


#ifndef USE_GENERIC_TIMER

unsigned long long int rdtsc()
                    {
                      unsigned long long int x;
                      __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));     return x;
                    }
#else

double lat_gettime(void)
{
  static struct timeval mytv;
  gettimeofday(&mytv,NULL);
  return(mytv.tv_sec+mytv.tv_usec/1000000.0);
}
#endif


// sets the process to "policy" policy,  if max=1 then set at max priority,
// else use min priority

int set_realtime_priority(int policy,int max)
{
struct sched_param schp;
	/*
	 * set the process to realtime privs
	 */
        memset(&schp, 0, sizeof(schp));
	if(max) 
        {  
          schp.sched_priority = sched_get_priority_max(policy);
        }
        else
        {  
          schp.sched_priority = sched_get_priority_min(policy);
        }
        

	if (sched_setscheduler(0, policy, &schp) != 0) {
		perror("sched_setscheduler");
		return -1;
	}

	return 0;

}




/* determines first the CPU frequency which is then stored in a global
   variable ( cpu_hz) which is needed to get accurate timing information
   through the Pentium RDTSC instruction
*/

void determine_cpu_hz(void)
{
  FILE *f;
  char *res;
  char s1[100];
  int num_cpus;

/* read the CPU frequency from /proc/cpuinfo  */

  f=fopen("/proc/cpuinfo","r");
  if(f==NULL)
  {
    perror("can't open /proc/cpuinfo, exiting. open");
    exit(1);
  }

  for(;;)
  {
    res=fgets(s1,100,f);
    if(res==NULL) break;
    if(!memcmp(s1,"cpu MHz",7))
    {
      cpu_hz=atof(&s1[10])*1000000.0;
    }
    if(!memcmp(s1,"processor",9))
    {
      //printf("processor string='%s'\n",&s1[11]);
      num_cpus=atoi(&s1[11]);
      //printf("num_cpus=%d\n",num_cpus);
    }

  }
  fclose(f);
  if(cpu_hz < 1.0)
  {
    fprintf(stderr,"can't determine CPU clock frequency, exiting.\n");
    fprintf(stderr,"maybe you are running on a 2.0.x kernel ? (missing cpu Mhz line.\n");
    fprintf(stderr,"please set \n");
    exit(0);

  }
}

/*
   determines how many loops per second the CPU is able to execute 
   returns the number of loops per sec, this is used for busywaiting
   through the mydelay() function
*/

int calibrate_loop(void)
{
  double tmp_loops_per_sec;
  double mytime1,mytime2;

  dprintf("calibrating loop ....\n");

#define CALIB_LOOPS 800000000

/* run mydelay for CALIB_LOOPS times, and measure how long this takes */

mytime1=lat_gettime();
mydelay(CALIB_LOOPS);
mytime2=lat_gettime();


  dprintf("time diff= %f \n",mytime2-mytime1);
  dprintf("loops/sec = %f\n",CALIB_LOOPS/(mytime2-mytime1));

/* calculate the "loops per second" value  */

  tmp_loops_per_sec=CALIB_LOOPS/(mytime2-mytime1);
  return(tmp_loops_per_sec);
}

/* simple busywaiting loop used to waste CPU time , useful to simulate
   heavy CPU bound computations , since some classes of realtime processes
   need a huge amount of CPU, like software synthesizers for example
*/

void mydelay(int loops)
{
  int k=0;
  int u;
  for(u=0;u<loops;u++) k+=1;
}






#define CHART_X1 50
#define CHART_X2 840

#define CHART_Y1 20
#define CHART_Y2 599

//#define PIXEL_PER_MS 90
#define GRID_MS 5

#define my_round(a) floor(a+0.5)

double PIXEL_PER_MS;

void draw_latency_graph(char *filename) 
{
  /* Declare the image */
  gdImagePtr im;
  /* Declare an output file */
  FILE *out;
  /* Declare color indexes */
  int black;
  int gray;
  int gray2;
  int yellow;
  int white;
  int white2;
  int lat_color_main[MAXLATENCIES];
  int lat_color_deadline[MAXLATENCIES];
  int lat_color_range1[MAXLATENCIES];
  int lat_color_range2[MAXLATENCIES];
  int oldx,oldy;
  int x,y;
  int index,i;
  int i_lat;
  double i2;
  double lat;
  double time_division;
  double total_time;
  int tmp_ms;
  int over_flag=0;
  int old_x1;
  char str1[80];
  
  int deadlinecounter[MAXLATENCIES];
  int range1counter[MAXLATENCIES];
  int range2counter[MAXLATENCIES];
  
  double max_deadline;
  double grid_factor;
  
  int linestyle1[4];
  
  max_deadline=0.0;
  for(i=0;i<lat_numlatencies;i++) {
    if(lat_flags[i] & LAT_SHOW_DEADLINE) if(lat_deadline[i] > max_deadline) max_deadline=lat_deadline[i];
    if(lat_flags[i] & LAT_SHOW_RANGE1) if(lat_range1high[i] > max_deadline) max_deadline=lat_range1high[i];
    if(lat_flags[i] & LAT_SHOW_RANGE2) if(lat_range2high[i] > max_deadline) max_deadline=lat_range2high[i];
  }
  
  if(max_deadline<0.0001) {
    dprintf("determining max peak because max_deadline was 0 ....\n");
    for(i_lat=0;i_lat<lat_numlatencies;i_lat++) {
      for(i=0;i<lat_numsamples;i++) {
	if(lat_latency_arr[i_lat][i] > max_deadline) max_deadline=lat_latency_arr[i_lat][i];
      }
    }
  }
  
  
  PIXEL_PER_MS=550.0/(max_deadline*1000.0);
  dprintf("PIXEL_PER_MS=%f\n",PIXEL_PER_MS);
  
  for ( index=0; index<lat_numlatencies; index++ )
    if ( lat_numsamples < lat_actual_sample[index] )
      lat_numsamples = lat_actual_sample[index];
  
  
  /* Allocate the image: 64 pixels across by 64 pixels tall */
  im = gdImageCreate(CHART_X2+20,680 );
  
  /* Allocate the color black (red, green and blue all minimum).
     Since this is the first color in a new image, it will
     be the background color. */
  black = gdImageColorAllocate(im,0 , 0, 0);
  gray = gdImageColorAllocate(im,64 , 64, 64);
  gray2 = gdImageColorAllocate(im,110 , 110, 110);
  white2 = gdImageColorAllocate(im,200 , 200, 200);
  yellow = gdImageColorAllocate(im,230 , 230, 0);
  
  for(i=0;i<lat_numlatencies;i++) {
    lat_color_main[i] = gdImageColorAllocate(im,lat_color_main_R[i] ,lat_color_main_G[i],lat_color_main_B[i]);
    lat_color_deadline[i] = gdImageColorAllocate(im,lat_color_deadline_R[i] ,lat_color_deadline_G[i],lat_color_deadline_B[i]);
    lat_color_range1[i] = gdImageColorAllocate(im,lat_color_range1_R[i] ,lat_color_range1_G[i],lat_color_range1_B[i]);
    lat_color_range2[i] = gdImageColorAllocate(im,lat_color_range2_R[i] ,lat_color_range2_G[i],lat_color_range2_B[i]);
  }
  
  /* Allocate the color white (red, green and blue all maximum). */
  white = gdImageColorAllocate(im, 255, 255, 255);
  
  
  dprintf("color allocation done\n");
  
  
  if(PIXEL_PER_MS > 2) {
    for(i2=CHART_Y2;i2>=CHART_Y1;i2-=PIXEL_PER_MS) {
      gdImageLine(im, CHART_X1,my_round(i2), CHART_X2, my_round(i2), gray);
    }
  }
  
  dprintf("grid1 done\n");
  
  tmp_ms=0;
  
  grid_factor=1.0;
  for(i=0;i<8;i++) {
    if((PIXEL_PER_MS*GRID_MS*grid_factor) < 8.0) {
      if((i % 2)==0) {
	grid_factor*=2.0;
      }
      else {
	grid_factor*=5.0;
      }
    }
  }
  
  for(i2=CHART_Y2;i2>=CHART_Y1;i2-=PIXEL_PER_MS*GRID_MS*grid_factor) {
    gdImageLine(im, CHART_X1,my_round(i2), CHART_X2, my_round(i2), gray2);
    sprintf(str1,"%3.1d ms",tmp_ms);
    tmp_ms+=GRID_MS*grid_factor;
    gdImageString(im, gdFontSmall,7, i2-8, str1, white); 
  }
  dprintf("grid2 done\n");
  
  gdImageLine(im, CHART_X1,CHART_Y2, CHART_X2, CHART_Y2, white2);
  gdImageLine(im, CHART_X1,CHART_Y2, CHART_X1, CHART_Y1, white2);
  
  
  total_time=0.0;
  for(index=0; index<lat_numlatencies; index++ ) {
    double min, max;
    min=max=lat_time_arr[index][0];
    for(i=0;i<lat_numsamples;i++)
      {
	if( lat_time_arr[index][i] < min )
          min = lat_time_arr[index][i];
	if( lat_time_arr[index][i] > max )
          max = lat_time_arr[index][i];
      }

    if ( (max-min) > total_time )
      total_time = max-min;
  }
  
  //dprintf("lat_numsamples=%d\n",lat_numsamples);
  //dprintf("lat_time_arr[%d]=%f\n",lat_numsamples-1,lat_time_arr[num][lat_numsamples-1]);
  //dprintf("lat_time_arr[0]=%f\n",lat_time_arr[num][0]);
  //dprintf("total time=%f\n",total_time);
  
  //dprintf("total time=%f\n",total_time);
  
  //dprintf("ARRAY BEGIN\n");
  //for(i=0;i<lat_numsamples;i++)
  //{
  //  dprintf("lat_time_arr[%d]=%f  lat_latency_arr[0][%d]=%f\n",i,lat_time_arr[i],i,lat_latency_arr[0][i]);
  //}
  //dprintf("ARRAY END\n");
  
  time_division=1.0;
  if(total_time>200.0) time_division=10.0;
  
  for(i2=0;i2<total_time;i2+=time_division) {
    x=CHART_X1+(CHART_X2-CHART_X1)*i2/total_time;
    gdImageLine(im, x,CHART_Y2+3, x, CHART_Y2-3, white2);
  }



  for(i2=0;i2<total_time;i2+=time_division*10.0) {
    x=CHART_X1+(CHART_X2-CHART_X1)*i2/total_time;
    gdImageLine(im, x,CHART_Y2+5, x, CHART_Y2-5, white2);
    sprintf(str1,"%.0f",i2);
    gdImageString(im, gdFontSmall,x+2,CHART_Y2+8, str1, white); 
  }
  
  gdImageString(im, gdFontSmall,CHART_X2-70,CHART_Y2+16, "time (sec)", white); 
  
  dprintf("time axis done\n");
  
  /* First draw ranges */
  for(i_lat=0;i_lat<lat_numlatencies;i_lat++) {
    if(lat_flags[i_lat] & LAT_SHOW_DEADLINE) {
      y=CHART_Y2-lat_deadline[i_lat]*1000.0*PIXEL_PER_MS;
      gdImageLine(im,CHART_X1,y,CHART_X2,y,lat_color_deadline[i_lat]);
      sprintf(str1,"%4.2f ms",(lat_deadline[i_lat]*1000.0));
      gdImageString(im, gdFontSmall,CHART_X1+2, y-12, str1, lat_color_deadline[i_lat]); 
    }
    
    if(lat_flags[i_lat] & LAT_SHOW_RANGE1) {
      linestyle1[0]=lat_color_range1[i_lat];
      linestyle1[1]=linestyle1[0];
      linestyle1[2]=linestyle1[0];
      linestyle1[3]=black;
      gdImageSetStyle(im,(int *)&linestyle1,4);
      
      y=CHART_Y2-lat_range1high[i_lat]*1000.0*PIXEL_PER_MS;
      gdImageLine(im,CHART_X1,y,CHART_X2,y,gdStyled);
      y=CHART_Y2-lat_range1low[i_lat]*1000.0*PIXEL_PER_MS;
      gdImageLine(im,CHART_X1,y,CHART_X2,y,gdStyled);
    }
    
    if(lat_flags[i_lat] & LAT_SHOW_RANGE2) {
      linestyle1[0]=lat_color_range2[i_lat];
      linestyle1[1]=linestyle1[0];
      linestyle1[2]=black;
      linestyle1[3]=black;
      gdImageSetStyle(im,(int *)&linestyle1,4);
      
      y=CHART_Y2-lat_range2high[i_lat]*1000.0*PIXEL_PER_MS;
      gdImageLine(im,CHART_X1,y,CHART_X2,y,gdStyled);
      y=CHART_Y2-lat_range2low[i_lat]*1000.0*PIXEL_PER_MS;
      gdImageLine(im,CHART_X1,y,CHART_X2,y,gdStyled);
    }
  }
  
  /* Then draw the latency */
  for(i_lat=0;i_lat<lat_numlatencies;i_lat++) {
    oldx=CHART_X1+1;
    oldy=CHART_Y2-lat_latency_arr[i_lat][0]*1000*PIXEL_PER_MS;
    old_x1=CHART_X1;
    
    deadlinecounter[i_lat]=0;
    range1counter[i_lat]=0;
    range2counter[i_lat]=0;
    lat_max_latency[i_lat]=0.0;
    
    for(index=0;index<lat_numsamples;index++) {
      lat=lat_latency_arr[i_lat][index];
      if(lat > lat_max_latency[i_lat]) lat_max_latency[i_lat]=lat;
      
      if(lat_flags[i_lat]  & LAT_SHOW_DEADLINE) {
	if(lat > lat_deadline[i_lat]) deadlinecounter[i_lat]++;
      }
      if(lat_flags[i_lat] & LAT_SHOW_RANGE1) if(lat < lat_range1low[i_lat] || lat > lat_range1high[i_lat]) range1counter[i_lat]++;
      if(lat_flags[i_lat] & LAT_SHOW_RANGE2) if(lat < lat_range2low[i_lat] || lat > lat_range2high[i_lat]) range2counter[i_lat]++;
      
      x=CHART_X1+(lat_time_arr[i_lat][index]-lat_time_arr[i_lat][0])/total_time*(CHART_X2-CHART_X1);
      
      y=CHART_Y2-lat_latency_arr[i_lat][index]*1000*PIXEL_PER_MS;
      
      //dprintf("DRAW: time=%f lat[%d]=%f x=%d  y=%d index=%d\n",lat_time_arr[num][index],i_lat,lat_latency_arr[i_lat][index],x,y,index);
      
      if(y<CHART_Y1) {
	y=CHART_Y1;
	if((x-old_x1)>30) {
	  sprintf(str1,"%.1fms",(lat_latency_arr[i_lat][index]*1000.0));
	  gdImageString(im, gdFontSmall,x+2,y-4+10*over_flag, str1, white); 
	  over_flag=1-over_flag;
	  old_x1=x;
	}
      }

      gdImageLine(im, oldx,oldy, x, y, lat_color_main[i_lat]);
      
      oldx=x;
      oldy=y;
      
    }

    sprintf(str1,"max latency:   %.1f ms",(lat_max_latency[i_lat]*1000.0));
    gdImageString(im, gdFontSmall,CHART_X1,CHART_Y2+25+40*i_lat , str1, lat_color_main[i_lat]); 
    
    if(lat_flags[i_lat] & LAT_SHOW_DEADLINE) {
      sprintf(str1,"deadline misses: %d (%.2f%%)",deadlinecounter[i_lat],100.0*(double)deadlinecounter[i_lat]/(double)lat_numsamples);
      gdImageString(im, gdFontSmall,CHART_X1+300,CHART_Y2+25+40*i_lat , str1, lat_color_deadline[i_lat]); 
    }

    if(lat_flags[i_lat] & LAT_SHOW_RANGE1) {
      sprintf(str1,"n.out of %.2f - %.2f ms range: %d (%.2f%%)",1000.0*lat_range1low[i_lat],1000.0*lat_range1high[i_lat],range1counter[i_lat],100.0*(double)range1counter[i_lat]/(double)lat_numsamples);
      gdImageString(im, gdFontSmall,CHART_X1,CHART_Y2+25+20+40*i_lat , str1, lat_color_range1[i_lat]); 
      
    }
    
    if(lat_flags[i_lat] & LAT_SHOW_RANGE2) {
      sprintf(str1,"n.out of %.2f - %.2f ms range: %d (%.2f%%)",1000.0*lat_range2low[i_lat],1000.0*lat_range2high[i_lat],range2counter[i_lat],100.0*(double)range2counter[i_lat]/(double)lat_numsamples);
      gdImageString(im, gdFontSmall,CHART_X1+300,CHART_Y2+25+20+40*i_lat , str1, lat_color_range2[i_lat]); 
    }


    sprintf(str1,"num samples: %d",lat_numsamples);
    gdImageString(im, gdFontSmall,CHART_X1+600,CHART_Y2+25 , str1, white); 
    
  }

  dprintf("drawing done\n");
  
  /* Open a file for writing. "wb" means "write binary", important
     under MSDOS, harmless under Unix. */
  out = fopen(filename, "wb");
  
  /* Output the image to the disk file. */
  gdImageGif(im, out);
  
  /* Close the file. */
  fclose(out);
  
  /* Destroy the image in memory. */
  gdImageDestroy(im);
}

void dump_samples ( char *file )
{
  FILE *stream;
  int idx,line;
  
  stream = fopen ( file, "wb" );
  
  if ( stream != NULL ) {
    fprintf ( stream, "S;" );
    for ( line=0; line < lat_numlatencies; line++ )
      fprintf ( stream, "%f;", lat_time_start[line] );
    fprintf ( stream, "\n" );
    
    for ( idx=0; idx < lat_numsamples; idx++ ) {
      fprintf ( stream, "%d;", idx );
      for ( line=0; line < lat_numlatencies; line++ ) 
	fprintf ( stream, "%f;", lat_latency_arr[line][idx] );
      fprintf ( stream, "\n" );
    }
    
    fclose ( stream );
  }
}

void dump_sample ( int line, char *file )
{
  FILE *stream;
  int idx;
  
  stream = fopen ( file, "wb" );
  
  if ( stream != NULL ) {
    fprintf ( stream, "S;%f;\n", lat_time_start[line] );
    
    for ( idx=0; idx < lat_numsamples; idx++ ) {
      fprintf ( stream, "%d;", idx );
      fprintf ( stream, "%f;", lat_latency_arr[line][idx] );
      fprintf ( stream, "\n" );
    }
    
    fclose ( stream );
  }
}

int lat_check_sampleranges ( void )
{
  int idx;
  
  for ( idx=0; idx < lat_numlatencies; idx++ ) {
    if ( !lat_check_samplerange(idx) )
      return 0;
  }
  return 1;
}
