#ifndef LATENCY_GRAPH_H
#define LATENCY_GRAPH_H

#ifdef __cplusplus
#define C_LINKAGE     extern  "C"
#else
#define C_LINKAGE
#endif /* __cplusplus */

// uncomment the #define below if you are running on a non-x86 box
// or <=486 , this will use gettimeofday() instead of the Pentium RDTSC
// instruction

#define USE_GENERIC_TIMER

// if you are running on a older  kernels ( 2.0.x ) where the
// CPU Mhz line is missing in /proc/cpuinfo, please set the frequency
// of your CPU manually in latency-graph.c
// for example
// cpu_hz=400000000.0;  for a 400Mhz CPU


extern double cpu_hz;

//#define DEBUG 1

#ifdef DEBUG
#define dprintf printf
#else
#define dprintf if(0) printf
#endif


#define MAXLATENCIES 8


extern  int lat_numsamples;
extern  int lat_numlatencies;
extern  float *lat_time_arr[MAXLATENCIES];
extern  float *lat_latency_arr[MAXLATENCIES];
extern  float lat_max_timediffs[MAXLATENCIES];
extern  double lat_deadline[MAXLATENCIES];
extern  int lat_overruns;
extern  int lat_actual_sample[MAXLATENCIES];
extern  double lat_time_start[MAXLATENCIES];

extern  unsigned long lat_flags[MAXLATENCIES];

extern  double lat_range1low[MAXLATENCIES];
extern  double lat_range1high[MAXLATENCIES];
extern  double lat_range2low[MAXLATENCIES];
extern  double lat_range2high[MAXLATENCIES];

extern  double lat_range1frac[MAXLATENCIES];
extern  double lat_range2frac[MAXLATENCIES];

extern  char lat_color_R[MAXLATENCIES];
extern  char lat_color_G[MAXLATENCIES];
extern  char lat_color_B[MAXLATENCIES];



#include <sched.h>

int calibrate_loop(void);
void mydelay(int loops);


#ifdef USE_GENERIC_TIMER
C_LINKAGE double lat_gettime(void);
#endif

#ifndef USE_GENERIC_TIMER
  C_LINKAGE unsigned long long int rdtsc();
  #define lat_gettime() (rdtsc() / cpu_hz)
#endif

#define LAT_SHOW_DEADLINE 1
#define LAT_SHOW_RANGE1   2
#define LAT_SHOW_RANGE2   4

C_LINKAGE int set_realtime_priority(int policy,int max);
C_LINKAGE int lat_init(int numlatencies,int numsamples);
C_LINKAGE void lat_set_flags(int numlatency,unsigned long flags);
C_LINKAGE void lat_set_deadline(int numlatency,double deadline);
C_LINKAGE void lat_set_range1(int numlatency,double low,double high);
C_LINKAGE void lat_set_range2(int numlatency,double low,double high);
C_LINKAGE void lat_set_color_main(int numlatency,char R,char G,char B);
C_LINKAGE void lat_set_color_deadline(int numlatency,char R,char G,char B);
C_LINKAGE void lat_set_color_range1(int numlatency,char R,char G,char B);
C_LINKAGE void lat_set_color_range2(int numlatency,char R,char G,char B);
C_LINKAGE void draw_latency_graph(char *filename);
C_LINKAGE void dump_samples( char *filename);
C_LINKAGE void dump_sample( int num, char *filename);
C_LINKAGE int lat_check_sampleranges ( void );

#define lat_record_start(a) lat_time_start[a]=lat_gettime()
#define lat_record_actual_time(b,a) lat_time_arr[b][lat_actual_sample[b]]=a-lat_time_start[b]
#define lat_record_latency(a,b) lat_latency_arr[a][lat_actual_sample[a]]=b
#define lat_next_sample(a) lat_actual_sample[a]++
#define lat_check_samplerange(a) ( (lat_actual_sample[a] < lat_numsamples) ? 0 : 1)

#endif

