/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
**!!!                                                                   
**!!!                   INFO SOLUTION SPA                     
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                    
**!!! PROJECT       : ROBY                                 
**!!!                                                                    
**!!! CLASS         : ModuleExpert                     
**!!!
**!!! PACKAGE       : Middleware
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                   
**!!! FILE NAME     : ModuleExpert.h
**!!!                                                                   
**!!! CREATION DATE : 25.10.2001         
**!!!                                                                   
**!!! AUTHOR        : P.M.  Roby  Team                             
**!!!                 Cristian Giussani
**!!!                                                                   
**!!!                                                                   
**!!!___________________________________________________________________
**!!!                                                                   
**!!! SPECS ON DOCS : <Spec. Doc Name A>  Rev.: AA.aa                   
**!!!                                                                     
**!!!                 <Spec. Doc Name B>  Rev.: BB.bb                   
**!!!                ............                                       
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
**!!!                                                                   
**!!! OVERVIEW      : ModuleExpert makes MIDDLEWARE transparent to 
**!!!	              expert modules (Task execution and message passing)
**!!!                 MIDDLEWARE is DCDT.       
**!!!                                                                   
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
*/

#ifndef __MODULE_EXPERT_H
#define __MODULE_EXPERT_H
/*----------------------------------------------------------------------
 *--------------------- SYSTEM  INCLUDE FILES --------------------------
 *---------------------------------------------------------------------*/
#include <DCDT_Member.h>
#include <DCDT_Agora.h>
#include <DCDT_Msg.h>

#include "assert.h"
#include "stl.h"
/*----------------------------------------------------------------------
 *--------------------- PROJECT INCLUDE FILES --------------------------
 *---------------------------------------------------------------------*/
#include "ModuleDispatch.h"

/*----------------------------------------------------------------------
 *-----------------------  DEFINE MACRO  -------------------------------
 *---------------------------------------------------------------------*/
const unsigned short ALL_TYPES=0;
/*----------------------------------------------------------------------
 *-----------------------  DEFINE TYPES  -------------------------------
 *---------------------------------------------------------------------*/

typedef enum {LOCAL,REMOTE} RequestType;
typedef list <DCDT_Msg *> PrivateMessagesList;

typedef vector <int> ReqMessagesList;
/*----------------------------------------------------------------------
 *-----------------------  DEFINE CLASS  -------------------------------
 *---------------------------------------------------------------------*/

/*! \brief This class is the interface that provided the basic functions for creating and senting messages.
 *
 *  The ModuleExpert class is an abstract class ( since in contains three pure
 *  pure virtual methods which are inherited from the class @ref DCDT_Member )
 *  that gives the necessary interface to build messages.
 */
class ModuleExpert: public DCDT_Member
{

  /*------------------- DEFINE CONSTRUCTOR/DESTRUCTOR -------------------*/
 protected:
  /** default Constructor **/
  ModuleExpert( ModuleDispatch *Dispatch,
		unsigned int period=10000 );

  /** copy Constructor **/ 
   
  /** Destructor **/ 
 virtual  ~ModuleExpert(void);
  
  /*---------------------------- PUBLIC ---------------------------*/
 public:
 /** Attribute: **/
 /** Methods: **/
 // ModuleExpert's sons have to implement these methods
 virtual void RunDuty(void)=0;
 virtual void RunInit(void)=0;
 virtual void RunClose(void)=0;
 /*********************************************
  **--          MESSAGE MANAGEMENT FUNCTIONS
  **-------------------------------------------
  **/
 /*! \brief Send the message pointed by \a Buffer through DCDT.
  *
  *  Copy the message pointed by \a Buffer in a new memory block and then send
  *  the newly allocated message through DCDT.
  *
  *  \param Buffer - Pointer to the message
  *  \param Length - Message length
  *  \param Type - Message type
  *  \param Priority - Message priority
  *  \return RC_OK - Messages successfully sent
  *  \return RC_ERROR - Error while sending the message
  */
 int SendMessage(unsigned char * Buffer, 
		 unsigned short  Length,
		 unsigned short Type,
		 char Priority=0);

 /*! \brief Receive a message of type \a Type through DCDT ( it allocates the buffer )
  *
  *   NB: You have to delete the buffer returned.
  *   \param Type - type of message to be received
  *   \param Buffer - pointer to the message received
  *   \param wait - If wait == true this method will wait until a message was found
  *   \return 0 - No new message available
  *   \return >0 - Size of the received message
  */
 unsigned short ReceiveMessage(unsigned short &Type, 
			       unsigned char * &Buffer,
                               bool wait=false );

 /*! \brief Get the last message received ( it allocates the buffer ).
  *
  *   Get the last message received, first check in the member queue then if no message is found check the private message list
  *   The buffer returned must be freed by the calling process.
  *
  *   \param Type - Type of message to be received
  *   \param Buffer - Pointer to the message received
  *   \param wait - If wait == true this method will wait until a message was found
  *   \return 0 - No new message available
  *   \return >0 - Size of the received message
  */
 unsigned short ReceiveLastMessage( unsigned short &Type,
				    unsigned char * &Buffer,
				    bool wait=false );

 /*! \brief Get the last message received ( it allocates the buffer ).
  *
  *   Get the last message received, first check in the member queue then if no message is found check the private message list
  *   The buffer returned must be freed by the calling process.
  *
  *   \param Buffer - Pointer to the message received
  *   \param Type - Type of the received message
  *   \return 0 - No new message available
  *   \return >0 - Size of the received message
  */

 //unsigned short ReceiveMessage(unsigned char * &Buffer, unsigned short &Type);

 /*! \brief Clear all messages of type \a Type in queue.
  *
  *  \param Type - Type of messages ( default ALL_TYPES )
  *  \return RC_OK - Messages successfully deleted
  *  \return RC_ERROR - Error deleting messages
  */
 int ClearMessages(unsigned short Type=ALL_TYPES);

 /*! \brief Subscribe to the messages of type \a Type.
  *
  *  \param Type - Type of messages to subscribe to
  *  \param Flag - Subscribe to local ( LOCAL ) or remote ( REMOTE ) messages
  *  \return RC_OK - Subscription ok
  *  \return RC_ERROR - Error during subscription
  */
 int AddMessageRequest(int Type, RequestType Flag);

 /*! \brief Unsubscribe to the messages of type \a Type.
  *
  *  \param Type - Type of messages to subscribe to
  *  \param Flag - Subscribe to local ( LOCAL ) or remote ( REMOTE ) messages
  *  \return RC_OK - Subscription ok
  *  \return RC_ERROR - Error during subscription
  */
 int RemMessageRequest(int Type, RequestType Flag);

 /*********************************************
  **--          MEMORY  MANAGEMENT FUNCTIONS
  **--          ( NOT IMPLEMENTED IN DCDT )
  **-------------------------------------------
  **/  
 /// Create the Memory Area ( Not implemented in DCDT )
 int CreateMemArea(int Size,int Name,int Type,int SecondType);
 /// Destroy The Memory Area ( Not implemented in DCDT )
 int DestroyMemArea(int Name);
 /// Check In to the memory Area ( Not implemented in DCDT )
 int MemCheckIn (int Name);
 /// Check Out to the memory Area ( Not implemented in DCDT )
 int MemCheckOut (int Name);
 /// Memory read function ( Not implemented in DCDT )
 int MemWrite (int Name,void * SourceBuffer);
 /// Memory Read function ( Not implemented in DCDT )
 int MemRead (int Name,void * DestBuffer);
 /*********************************************
  **--          UTILITY  FUNCTIONS
  **-------------------------------------------
  **/  
 /// Get Expert Id (Attributed at creation time by ModuleDispatch
 int GetId();

 /*********************************************
  **--          SHUTDOWN  
  **-------------------------------------------
  **/ 
 int Terminate(int uSec);
 /*--------------------------- PROTECTED -------------------------*/
 protected:
 /** Attribute: **/
 /* \brief List of local messages requested by this expert
  * 
  *  The expert uses this list to check if a messages published by
  *  the expert himself have to be saved in @ref Plist.
  */
 ReqMessagesList ReqMessages;
 string mModuleName;

 /// Dispatcher Reference
 ModuleDispatch* Dispatcher;

 /** Methods: **/ 
 
 /*--------------------------- PRIVATE ---------------------------*/
 private:
 /** Attribute: **/
  
 /// Expert Identifier
 int MyId;

#ifdef EXPERTS_TIME
 ///Time variable
 unsigned int mExpertTime;
#endif

 /// Pointer to the incoming message, used in Receive methods as temporary variable
 const DCDT_Msg *IncomingMessage;
 
 /// Messages published by the Expert himself
 PrivateMessagesList PList;

 // Overflow buffer ??? ( seems a virtual pointer added by someone .... ) 
 char buffer[4];
 
 /** Methods: **/
 // Redefinition of Middleware Init,Close and Main Process functions
 void DoYourJob(int wc);
 void Init(void);
 void Close(void);
 // Check if messages of type \a Type has been requested 
 /*! \brief Check if this Type of messages has been requested
  *
  *  \param Type - Type of messages
  *  \return true - Type of message requested
  *  \return false - Type of message not requested
  */
 bool IsPrivateReqMessages(int Type);
};
/*------------------- INLINE IMPLEMENTATION ------------------------*/

#endif
/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*-------------------------- End Of Header File --------------------*/
/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
