/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
**!!!                                                                    
**!!! PROJECT       : Middleware                                 
**!!!                                                                    
**!!! CLASS         : ModuleMember                     
**!!!
**!!! PACKAGE       : Middleware
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                   
**!!! FILE NAME     : ModuleExpert.h
**!!!                                                                   
**!!! CREATION DATE : 09-07-2005         
**!!!                                                                   
**!!! AUTHOR        : Cristian Giussani
**!!!
**!!!                                                                   
**!!! OVERVIEW      : ModuleMember makes MIDDLEWARE transparent to 
**!!!	              expert modules (Task execution and string message 
**!!!                 passing)
**!!!                 MIDDLEWARE is DCDT.       
**!!!                                                                   
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
*/

#ifndef __STRING_MODULE_MEMBER_H
#define __STRING_MODULE_MEMBER_H

#include "ModuleDispatch.h"
#include "ModuleMember.h"

#include <sstream>
#include <string>

#include <type.h>

/*! \brief This class is the interface that provided the basic functions for creating and senting messages.
 *
 *  The ModuleMember class is an abstract class ( since in contains three pure
 *  pure virtual methods which are inherited from the class @ref DCDT_Member )
 *  that gives the necessary interface to build messages.
 */
class StringModuleMember: public ModuleMember
{
  
  /*------------------- DEFINE CONSTRUCTOR/DESTRUCTOR -------------------*/
 protected:
  /** default Constructor **/
  StringModuleMember(ModuleDispatch *Dispatch,
		     unsigned int period=10000, string name="StringModuleMember" );
  
  /** Destructor **/ 
  virtual  ~StringModuleMember(void);
  
 public:
  // ModuleMember's sons have to implement these methods
  virtual void RunDuty(void)=0;
  virtual void RunInit(void)=0;
  virtual void RunClose(void)=0;

  /*
   * 
   */
  void NewMessage( int msg_id, string type );

  void SendStringMessage( int msg_id );

  void SendAllStringMessages( );

  void CloseCurrentMessage( );
  
  void CloseMessage();
  
  int strtovalue(char* buf, char* del,float* ris);
  
  weightfloat getFilteredData(weightfloat,weightfloat [],int,float,int*);
  pdistance ppdistance(pos,pos);
  float norangle(float);
  float a360to180(float);
  float a180to360(float);
  float degtorad(float);
  float radtodeg(float);

 protected:
  stringstream mCurrentMessage;
  bool mIsMessageOpened;
 
 private:
  int mMessageId;
  map<int, vector<string> > mMessages;

  /// Expert Identifier
  int MyId;
  
  /// Pointer to the incoming message, used in Receive methods as temporary variable
  const DCDT_Msg *IncomingMessage;
  
};

#endif
