/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
**!!!                                                                   
**!!!                   INFO SOLUTION SPA                     
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                    
**!!! PROJECT       : ROBY                                 
**!!!                                                                    
**!!! MODULE        : DATA_HANDLING                     
**!!!
**!!! PACKAGE       : DATA_HANDLING                                       
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                   
**!!! FILE NAME     : MessageManager.cpp
**!!!                                                                   
**!!! CREATION DATE : 18.10.2001         
**!!!                                                                   
**!!! AUTHOR        : P.M.  Roby  Team
**!!!                 Cristian Giussani
**!!!                                                                   
**!!!                                                                   
**!!!___________________________________________________________________
**!!!                                                                   
**!!! SPECS ON DOCS : <Spec. Doc Name A>  Rev.: AA.aa                   
**!!!                                                                     
**!!!                 <Spec. Doc Name B>  Rev.: BB.bb                   
**!!!                ............                                       
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
*/


/*----------------------------------------------------------------------
 *--------------------- SYSTEM  INCLUDE FILES --------------------------
 *---------------------------------------------------------------------*/
#include <stdio.h>
#include <assert.h>
#include <iostream>
/*----------------------------------------------------------------------
 *--------------------- PROJECT INCLUDE FILES --------------------------
 *---------------------------------------------------------------------*/
#include "MessageManager.h"
/*----------------------------------------------------------------------
 *-----------------------  DEFINE MACRO  -------------------------------
 *---------------------------------------------------------------------*/

/*----------------------------------------------------------------------
 *-----------------------  TYPES DECLARATION  --------------------------
 *---------------------------------------------------------------------*/

/*----------------------------------------------------------------------
 *--------------------  STATIC VARIABLE DECLARATION  -------------------
 *---------------------------------------------------------------------*/

/*----------------------------------------------------------------------
 *--------------------  STATIC FUNCTION DECLARATION  -------------------
 *---------------------------------------------------------------------*/

/*----------------------------------------------------------------------
 *--------------------  CONSTRUCTOR DESTRUCTOR  ------------------------
 *---------------------------------------------------------------------*/

void MsgObject::GetValues(GeneralMsgType* obj)
{
  assert(obj != 0);

  timestamp = obj->timestamp;
  kernel_id = obj->kernel_id;
  sender_id = obj->sender_id;
  strcpy(name,obj->name);
  //cout <<"\nGetValues "<<obj->num_attributes <<endl;
  for (int i = 0; i < obj->num_attributes; i++)
    {
      //cout <<"\nGetValues "<<obj->attributes[i].name <<"\t"<<obj->attributes[i].value <<endl;
      char* attr_name = (char*)malloc(sizeof(char) * (strlen(obj->attributes[i].name) + 1));
      strcpy(attr_name,obj->attributes[i].name);
      attributes.insert(pair<const char*,AttributeType> (attr_name, obj->attributes[i]));
    }
}

void BrowsingType::Initialize()
{
  message = 0;
  object = 0;
  attribute = 0;
}

void BrowsingType::Set(unsigned int msg, unsigned int obj, unsigned int att)
{
  message = msg;
  object = obj;
  attribute = att;
}

MessageManager::MessageManager(ModuleDispatch* Dispatch) : ModuleExpert(Dispatch,100000)
{
  mSendingMessages.clear();
  mRequestedMessages.clear();
  mBrowsingInfo.Initialize();
}

MessageManager::MessageManager(ModuleDispatch* Dispatch, DCDT_TIME period) : ModuleExpert(Dispatch,period)
{
  mSendingMessages.clear();
  mRequestedMessages.clear();
  mBrowsingInfo.Initialize();
}

MessageManager::~MessageManager()
{
}

/*----------------------------------------------------------------------
 *-------------------- PUBLIC METHODS DEFINITIONS ----------------------
 *---------------------------------------------------------------------*/
unsigned int MessageManager::AddObject(unsigned int message)
{
  MsgObject* pMsg = new MsgObject();

  gettimeofday(&(pMsg->timestamp),NULL);
  pMsg->kernel_id = 0; // ???
  pMsg->sender_id = GetId();
  pMsg->attributes.clear();

  mSendingMessages[message].push_back(pMsg);
  
  return (mSendingMessages[message].size() - 1);
}

unsigned int MessageManager::AddObject(unsigned int message, const char* name)
{
  MsgObject* pMsg = new MsgObject();

  gettimeofday(&(pMsg->timestamp),NULL);
  pMsg->kernel_id = 0; // ???
  pMsg->sender_id = GetId();
  pMsg->attributes.clear();
  if (name != 0 
      &&
      strlen(name) < STRING_LENGTH)
    {
      strcpy(pMsg->name,name);
    }
  else
    {
      cout << endl << "Object name NULL or too long" << endl;
    }

  mSendingMessages[message].push_back(pMsg);
  
  return (mSendingMessages[message].size() - 1);
}

void MessageManager::AddAttribute(unsigned int message,unsigned int object,const char* attribute, float value, float variance, float reliability)
{
#ifndef NDEBUG
  unsigned int init_num_attributes = mSendingMessages[message][object]->attributes.size();
#endif

  if ((mSendingMessages.find(message) != mSendingMessages.end()) && (object < mSendingMessages[message].size()))
    {
      AttributeType attr(value,variance,reliability,attribute);
      char* attr_name = (char*)malloc(sizeof(char) * (strlen(attribute) + 1));
      strcpy(attr_name,attribute);
      mSendingMessages[message][object]->attributes.insert(pair<const char*,AttributeType> ((const char*)attr_name,attr));
    }
  else
    {
      cout << "The specified message or object is unknown"<<endl;
    }
  
  assert(mSendingMessages[message][object]->attributes.size() == init_num_attributes + 1);
}

void MessageManager::SendGeneralMessage(unsigned int message)
{
  if (mSendingMessages.find(message) != mSendingMessages.end())
    {
      //cout << "Sending Message " << message << "\t" << mSendingMessages[message].size()<<endl;
      
      for (unsigned int num_obj = 0; num_obj < mSendingMessages[message].size(); num_obj++)
	{
	  GeneralMsgType* obj = (GeneralMsgType*)malloc(SizeOfMessage(mSendingMessages[message][num_obj]));

	  obj->kernel_id = mSendingMessages[message][num_obj]->kernel_id;
	  obj->sender_id = mSendingMessages[message][num_obj]->sender_id;
	  obj->num_attributes = mSendingMessages[message][num_obj]->attributes.size();
	  obj->timestamp = mSendingMessages[message][num_obj]->timestamp;
	  strcpy(obj->name,mSendingMessages[message][num_obj]->name);
	  map<const char*,AttributeType,ins_ord>::iterator it;
	  it = mSendingMessages[message][num_obj]->attributes.begin();
	  for (unsigned int num_attr = 0; num_attr < mSendingMessages[message][num_obj]->attributes.size(); num_attr++)
	    {
	      obj->attributes[num_attr] = it->second;
	      it++;
	    }

	  SendMessage((unsigned char*)obj,(unsigned short)SizeOfMessage(obj),(unsigned short)message,0 );
	  //:WARNING: obj has to be deleted
	  free(obj);
	}

      unsigned int n = mSendingMessages[message].size();
      for (unsigned int i = 0; i < n; i++)
	{
	  delete(mSendingMessages[message][i]);
	  //:WARNING: chi distrugge le stringhe? ???
	}
  
      mSendingMessages[message].clear(); 
    }
  else
    {
      cout << "The specified message is unknown" << endl;      
    }
}

void MessageManager::SendGeneralMessages()
{
  GeneralMsgMapType::iterator it;
  for (it = mSendingMessages.begin(); it != mSendingMessages.end(); it++)
    {
      //cout << "\nSending " << it->first << endl;
      SendGeneralMessage(it->first);
    }
}

void MessageManager::AddGeneralMsgRequest(unsigned int message, RequestType request)
{
  AddMessageRequest(message,request);

  MsgObject* pMsg = new MsgObject();
  gettimeofday(&(pMsg->timestamp),NULL);
  pMsg->sender_id = GetId();
  pMsg->kernel_id = 0; // ???

  mRequestedMessages[message].push_back(pMsg);
  mRequestedMessages[message].clear();
}

void MessageManager::RemGeneralMsgRequest(unsigned int message, RequestType request)
{
  RemMessageRequest(message,request);
  
  mRequestedMessages.erase(message);
}

/*
  Cancella tutti i messaggi in coda di tipo "message" quindi accoda tutti i nuovi messaggi ricevuti.

  NB Non e' possibile ricevere messaggi a lunghezza 0.
*/
void MessageManager::ReceiveGeneralMessage( unsigned short message )
{
  GeneralMsgType *incoming_msg;
  unsigned char *buffer;
  unsigned short type=message;

  mBrowsingInfo.Initialize();
  EraseGeneralMessage( message );
  
  while ( ReceiveMessage( type, buffer ) != 0 )
    {
      incoming_msg = (GeneralMsgType *)buffer;
      MsgObject* msg = new MsgObject();
      msg->GetValues( incoming_msg );

      delete incoming_msg;

      mRequestedMessages[type].push_back(msg);
      type=message;
    }
}

void MessageManager::ReceiveGeneralMessages()
{
  mBrowsingInfo.Initialize();
  EraseGeneralMessages();
  ReceiveGeneralMessage( ALL_TYPES );
}

void MessageManager::EraseGeneralMessage(unsigned int message)
{
  unsigned int n = mRequestedMessages[message].size();

  if ( message != ALL_TYPES ) {
    for (unsigned int i = 0; i < n; i++)
      {
	delete(mRequestedMessages[message][i]);
	//:WARNING: chi distrugge le stringhe?
      }
    mRequestedMessages[message].clear();
  }
  else
    EraseGeneralMessages();
}


void MessageManager::EraseGeneralMessages()
{
  unsigned int n = 0;
  
  GeneralMsgMapType::iterator it;
  for (it = mRequestedMessages.begin(); it != mRequestedMessages.end(); it++)
    {
      n = mRequestedMessages[it->first].size();
      
      for (unsigned int i = 0; i < n; i++)
	{
	  delete(mRequestedMessages[it->first][i]);
	  //:WARNING: chi distrugge le stringhe?
	}

      mRequestedMessages[it->first].clear();
    }

}

AttributeType* MessageManager::GetFirstAttribute(unsigned int msg, unsigned int obj)
{
  return GetAttribute(msg,obj,0);
}

AttributeType* MessageManager::GetFirstAttribute(unsigned int msg)
{
  return GetAttribute(msg,0,0);
}

AttributeType* MessageManager::GetFirstAttribute()
{
  //cout <<"\nGetFirstAttribute "<< mRequestedMessages.size() <<endl;
  return GetAttribute(0,0,0);
}

AttributeType* MessageManager::GetNextAttribute()
{
  //cout <<"\nGetAttribute "<< mRequestedMessages.size()<<endl;
  AttributeType* attribute = 0;

  if (IsLastAttribute())
    {
     //cout<<"\nLAST_ATTRIBUTE";fflush(stdout);
      attribute = 0;
    }
  else if (IsLastAttributeOfMessage())
    {
     //cout<<"\nLAST_OF_MESSAGE";fflush(stdout);
      attribute = GetAttribute(++mBrowsingInfo.message,0,0);
    }
  else if (IsLastAttributeOfObject())
    {
     //cout<<"\nLAST_OF_OBJECT";fflush(stdout);
      attribute = GetAttribute(mBrowsingInfo.message,++mBrowsingInfo.object,0);     
    }
  else
    {
     //cout<<"\nNEXT_ATTRIBUTE";fflush(stdout);
      attribute = GetAttribute(mBrowsingInfo.message,mBrowsingInfo.object,++mBrowsingInfo.attribute);
    }

  return  attribute; 
}

bool MessageManager::IsLastAttributeOfObject()
{
  bool last = false;
  unsigned int msg_num = GetMessage(mBrowsingInfo.message);

  if ((mRequestedMessages[msg_num][mBrowsingInfo.object]->attributes.size() == 0)
      ||
      (mBrowsingInfo.attribute >= (mRequestedMessages[msg_num][mBrowsingInfo.object]->attributes.size() - 1)))
    {
      last = true;
    }

  return  last;
}

bool MessageManager::IsLastAttributeOfMessage()
{
  bool last = false;
  unsigned int msg_num = GetMessage(mBrowsingInfo.message);

  if ( (mRequestedMessages[msg_num].size() == 0)
       ||
       ((mBrowsingInfo.object >= mRequestedMessages[msg_num].size() - 1)
	&& 
	((mRequestedMessages[msg_num][mBrowsingInfo.object]->attributes.size() == 0)
	 ||
	 (mBrowsingInfo.attribute >= mRequestedMessages[msg_num][mBrowsingInfo.object]->attributes.size() - 1))))
    {
      last = true;
    }

  return  last;
}

bool MessageManager::IsLastAttribute()
{
  bool last = false;
  unsigned int msg_num = GetMessage(mBrowsingInfo.message);

  if ( ((mRequestedMessages.size() == 0)
	||
	(mBrowsingInfo.message >= mRequestedMessages.size() - 1))
       && 
       ((mRequestedMessages[msg_num].size() == 0)
	||
	(mBrowsingInfo.object >= mRequestedMessages[msg_num].size() - 1))
       &&
       ((mRequestedMessages[msg_num][mBrowsingInfo.object]->attributes.size() == 0)
	||
	(mBrowsingInfo.attribute >= mRequestedMessages[msg_num][mBrowsingInfo.object]->attributes.size() - 1)))
    {
      last = true;
    }

  return  last;
}

AttributeType* MessageManager::GetAttribute(unsigned int msg, unsigned int obj, unsigned int num)
{
  AttributeType* attribute = 0;
  unsigned int msg_num;
  //cout <<"\nGetAttribute "<< mRequestedMessages.size()<<endl;

  mBrowsingInfo.Set(msg,obj,num);

  if (msg < mRequestedMessages.size())
    {
      msg_num = GetMessage(msg);
      //cout <<"\nGetAttribute "<< mRequestedMessages.size()<<"\t"<<msg_num<<endl;
    }
  else
    {
      return 0;
    }
 //cout <<"\nGetAttribute "<< msg <<"\t"<<msg_num<<"\t"<<mRequestedMessages[msg_num].size() <<endl;
  

  if (mRequestedMessages[msg_num].size() == 0)
    {
      attribute = GetNextAttribute();//:TEST: seri dubbi!!!
    }
  else
    {
      map<const char*,AttributeType,ins_ord>::iterator it;
      it = mRequestedMessages[msg_num][obj]->attributes.begin();
      for (unsigned int i = 0; i < num; i++)
	{
	  it++;
	}
      attribute = &(it->second);                                    ;
    }
  //cout <<"\nGetAttribute "<< mRequestedMessages.size() << "\t" <<endl;

  return attribute;
}

vector<AttributeType*> MessageManager::GetAttributeByName(const char* name)
{
  vector<AttributeType*> attributes;
  attributes.clear();

  AttributeType* pAttribute;
  pAttribute = GetFirstAttribute();
  while (pAttribute != 0)
    {
      if (strcmp(name,pAttribute->name) == 0)
	{
	  attributes.push_back(pAttribute);
	}
      pAttribute = GetNextAttribute();
    } 
  
  return attributes;
}

AttributeType* MessageManager::GetLatestAttributeByName(const char* name)
{
  AttributeType* attribute = 0;
  AttributeType* last_attribute = 0;
  BrowsingType last_browsed;
 //cout <<"\nLatest 1 "<<endl;
   
  attribute = GetFirstAttribute();
  last_browsed = mBrowsingInfo;
 //cout <<"\nLatest 2 "<<endl;
  while (attribute != 0)
    {
      if ((strcmp(name,attribute->name) == 0 )
	  && 
	  ((mRequestedMessages[GetMessage(mBrowsingInfo.message)][mBrowsingInfo.object]->timestamp >=  mRequestedMessages[GetMessage(last_browsed.message)][last_browsed.object]->timestamp)
	   ||
	   (last_attribute == 0)))
	{
	 //cout <<"\nLatest 3 "<<endl;
	  last_attribute = attribute;
	  last_browsed = mBrowsingInfo;
	}
     //cout <<"\nLatest 4 "<<endl;
      attribute = GetNextAttribute();
     //cout <<"\nLatest 5"<<endl;
    } 
  
  mBrowsingInfo = last_browsed;

  return last_attribute;
}


unsigned int MessageManager::GetAttributesOfObject(unsigned int msg, unsigned int num)
{
  return mRequestedMessages[msg][num]->attributes.size();
}

unsigned short MessageManager::GetKernelOfObject(unsigned int msg, unsigned int num)
{
  return mRequestedMessages[msg][num]->kernel_id;
}

int MessageManager::GetSenderOfObject(unsigned int msg, unsigned int num)
{
  return mRequestedMessages[msg][num]->sender_id;
}


unsigned int MessageManager::GetFirstMessage()
{
  mBrowsingInfo.message = mRequestedMessages.begin()->first;

  return mBrowsingInfo.message;
}

unsigned int MessageManager::GetNextMessage()
{
  mBrowsingInfo.message++;

  return mBrowsingInfo.message;
}

bool MessageManager::IsLastMessage()
{
  bool last;
  GeneralMsgMapType::iterator it;

  it = --mRequestedMessages.end();
  if (GetMessage(mBrowsingInfo.message) == it->first)
    {
      last =true;
    }
  else
    {
      last = false;
    }

  return last;
}

bool MessageManager::IsArrivedMessage(unsigned int num)
{
  bool arrived = true;
  
  if (mRequestedMessages[num].size() == 0)
    {
      arrived = false;
    }

  return arrived;
}

unsigned int MessageManager::GetMessage(unsigned int pos)
{
  GeneralMsgMapType::iterator it;
 
  it = mRequestedMessages.begin();
  for (unsigned int i = 0; i < pos; i++)
    {
      it++;
    }

  return it->first;
}

unsigned int MessageManager::GetObjectsOfMessage(unsigned int num)
{
  return mRequestedMessages[num].size();
}


MsgObject* MessageManager::GetFirstObject()
{
  mBrowsingInfo.Set(GetFirstMessage(),0,0);  
  
  while ( mRequestedMessages[GetMessage(mBrowsingInfo.message)].size() == 0 ) {
    if ( IsLastMessage() )
      return NULL;

    mBrowsingInfo.message++;
  }

  return mRequestedMessages[GetMessage(mBrowsingInfo.message)][mBrowsingInfo.object];

  /*
  if (mRequestedMessages[GetMessage(mBrowsingInfo.message)].size() != 0)
    {
      return mRequestedMessages[GetMessage(mBrowsingInfo.message)][mBrowsingInfo.object];
    }
  else
    {
      return GetNextObject();
    }
  */
}

MsgObject* MessageManager::GetNextObject()
{

  mBrowsingInfo.object++;
  if ( mBrowsingInfo.object >= GetObjectsOfMessage(GetMessage(mBrowsingInfo.message)) ) {
    // End of object, go to the next message type
    mBrowsingInfo.message++;
    mBrowsingInfo.object=0;

    while ( mRequestedMessages[GetMessage(mBrowsingInfo.message)].size() == 0 ) {
      if ( IsLastMessage() )
	return NULL;
      
      mBrowsingInfo.message++;
    }
  }

  return mRequestedMessages[GetMessage(mBrowsingInfo.message)][mBrowsingInfo.object];
    
  /*
  mBrowsingInfo.object++;
  if (mBrowsingInfo.object >= GetObjectsOfMessage(GetMessage(mBrowsingInfo.message)))
    {
      mBrowsingInfo.object = 0;
      if (IsLastMessage())
	{
	  return 0;
	}
      else
	{
	  mBrowsingInfo.message++;
	}
    }
  
  return mRequestedMessages[GetMessage(mBrowsingInfo.message)][mBrowsingInfo.object];
  */
}

MsgObject* MessageManager::GetCurrentObject()
{
  return mRequestedMessages[GetMessage(mBrowsingInfo.message)][mBrowsingInfo.object];
}

map<const char*,AttributeType,ins_ord>* MessageManager::GetAttributesOfCurrentObject()
{
  return &(mRequestedMessages[GetMessage(mBrowsingInfo.message)][mBrowsingInfo.object]->attributes);  
}

void MessageManager::RemoveCurrentObject()
{
  vector<MsgObject*>::iterator it = mRequestedMessages[GetMessage(mBrowsingInfo.message)].begin();
  int cont = mBrowsingInfo.object;

  while (it != mRequestedMessages[GetMessage(mBrowsingInfo.message)].end() && cont > 0)
    {
      it++;
      cont--;
    }

  MsgObject* temp = *it;
  mRequestedMessages[GetMessage(mBrowsingInfo.message)].erase(it);
  delete temp;
}

vector<MsgObject*>* MessageManager::GetAllObjectsOfMessage(unsigned int message)
{
  return &mRequestedMessages[message];
}


/*----------------------------------------------------------------------
 *-------------------- PRIVATE  METHODS DEFINITIONS --------------------
 *---------------------------------------------------------------------*/
int MessageManager::SizeOfMessage(GeneralMsgType* message)
{
  int msg_length = 0;
  msg_length = sizeof(GeneralMsgType) + message->num_attributes * sizeof(AttributeType);
  return msg_length;
}

int MessageManager::SizeOfMessage(MsgObject* message)
{
  int msg_length = 0;
  msg_length = sizeof(GeneralMsgType) + message->attributes.size() * sizeof(AttributeType);
  return msg_length;
}

/*----------------------------------------------------------------------
 *-------------------- OPERATOR   DEFINITIONS --------------------------
 *---------------------------------------------------------------------*/

/*----------------------------------------------------------------------
 *----------------  STATIC FUNCTION  DEFINITIONS -----------------------
 *---------------------------------------------------------------------*/

/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : 
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  :   
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      :
**---                
**---                                                                   
**-------------------------------------------------------------------------
*/
