/*
RS232.C
RS232 module.

This file contains all function to manage communications on rs232 bus.
We don not use the Microchip original because it contains some bugs.

The subroutines are
putsUSART: it is the subroutine that writes a string on the serial bus. The string has placed in the PIC RAM memory: this means that the string must not be within the double quote (constant string). Example of use is putsUSART(variable); where variable is a string variable.
putrsUSART: it is the subroutine that writes a string on the serial bus. This string has placed in the PIC ROM memory: this means that the string is a constant string (write within the double quote). Example of use is putrsUSART("Welcome in AIRboard");.
ReadUSART: it is the subroutine that read a single character from the serial bus. Example of use is character = ReadUsart();.
WriteUSART: it is the subroutine that write a single character from the serial bus. Example of use is WriteUSART('\r');.
read_rs232: it is the subroutine called by serial bus isr. It retrieves the characters and forms the command string. The command string is created when the last received character is the terminal character (carriage return)
*/

#include <stdlib.h>
#include <string.h>
#include "rs232.h"
#include "definitions.h"




//write on rs232 a string
void putsUSART(char *data) {
/*	do
	{
		while(!(TXSTA & 0x02)); //wait for busy rs232
		TXREG = *data; //copy the data into transmission register
	} while( *(++data) );
*/

	while (*data){
		while(TXSTAbits.TRMT==0);
		TXREG=*data;
		data++;
	}
}

//write on rs232 a constant string
void putrsUSART(const rom char *data) {
	/*do
	{
		while(!(TXSTA & 0x02)); //wait for busy rs232
		TXREG = *data; //copy the data into transmission register
	} while( *(++data) );*/

	while (*data){
		while(TXSTAbits.TRMT==0);
		TXREG=*data;
		data++;
	}
}

//read a single byte (character) from rs232
char ReadUSART(void) {
	unsigned char temp;
	temp=RCREG;
	if(temp=='\n') return '\r';
	return temp; //return reception register
}

//write a single byte (character) on rs232
void WriteUSART(char c) {
	TXREG = c; //put character into transmission register
}

/*subroutines called by rs232 interrupt service routine.
It is called when a byte arrives on the serial bus.*/
void read_rs232() {
	static unsigned char inputBuffer[DIM_INPUT_BUFFER];
	static unsigned int inputBufferCounter = 0;
	
	
	//read a single character from serial bus
	inputBuffer[inputBufferCounter] = ReadUSART();

	//echo
	//WriteUSART(inputBuffer[inputBufferCounter]);

	//increase bytes number
	inputBufferCounter++;

	

	if (inputBufferCounter >= DIM_INPUT_BUFFER) {
		//buffer has reached its maximum capacity: counter and command buffer data must be reset
		inputBufferCounter = 0;
		//enable low priority interrupt
	}
  	else if (inputBuffer[inputBufferCounter-1] == '\r') {
	  	//insert string terminator
	  	inputBuffer[inputBufferCounter-1]='\0';

  		// copy inputBuffer into parsingInputBuffer
  		/*When we have a CR from serial, we copy the received buffer (inputBuffer)
		to this buffer (parsingInputBuffer) and we start to parse this one*/
  		memcpy((char *)parsingInputBuffer, (const char *)inputBuffer, inputBufferCounter-1);

  		parsingInputBuffer[inputBufferCounter-1]='\0';

   	//decrease bytes received because \r is not a valid byte for the command
   	byteReceived = inputBufferCounter-1;

    	//reset counter
    	inputBufferCounter = 0;

    	//set dataReady to true
 		dataReady = 1;
	}
}




