/*
 * AIRboardUI: graphical interface program to interact with airboard
 * through RS232 serial port.
 *
 * In order to compile this package you have to install the
 * Java Communications API on your system.
 * http://java.sun.com/products/javacomm/downloads/index.html
 * 
 * Bruschieri Stefano
 * Campari PierPaolo
 */

package airboardUI;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
//import java.io.File;
//import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import javax.swing.Timer;
import javax.comm.SerialPort;
import java.util.Vector;

/**
 * @author Bruschieri Stefano 
 */
public class Rules {

	private Integer rulesNumber;
	private Vector rules = new Vector();
	private String rulesNumberString;
    private AirboardFrame parent;	
    private SerialConnection connection;
    private int MAX_ERRORS = 5; /*max errors number*/
    private int MAX_NO_RESPONSE = 5;
    
    Timer timer;
    ProgressDialog pd;
    int i=0;
	int noResponseCounter = 0;
    
    /**
    * Creates a new vector containing commands. It loads the list of
    * rules from a file (AirboardFrame.RULES_FILE) and it creates the vector
    * putting the values of rules into the string representing commands. 
    *
    * @param parent The main frame
    */
    public Rules(AirboardFrame parent)throws IOException {
    	this.parent = parent;
    	this.connection = parent.getConnection();
    
    	rules = (this.parent).loadFuzzyData(/*"airboardUI/data/rules.dat"*/ AirboardFrame.RULES_FILE);
        this.rulesNumber = new Integer(this.rules.size());
        
        /* Create a vector of commands to send rules to pic */
        String rulesNumberString = (this.rulesNumber).toString();
		if(rulesNumberString.length()==1) rulesNumberString = "00"+rulesNumberString;
		if(rulesNumberString.length()==2) rulesNumberString = "0"+rulesNumberString;
		if(rulesNumberString.length()==0 || rulesNumberString.length()>3) 
			System.out.println("Error");
		rulesNumberString = "ewfrulenum=+"+rulesNumberString+"&+000";
        this.rulesNumberString = rulesNumberString;
     
    }
	
    /**
     *  Sends one rule each time through serial port. The <code>Rules</code>
     *  constructor generates a Vector containing the sequence of commands, but
     *  it doesn't send rules. In order to execute commands we have to call
     *  explicity this method. <br>
     *  Before sending each command, waits a period <code>DELAY_TIME</code> to
     *  ensure a little probability of errors. <br>
     *  In order to show what it is doing, it opens a progress bar. If any 
     *  errors occurs the bar becomes red and some diagnostic information is
     *  written under the bar.
     */
    public void transmitRules(){    	
    	pd = new ProgressDialog(this.parent, (this.rulesNumber).intValue() );
		
    	/* Setting number of rules*/
    	pd.setText("Setting number of rules");
		SerialPort sPort = this.connection.getSerialPort();
		OutputStream os;
		try {
			os = sPort.getOutputStream();
			for(int j=0; j<(this.rulesNumberString).length(); j++){
				os.write((int)((this.rulesNumberString).charAt(j)));
			} 
			os.write((int)'\n');

		} catch (IOException e) {
			sPort.close();
			System.err.println("OutputStream write error: " + e);
		}

    	/* timer */
    	timer = new Timer(AirboardFrame.DELAY_TIME, new ActionListener() {		
    		public void actionPerformed(ActionEvent evt) {
    		//int mf = 1;
    		SerialPort sPort;
    		OutputStream os;
    		InputStream is;
    	    	
    		if(!pd.hasBeenStopped() && 
    				i<=( (rulesNumber).intValue()-1)){
    					pd.progressBar.setValue(i);
    					pd.setText("Copying fuzzy rules");
    	
    					String lastInput = (connection.getLastInputString());
    					if(lastInput.length()>0){
    						if(lastInput.lastIndexOf("ERROR")!=-1 ){
    							//if( i==0 ) i=1;
    							System.out.println(i+" An error was occurred "/*+connection.getLastInputString()*/);
    							return;
    						} else { /*System.out.println(i+" OK");*/ i++; }
    			
    					} else {
    						noResponseCounter++;
    						System.out.println("No response");
    						//if( i==0 ) i=1;
    						if(noResponseCounter>=MAX_NO_RESPONSE) { 
    							pd.setCancelled("No response"); 
    							pd.setCompleted(); 
    						}
    						return;
    					}

    	
    					String toOutput;
    					toOutput = (rules.elementAt((i-1))).toString();

    					if(toOutput.length()==1) toOutput = "00"+toOutput;
    					if(toOutput.length()==2) toOutput = "0"+toOutput;
    					if(toOutput.length()==0 || toOutput.length()>3) 
    						System.out.println("Error");
    					String end;
    					end = String.valueOf((i-1));
    					if(end.length()==1) end = "00"+end;
    					if(end.length()==2) end = "0"+end;
    					if(end.length()==0 || end.length()>3) 
    						System.out.println("Error");
    		    		//ewfuzrules=+017&+006 scrivi regola numero 6
    					toOutput = "ewfuzrules=+"+toOutput+"&+"+end;
    					System.out.println(toOutput);
    	
    					connection.sendAString(toOutput);
    				} else { 
    					pd.setCompleted();
    					timer.stop(); 
    				}}
    		});

    		timer.start();
    }
	
}
