#ifndef MultiSerialCommunicationAIRBoard_h
#define MultiSerialCommunicationAIRBoard_h

#include "MultiSerialCommunication.h"


/**
 * This class implements the communication through serial devices toward the
 * control board produced inside the AIRLab.
 *
 * The SerialCommunicationAIRBoard class implements the methods to write to and
 * read from the AIRBoard control board. In particular, this class implements
 * two pure virtual methods defined in the @ref LowLevelCommunication interface.
 *
 * @short	Interface for the AIRBoard control board
 * @author	Marcello Restelli (restelli@elet.polimi.it)
 */

class MultiSerialCommunicationAIRBoard : public MultiSerialCommunication
{
 public:
  // LIFECYCLE

  /**
   * Default constructor
   */
  MultiSerialCommunicationAIRBoard();

  /**
   * Destructor
   */
  virtual ~MultiSerialCommunicationAIRBoard();


  // OPERATORS

  //Use compiler-generated copy constructor and assignment
  //MultiSerialCommunicationAIRBoard(const MultiSerialCommunicationAIRBoard& from);
  //MultiSerialCommunicationAIRBoard& operator=(const MultiSerialCommunicationAIRBoard& from);

  // OPERATIONS
  
  /**
   * This is a pure virtual method that communicates the data stored in 
   * mActuationValues to the AIRBoard
   */
  void WriteActuations();

  /**
   * This is a pure virtual method that receives the data sent by the AIRBoard
   * and stores them in mEncoderValues
   */
  void ReadEncoders();

  /**
   * Set to zero the speed of the motors and  enables them
   */
  void StartMotors();

  /**
   * Set to zero the speed of the motors and  disables them
   */
  void StopMotors();

 protected:

  /**
   * Set the speeds for motor 0, motor 1 and motor 2
   *
   * @param speed_motor_0	The speed of motor 0 (between -99 and +99)
   * @param speed_motor_1	The speed of motor 1 (between -99 and +99)
   * @param speed_motor_2	The speed of motor 2 (between -99 and +99)
   */
  void SetMotorsSpeed(int speed_motor_0, int speed_motor_1, int speed_motor_2);


  /**
   * Send the kick command
   *
   * @param kicker	The kicker that must be activated
   */
  void Kick(int kicker);

 private:

};

#endif
