#include "MultiSerialCommunication.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <iostream>

using namespace std;

// Definizioni del modulo
#define STATUS_OFF 0
#define STATUS_ON 1
#define SX 0
#define DX 1
#define BOTH 2

#define FALSE 0
#define TRUE 1
#define _POSIX_SOURCE 1


MultiSerialCommunication::MultiSerialCommunication()
{}

MultiSerialCommunication::~MultiSerialCommunication()
{
  delete mModemDevice;
}


void MultiSerialCommunication::InitCommunication(void)
{ short int i = 0;
  for(;i<DEVICE_NO;i++)
    {
      // Open modem device for reading and writing and not as controlling tty
      // because we don't want to get killed if linenoise sends CTRL-C.
      if ((fd[i] = open(mModemDevice[i], O_RDWR | O_NOCTTY )) <0)
         {perror(mModemDevice[i]); exit(-1); }
      tcgetattr(fd[i],&oldtio[i]); /* save current serial port settings */
      bzero(&newtio[i], sizeof(newtio[i])); /* clear struct for new port settings */
  
      // BAUDRATE: Set bps rate. You could also use cfsetispeed and cfsetospeed.
      //  CRTSCTS : output hardware flow control (only used if the cable has
      //            all necessary lines. See sect. 7 of Serial-HOWTO)
      //  CS8     : 8n1 (8bit,no parity,1 stopbit)
      //  CLOCAL  : local connection, no modem contol
      //  CREAD   : enable receiving characters
      if (mHardwareFlowControl[i])
	{
	  newtio[i].c_cflag = mBaudRate[i] | CRTSCTS | CS8 | CLOCAL | CREAD;
	}
      else
	{
	  newtio[i].c_cflag = mBaudRate[i] | CS8 | CLOCAL | CREAD;
	}
  
      //  IGNPAR  : ignore bytes with parity errors
      //  ICRNL   : map CR to NL (otherwise a CR input on the other computer
      //            will not terminate input)
      //  otherwise make device raw (no other input processing)
      newtio[i].c_iflag = IGNPAR | ICRNL;
  
      // Raw output.
      newtio[i].c_oflag = 0;
  
      //  ICANON  : enable canonical input
      //  disable all echo functionality, and don't send signals to calling program
      newtio[i].c_lflag = ICANON;
  
      //  initialize all control characters
      //  default values can be found in /usr/include/termios.h, and are given
      //  in the comments, but we don't need them here
      newtio[i].c_cc[VINTR]    = 0;     /* Ctrl-c */
      newtio[i].c_cc[VQUIT]    = 0;     /* Ctrl-\ */
      newtio[i].c_cc[VERASE]   = 0;     /* del */
      newtio[i].c_cc[VKILL]    = 0;     /* @ */
      newtio[i].c_cc[VEOF]     = 4;     /* Ctrl-d */
      newtio[i].c_cc[VTIME]    = 0;     /* inter-character timer unused */
      newtio[i].c_cc[VMIN]     = 0;//1;     /* blocking read until 1 character arrives */
      newtio[i].c_cc[VSWTC]    = 0;     /* '\0' */
      newtio[i].c_cc[VSTART]   = 0;     /* Ctrl-q */
      newtio[i].c_cc[VSTOP]    = 0;     /* Ctrl-s */
      newtio[i].c_cc[VSUSP]    = 0;     /* Ctrl-z */
      newtio[i].c_cc[VEOL]     = 0;     /* '\0' */
      newtio[i].c_cc[VREPRINT] = 0;     /* Ctrl-r */
      newtio[i].c_cc[VDISCARD] = 0;     /* Ctrl-u */
      newtio[i].c_cc[VWERASE]  = 0;     /* Ctrl-w */
      newtio[i].c_cc[VLNEXT]   = 0;     /* Ctrl-v */
      newtio[i].c_cc[VEOL2]    = 0;     /* '\0' */

      //  now clean the modem line and activate the settings for the port
      tcflush(fd[i], TCIFLUSH);
      tcsetattr(fd[i],TCSANOW,&newtio[i]);

      // terminal settings done, now handle input
  
    }
}


void MultiSerialCommunication::CloseCommunication(void)
{
 short int i = 0;
 for (;i<=DEVICE_NO;i++)
 {
  // restore the old port settings
  tcsetattr(fd[i],TCSANOW,&oldtio[i]);
  
  close (fd[i]);
 }
}

