#include "MultiSerialCommunicationAIRBOARD.h"
#include "MotorConstants.h"

#include <assert.h>
#include <sstream>
#include <string.h>

using namespace std;

const int WRITING_DELAY = 0;
const int DEVICE0 = 0;
const int DEVICE1 = 1;
const int MAX_SPEED = 100;

MultiSerialCommunicationAIRBoard::MultiSerialCommunicationAIRBoard()
{
  mModemDevice[DEVICE0] = strdup(USB0);
  mModemDevice[DEVICE1] = strdup(USB1);
  short int i = 0;
  for(;i<DEVICE_NO;i++)
   {  mBaudRate[i] = B115200;
      mHardwareFlowControl[i] = false;
   }
  InitCommunication();
  StartMotors();
}

MultiSerialCommunicationAIRBoard::~MultiSerialCommunicationAIRBoard()
{
  StopMotors();
  CloseCommunication();
}

void MultiSerialCommunicationAIRBoard::StartMotors()
{
  SetMotorsSpeed(0,0,0);
  Write("a2\r",DEVICE0);
  Write("a0\r",DEVICE1);
}

void MultiSerialCommunicationAIRBoard::StopMotors()
{
  SetMotorsSpeed(0,0,0);
  Write("b2\r",DEVICE0);
  Write("b0\r",DEVICE1);
}

void MultiSerialCommunicationAIRBoard::WriteActuations()
{
  Write("a2\r",DEVICE0);
  Write("a0\r",DEVICE1);

  SetMotorsSpeed(mActuationValues[SPEED_MOTOR_0],mActuationValues[SPEED_MOTOR_1],mActuationValues[SPEED_MOTOR_2]);

  if (mActuationValues.find(KICK) != mActuationValues.end())
    Kick(mActuationValues[KICK]);
}

void MultiSerialCommunicationAIRBoard::ReadEncoders()
{
  int n;


  Write(READ_SPEEDS,DEVICE0);
  ReadNumber(&n,DEVICE0);
  mEncoderValues[SPEED_MOTOR_0] = n;
  ReadNumber(&n,DEVICE0);
  mEncoderValues[SPEED_MOTOR_1] = n;

  Write(READ_SPEEDS,DEVICE1);
  ReadNumber(&n,DEVICE1);
  mEncoderValues[SPEED_MOTOR_2] = n;
  ReadNumber(&n,DEVICE1);

  Write(READ_ENCODERS,DEVICE0);
  ReadNumber(&n,DEVICE0);
  mEncoderValues[ENCODER_MOTOR_0] = n;
  ReadNumber(&n,DEVICE0);
  mEncoderValues[ENCODER_MOTOR_1] = n;

  Write(READ_ENCODERS,DEVICE1);
  ReadNumber(&n,DEVICE1);
  mEncoderValues[ENCODER_MOTOR_2] = n;
  ReadNumber(&n,DEVICE1);

}

void MultiSerialCommunicationAIRBoard::SetMotorsSpeed(int speed_motor_0, int speed_motor_1, int speed_motor_2)
{
  stringstream serial_message;
  string motor0("m2");
  string motor1("m0");
  string motor2("");
  int abs_speed_motor_0 = 0;
  int abs_speed_motor_1 = 0;
  int abs_speed_motor_2 = 0;

  if (speed_motor_0 >= 0)
    {
      motor0.append("+");
    }
  else
    {
      motor0.append("-");
    }

  if (speed_motor_1 >= 0)
    {
      motor1.append("+");
    }
  else
    {
      motor1.append("-");
    }

  if (speed_motor_2 >= 0)
    {
      motor2.append("+");
    }
  else
    {
      motor2.append("-");
    }

  abs_speed_motor_0 = abs(speed_motor_0);
  abs_speed_motor_1 = abs(speed_motor_1);
  abs_speed_motor_2 = abs(speed_motor_2);

  if (abs_speed_motor_0 > MAX_SPEED)
    {
      abs_speed_motor_0 = MAX_SPEED;
    }
  else if (abs_speed_motor_0 < 10)
    {
      motor0.append("00");
    }
  else if (abs_speed_motor_0 < 100)
    {
      motor0.append("0");
    }

  if (abs_speed_motor_1 > MAX_SPEED)
    {
      abs_speed_motor_1 = MAX_SPEED;
    }
  else if (abs_speed_motor_1 < 10)
    {
      motor1.append("00");
    }
  else if (abs_speed_motor_1 < 100)
    {
      motor1.append("0");
    }

  if (abs_speed_motor_2 > MAX_SPEED)
    {
      abs_speed_motor_2 = MAX_SPEED;
    }
  else if (abs_speed_motor_2 < 10)
    {
      motor2.append("00");
    }
  else if (abs_speed_motor_2 < 100)
    {
      motor2.append("0");
    }

  serial_message.str(string());

  serial_message << motor0 << abs_speed_motor_0 
                 << motor2 << abs_speed_motor_2
		 << "\r"; 
  
  Write(serial_message.str(),DEVICE0);

  serial_message.str(string());

  serial_message << motor1 << abs_speed_motor_1
		 << "\r"; 
  
  Write(serial_message.str(),DEVICE1);
}


void MultiSerialCommunicationAIRBoard::Kick(int kicker)
{
  if (kicker == 1)
    Write("k0\r",DEVICE1);

}
