#include "SerialCommunicationAIRBoard.h"
#include "MotorConstants.h"

#include <assert.h>
#include <sstream>
#include <string.h>
#include <const.h>

using namespace std;

const int WRITING_DELAY = 10000;

SerialCommunicationAIRBoard::SerialCommunicationAIRBoard()
{
#ifdef rabbiati
  mBaudRate = B115200;
#else
  mBaudRate = B115200;
#endif
  mModemDevice = strdup(AIRBOARD_OUT);
  mHardwareFlowControl = false;
  InitCommunication();
  StartMotors();
}

SerialCommunicationAIRBoard::~SerialCommunicationAIRBoard()
{
  StopMotors();
  CloseCommunication();
}

void SerialCommunicationAIRBoard::StartMotors()
{
  SetMotorsSpeed(0,0);
  usleep(WRITING_DELAY);
  //Write ("d2\r");
  Write("b2\r");
  usleep(WRITING_DELAY);
  Write("a2\r");
  usleep(WRITING_DELAY);
}

void SerialCommunicationAIRBoard::StopMotors()
{
  SetMotorsSpeed(0,0);
  usleep(WRITING_DELAY);
    Write("b2\r");
}

void SerialCommunicationAIRBoard::WriteActuations()
{
  //Write("b2\r");
  //Write("m2+000+000\r");
  //usleep(WRITING_DELAY);
  Write("a2\r");
  usleep(WRITING_DELAY);
	
	/*static int i=0;
	
	i++;
	if(i>10){
		i=0;
		Write("b2\r");
  		usleep(WRITING_DELAY);
  		Write("a2\r");
  		usleep(WRITING_DELAY);
	}*/
  SetMotorsSpeed(mActuationValues[SPEED_MOTOR_SX],mActuationValues[SPEED_MOTOR_DX]);

  if (mActuationValues.find(KICK) != mActuationValues.end())
    {
 //     usleep(WRITING_DELAY); /*wiim*
      Kick(mActuationValues[KICK]);
    }

}

void SerialCommunicationAIRBoard::ReadEncoders()
{
  int n;

//printf("+\n");

  usleep(WRITING_DELAY);
  Write(READ_ENCODER_DX);
  usleep(WRITING_DELAY);
  ReadNumber(&n);
  mEncoderValues[ENCODER_MOTOR_DX] = n;
  //printf( "lettura: mEncoderValues[ENCODER_MOTOR_DX] = %d\n", mEncoderValues[ENCODER_MOTOR_DX] );
  usleep(WRITING_DELAY);
  Write(READ_ENCODER_SX);
  usleep(WRITING_DELAY);
  ReadNumber(&n);
  mEncoderValues[ENCODER_MOTOR_SX] = n;
  //printf( "lettura: mEncoderValues[ENCODER_MOTOR_SX] = %d\n", mEncoderValues[ENCODER_MOTOR_SX] );
  
 /* usleep(WRITING_DELAY);
  ReadNumber(&n);
  mEncoderValues[SPEED_MOTOR_DX] = n;
  printf( "lettura: mEncoderValues[SPEED_MOTOR_DX] = %d\n", mEncoderValues[SPEED_MOTOR_DX] );

  usleep(WRITING_DELAY);
  Write(READ_ENCODERS);
  usleep(WRITING_DELAY);
  ReadNumber(&n);
  mEncoderValues[ENCODER_MOTOR_SX] = n;
  printf( "lettura: mEncoderValues[ENCODER_MOTOR_SX] = %d\n", mEncoderValues[ENCODER_MOTOR_SX] );
  usleep(WRITING_DELAY);
  ReadNumber(&n);
  mEncoderValues[ENCODER_MOTOR_DX] = n;
  printf( "lettura: mEncoderValues[ENCODER_MOTOR_DX] = %d\n", mEncoderValues[ENCODER_MOTOR_DX] );*/
  
  //printf("++++++\n"); //*wiim*
}

void SerialCommunicationAIRBoard::SetMotorsSpeed(int speed_motor_0, int speed_motor_1)
{
  speed_motor_1 = -speed_motor_1;
  stringstream serial_message;
  string m0("m2");
  string m1("");
  int abs_speed_motor_0 = 0;
  int abs_speed_motor_1 = 0;
  
  if (speed_motor_0 >= 0)
    {
      m0.append("+");
    }
  else
    {
      m0.append("-");
    }

  if (speed_motor_1 >= 0)
    {
      m1.append("+");
    }
  else
    {
      m1.append("-");
    }

  abs_speed_motor_0 = abs(speed_motor_0);
  abs_speed_motor_1 = abs(speed_motor_1);

  if (abs_speed_motor_0 > 100)
    {
      abs_speed_motor_0 = 100;
    }
  else if (abs_speed_motor_0 < 10)
    {
      m0.append("00");
    }
  else if (abs_speed_motor_0 < 100)
    {
      m0.append("0");
    }

  if (abs_speed_motor_1 > 100)
    {
      abs_speed_motor_1 = 100;
    }
  else if (abs_speed_motor_1 < 10)
    {
      m1.append("00");
    }
  else if (abs_speed_motor_1 < 100)
    {
      m1.append("0");
    }

  serial_message.str(string());

  serial_message << m0 << abs_speed_motor_0 
                 << m1 << abs_speed_motor_1 
		 << "\r"; 
  
 
  Write(serial_message.str());

}


void SerialCommunicationAIRBoard::Kick(int kicker)
{
  if (kicker == 0)
    {
      Write("k2\r");
    }
  else if (kicker == 1)
    {
      Write("k0\r");
    }
  else if (kicker == -1)
    {
      Write("k1\r");
    }

}
