#include "LogExpert.h"
#include "MotorConstants.h"
#include <msg.h>
#include <wiiuse.h>

#include <iostream>

#include <string>

#define MAXSTR	200


#define SC  2
#define PI 3.14159265358979323846

GLint width = 1024/SC, height = 768/SC;

typedef struct point_s{
	int x;
	int y;
	int valid;
}point;
point center;

extern wiimote_t* wm0;

using namespace std;

#define DRAW_TRIANGLE(x, y, z, s)	do {							\
										glVertex3f(x/SC, (y-s*SC)/SC, z/SC);		\
										glVertex3f((x+s*SC)/SC, (y+s*SC)/SC, z/SC);	\
										glVertex3f((x-s*SC)/SC, (y+s*SC)/SC, z/SC);	\
									} while (0)

#define DRAW_TRIANGLE(x, y, z, s)	do {							\
										glVertex3f(x/SC, (y-s*SC)/SC, z/SC);		\
										glVertex3f((x+s*SC)/SC, (y+s*SC)/SC, z/SC);	\
										glVertex3f((x-s*SC)/SC, (y+s*SC)/SC, z/SC);	\
									} while (0)
									
struct light_t {
	GLfloat position[4];
	GLfloat spotDirection[3];
	GLfloat ambient[4];
	GLfloat diffuse[4];
	GLfloat specular[4];
	GLfloat spotCutoff;
	GLfloat spotExponent;
	GLfloat spotAttenuation[3];  /*  [0] = constant, [1] = linear, [2] = quadratic  */
};
struct light_t light = {
	{ 1.0, 1.0, -2.0, 1.0	},
	{ -1.0, -1.0, 2.0		},
	{ 0.0, 0.0, 0.0, 1.0	},
	{ 1.0, 1.0, 1.0, 1.0	},
	{ 1.0, 1.0, 1.0, 1.0	},
	180.0, 0.0,
	{ 1.0, 0.0, 0.0			}
};
struct material_t {
	GLfloat ambient[4];
	GLfloat diffuse[4];
	GLfloat specular[4];
	GLfloat emission[4];
	GLfloat shininess;
};
struct material_t red_plastic = {
	{ 0.3, 0.0, 0.0, 1.0 },
	{ 0.3, 0.0, 0.0, 1.0 },
	{ 0.8, 0.6, 0.6, 1.0 },
	{ 0.0, 0.0, 0.0, 1.0 },
	32.0
};

// la parola extern non dichiara un nuovo identificatore, 
// ma dice "La variabile intera a e` dichiarata da qualche altra parte, 
// lascia solo lo spazio per risolvere il riferimento"

LogExpert::LogExpert(ModuleDispatch* kernel,int period) : 
  StringModuleMember(kernel,period,"LogExpert")
{
  printf("\nLogExpert built");
  fflush(stdout);
};

LogExpert::~LogExpert()
{
  printf("\nLogExpert destroyed");
  fflush(stdout);
};

void LogExpert::RunInit()
{
	//usleep(5000000);
	
	AddMessageRequest( MSG_POSITION,LOCAL );
	AddMessageRequest( MSG_TO_LOG,LOCAL );
	
	center.valid=0;
	frame=0;
	gettimeofday(&last_render, NULL);
	
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		printf("Failed to initialize SDL: %s\n", SDL_GetError());
	}
	SDL_WM_SetCaption("ROBOWII", "ROBOWII");
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);

	SDL_SetVideoMode(width, height*2, 16, SDL_RESIZABLE | SDL_OPENGL);
	glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
	glEnable(GL_NORMALIZE);
	glEnable(GL_BLEND);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDepthFunc(GL_LEQUAL);
    glClearColor(0, 0, 0, 0);
    
	glViewport(0, 0, width, height*2);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, width, height*2, 0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
//	resize_window(width, height*2);
	display();		
	
  	printf("\nLogExpert initialized\n");
  	fflush(stdout);
};

/* ***********************************
   
   DoYourDuty()

   *********************************** */
   

void LogExpert::RunDuty()
{

	unsigned char* buffer;
	char* buf;
	float f;
	int i,j;
	
	/*if (wm0!=NULL){
	    for (j=0; j < 4; j++) {
			if (wm0->ir.dot[j].visible){
				printf("IR%i ---: (%u, %u) %u\t\t", j, wm0->ir.dot[j].x, wm0->ir.dot[j].y,wm0->ir.dot[j].size);
			}else{
				printf("IR%i ***: \t \t\t", j, wm0->ir.dot[j].x, wm0->ir.dot[j].y,wm0->ir.dot[j].size);
			}
		}
		printf("\n");
	}*/
	
	
	if((ReceiveLastMessage(MSG_TO_LOG, buffer,false) > 0)){
   	//	printf ("\nricevuto log: \n%s\n",buffer);
    		for(i=0;buffer[i]!=NULL;i++){
				;
			}
    		buf=(char*)malloc(i+1);
    		strcpy(buf,(char*)buffer);
    		if(strtovalue(buf,"wiimote_data",NULL)){    		
				strtovalue(buf,"valid ",&f);
				if((int)f){
					center.valid=1;
					strtovalue(buf,"xcenter ",&f);
					center.x=f;
					strtovalue(buf,"ycenter ",&f);
					center.y=f;
				}else{
					center.valid=0;
				}
			}
			free(buf);
	}

	
	
	if((ReceiveLastMessage(MSG_POSITION, buffer,false) > 0)){
		
		    for(i=0;buffer[i]!=NULL;i++){
				;
			}
    		buf=(char*)malloc(i+1);
    		strcpy(buf,(char*)buffer);

			if(strtovalue(buf,"position_data",NULL)){
				if(strtovalue(buf,"posx ",&f)){
					px=f;
				}
				if(strtovalue(buf,"posy ",&f)){
					py=f;
				}
				if(strtovalue(buf,"angle ",&f)){
					pa=f;
				}			
				//printf("*%f*%f*%f*\n",px,py,pa);
			}
    		/*if(strtovalue(buf,"TanSpeed ",&f) || strtovalue(buf,"RotSpeed ",&f) ){
    			printf("*%d*\n%s\n",i,buf);
    			f=0;
    			if(strtovalue(buf,"TanSpeed ",&f)){
    				printf("TanSpeed=%f\n",f);
				}
				f=0;
       			if(strtovalue(buf,"RotSpeed ",&f)){
       				printf("RotSpeed=%f\n",f);
				}
			}*/
			free(buf);
	}
	

	SDL_Event event;

		if (SDL_PollEvent(&event)) {
			switch (event.type) {
				case SDL_QUIT:
				{
					/* shutdown */
					SDL_Quit();
					return;
				}
				default:
				{
				}
			}
	}
	display();		

	
}


void LogExpert::RunClose()
{
  printf("\nLogExpert closed");
  fflush(stdout);
};

void LogExpert::AddAttribute(string name, 
				       float value, 
				       float std_dev, 
				       float reliability)
{
  //To add an attribute to an object it is necessary that the object has been opened
  //assert(mIsObjectOpened);
  //mCurrentMessage<< "<C>TanSpeed++ 2.0</C>\r";
  /*
  mCurrentMessage.precision(5); //wiim è una variabile std::stringstream
  mCurrentMessage << "<C>" 
		  << name << " " << value << " " << std_dev << " " << reliability 
		  << "</C>\n";*/
}

void LogExpert::resize_window(GLint new_width, GLint new_height) {
	/*width = new_width;
	height = new_height;
	if (new_height == 0)
		new_height = 1;
	SDL_SetVideoMode(width, height, 16, SDL_RESIZABLE | SDL_OPENGL);
	glViewport(0, 0, new_width, new_height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, width, height, 0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	width = new_width;
	height = new_height;*/
}

void LogExpert::display() {
	int i;
	float size = 5;
	
	frame++;

	
	if(elapsed()<50000){//limito a 20 fps
		return;
	}else{
		gettimeofday(&last_render, NULL);
	}
		
	/*if(elapsed()>1000000){
		gettimeofday(&last_render, NULL);
		printf("fps:%d\n",frame);
		frame=0;
	}*/


	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	

		glDisable(GL_LIGHTING);
		
	GLint ic;
   	GLfloat cosine, sine;   	

//cerchi wiimote
  glBegin(GL_POLYGON);
  		glColor3f(0.1, 0.1, 0.1);
      for(ic=0;ic<100;ic++){
         cosine=cos(ic*2*PI/100);
         sine=sin(ic*2*PI/100);
         glVertex2f(width/2+(cosine*IRCENTERLIMIT/SC),height/2+(sine*IRCENTERLIMIT/SC));
      }
   glEnd();
    glBegin(GL_POLYGON);
  		glColor3f(0, 0.9, 0.9);
      for(ic=0;ic<100;ic++){
         cosine=cos(ic*2*PI/100);
         sine=sin(ic*2*PI/100);
         glVertex2f(width/2+(cosine*IRCENTERHIT/SC),height/2+(sine*IRCENTERHIT/SC));
      }
   glEnd();
   
   
 //cerchi campo robot  
    glLoadIdentity();									
	glTranslatef(0,height,0.0f);	
   glBegin(GL_POLYGON);
  		glColor3f(0.1, 0.1, 0.1);
      for(ic=0;ic<100;ic++){
         cosine=cos(ic*2*PI/100);
         sine=sin(ic*2*PI/100);
         glVertex2f(width/2+(cosine*200),height/2+(sine*200));
      }
   glEnd();
      
   	glLoadIdentity();									
	glTranslatef(0,height,0.0f);	
   glBegin(GL_POLYGON);
  		glColor3f(0, 0.9, 0.9);
      for(ic=0;ic<100;ic++){
         cosine=cos(ic*2*PI/100);
         sine=sin(ic*2*PI/100);
         glVertex2f(width/2+(cosine*160),height/2+(sine*160));
      }
   glEnd();
   
//robot
//il sistema di riferimento ha O nel centro del cerchio e gradi antiorari
	glLoadIdentity();									// Reset The Current Modelview Matrix
	glRotatef(180,1.0f,0.0f,0.0f);	
	glTranslatef((width/2+px/SC),(+py/SC-(height*1.5)),-0.02f);		
		
	glRotatef(pa+90,0.0f,0.0f,1.0f);						
	glBegin(GL_TRIANGLES);								// Start Drawing A Triangle
		glColor3f(1.0f,0.0f,0.0f);						// Set Top Point Of Triangle To Red
		DRAW_TRIANGLE(0,0, 0, size*2);
	glEnd();
	
	glTranslatef(0.0f,0.0f,-0.01f);	
	glBegin(GL_POLYGON);
  	glColor3f(1.0, 1.0, 1.0);
      for(ic=0;ic<100;ic++){
         cosine=cos(ic*2*PI/100);
         sine=sin(ic*2*PI/100);
         glVertex2f((cosine*3),(sine*3));
      }
   	glEnd();
   	
//goal-start
	glLoadIdentity();									
	glRotatef(180,1.0f,0.0f,0.0f);	
	glTranslatef((width/2+GOALX/SC),(+GOALY/SC-(height*1.5)),-0.01f);		
	glBegin(GL_POLYGON);
  	glColor3f(0.0, 0.0, 0.0);
      for(ic=0;ic<100;ic++){
         cosine=cos(ic*2*PI/100);
         sine=sin(ic*2*PI/100);
         glVertex2f((cosine*GOALRADIUS/SC),(sine*GOALRADIUS/SC));
      }
   	glEnd();
   	glLoadIdentity();									
	glRotatef(180,1.0f,0.0f,0.0f);	
	glTranslatef((width/2+STARTX/SC),(+STARTY/SC-(height*1.5)),-0.01f);		
	glBegin(GL_POLYGON);
  	glColor3f(1.0, 1.0, 1.0);
      for(ic=0;ic<100;ic++){
         cosine=cos(ic*2*PI/100);
         sine=sin(ic*2*PI/100);
         glVertex2f((cosine*HOMERADIUS/SC),(sine*HOMERADIUS/SC));
      }
   	glEnd();

if(wm0!=NULL){

		glLoadIdentity();
		glBegin(GL_TRIANGLES);
			glColor3f(0.1, 0.9, 0.1);
			DRAW_TRIANGLE(width, height, 0, size);
			
		glEnd();

			glLoadIdentity();
			glBegin(GL_TRIANGLES);

				glColor3f(1.0, 0.0, 0.0);
				for (i = 0; i < 4; ++i) {
					if (wm0->ir.dot[i].visible)
						DRAW_TRIANGLE(wm0->ir.dot[i].rx, wm0->ir.dot[i].ry, 0, size);
				}

				glColor3f(1.0, 1.0, 0.0);
				for (i = 0; i < 4; ++i) {
					if (wm0->ir.dot[i].visible)
						DRAW_TRIANGLE(wm0->ir.dot[i].x, wm0->ir.dot[i].y, 0, size);
				}

				glColor3f(0.0, 0.0, 1.0);
				DRAW_TRIANGLE(wm0->ir.ax, wm0->ir.ay, 0, size);
			glEnd();

	if(center.valid){
	glLoadIdentity();
	glBegin(GL_TRIANGLES);
		glColor3f(1.0, 1.0, 1.0);
		DRAW_TRIANGLE(center.x, center.y-size, 0, size);
	glEnd();
	}


}
	SDL_GL_SwapBuffers();
	


}


long LogExpert::elapsed()
{ //possibile metodo di temporizzazione, alternativo a dcdt
	struct timeval now;
	long elapsed_usec = 0;

	gettimeofday(&now, NULL);

	if (now.tv_usec > 1000000) {
		now.tv_usec -= 1000000;
		++now.tv_sec;
	}

	if (now.tv_sec > last_render.tv_sec)
		elapsed_usec = ((now.tv_sec - last_render.tv_sec) * 1000000);

	if (now.tv_usec > last_render.tv_usec)
		elapsed_usec += now.tv_usec - last_render.tv_usec;
	else
		elapsed_usec += last_render.tv_usec - now.tv_usec;


	return elapsed_usec ;
}


/*
point getcenter() //andrebbe presa da wiim.cpp
{
	//estendere ai casi con punti != 2 (guardare le librerie)
	int x,y,x1,x2,y1,y2;
	int i1=-1,i2=-1;
	int i;
	point p;
	
	if (wm0->ir.num_dots == 2){
		for (i = 0; i < 4; ++i) {
			if (wm0->ir.dot[i].visible){
				if (i1==-1){
					i1=i;
				}else if(i2==-1){
					i2=i;
				}
			}
		}
	x1=wm0->ir.dot[i1].x;
	x2=wm0->ir.dot[i2].x;
	y1=wm0->ir.dot[i1].y;
	y2=wm0->ir.dot[i2].y;
	p.x=(x1+x2)/2;
	p.y=(y1+y2)/2;
	p.valid=1;
	
	//printf("%d-%d-%d-%d-%d-%d\n",x1,y1,x2,y2,p.x,p.y);
	
	}else{
		p.valid=0;
		return p;
	}
}
*/
