#include "ARToolKitWrapper.h"
#include "MatrixFunctions.h"
#include "WheelChairUtil.h"

using namespace ARToolKitPlus;
using namespace std;

class MyLogger : public Logger
{
    void artLog(const char* nStr)
    {
        printf(nStr);
    }
};

ARToolKitWrapper::ARToolKitWrapper(string cameraCfgFile, double tagDim){
	MyLogger logger;
	markerId=-1;
	tagDimension=tagDim;
	matrix=new Matrix3DRotoTraslation();
	tracker=new TrackerSingleMarkerImpl<6,6,6, 1, 8>(img_width ,img_height);
	tracker->setLogger(&logger);
	tracker->setPixelFormat(ARToolKitPlus::PIXEL_FORMAT_LUM);
	if(!tracker->init(cameraCfgFile.c_str(), 1.0f, 1000.0f)){
		printf("ERROR: init() tracker failed\n");
		delete tracker;
		/*
		 *   GESTIRE MEGLIO ERRORI!!!
		 */
		exit(-1);
		return ;	
	}
	// define size of the marker
	tracker->setPatternWidth(tagDim);
	tracker->setBorderWidth( 0.250f);
	tracker->activateAutoThreshold(true);
	//tracker->setUndistortionMode(ARToolKitPlus::UNDIST_LUT);
	tracker->setPoseEstimator(ARToolKitPlus::POSE_ESTIMATOR_RPP);
	tracker->setMarkerMode( ARToolKitPlus::MARKER_ID_SIMPLE);


}

ARToolKitWrapper::~ARToolKitWrapper(){
	delete matrix;
	delete tracker;
}


int ARToolKitWrapper::processImage(unsigned char * img){
	// restituisce il numero di tag trovato
	bool allZero=true;

	ARToolKitPlus::ARMarkerInfo* markerInfo;
	int numMarkers;
	//così trova il marker migliore
	markerId = tracker->calc(img,-1,true, &markerInfo,&numMarkers);
	//markerId=tracker->calc(img);


	if(markerId>=0){

		for(int i=0;i<numMarkers;i++){
			DBG("id[%d]=%d\n",i,markerInfo[i].id);
			if(markerInfo[i].id==markerId){
				for(int j=0;j<4;j++){
					DBG("x: %f y: %f\n",markerInfo[i].vertex[j][0],markerInfo[i].vertex[j][1]);
					xpoint[j]=markerInfo[i].vertex[j][0];
					ypoint[j]=markerInfo[i].vertex[j][1];
					DBG("*** x: %f y: %f\n",xpoint[j],ypoint[j]);
				}
				break;
			}
		}

		//printf("out\n");
		for(int j=0; j<3; j++){//faccio solo le prime tre righe, la quarta è [0 0 0 1]
			for(int i=0; i<4; i++){
				matrix->setValue(j,i,tracker->getModelViewMatrix()[i*4+j]);
				if(!(i==3 && j==3)){
					if(fabs(tracker->getModelViewMatrix()[i*4+j])>all_zero_esp){
						//printf("artoolkit::allzero=false\n");
						allZero=false;
					}
				}
			}
		}
		//se allZero=true, la matrice non è valida! (anche se markerId è >=0 capita)
		if(allZero)markerId=-1;
	}
	//print3DMatrix(matrix,"artoolkit::matrix:\n","\n\n");
	if(allZero)return -1;
	return markerId;
}

Matrix3DRotoTraslation * ARToolKitWrapper::getMatrix3D(){
	return matrix;
}

