#include "FireIAcquisition.h"
#include <stdlib.h>
#include <string.h>
using namespace std;

FireIAcquisition::FireIAcquisition(string nameDev,bool flagSaveFrame,string p,string s,string f)
{
	/*-----------------------------------------------------------------------
	*  Open ohci and asign handle to it
	*-----------------------------------------------------------------------*/
	count=0;
	saveFrame=flagSaveFrame;
	prefix=p;
	suffix=s;
	iformat=f;

	deviceName=nameDev;
	mHandle = dc1394_create_handle(0);
	if (mHandle==NULL)
	{
		fprintf( stderr, "Unable to aquire a raw1394 handle\n\n"
		       "Please check \n"
		       "  - if the kernel modules `ieee1394',`raw1394' and `ohci1394' are loaded \n"
		       "  - if you have read/write access to /dev/raw1394\n\n");
		return;
	}
  
	int numNodes = raw1394_get_nodecount(mHandle);
	int numCameras = 0;
	mCameraNodes = dc1394_get_camera_nodes(mHandle,&numCameras,1);
	fflush(stdout);
	if (numCameras<1){
		fprintf( stderr, "no cameras found :(\n");
		raw1394_destroy_handle(mHandle);
		mHandle=NULL;
		return;
	}
	printf("working with the first camera on the bus\n");
  
	/*-----------------------------------------------------------------------
	*  to prevent the iso-transfer bug from raw1394 system, check if
	*  camera is highest node. For details see 
	*  http://linux1394.sourceforge.net/faq.html#DCbusmgmt
	*  and
	*  http://sourceforge.net/tracker/index.php?func=detail&aid=435107&group_id=8157&atid=108157
	*-----------------------------------------------------------------------*/
	if( mCameraNodes[0] == numNodes-1){
		fprintf( stderr, "\n"
	       "Sorry, your camera is the highest numbered node\n"
	       "of the bus, and has therefore become the root node.\n"
	       "The root node is responsible for maintaining \n"
	       "the timing of isochronous transactions on the IEEE \n"
	       "1394 bus.  However, if the root node is not cycle master \n"
	       "capable (it doesn't have to be), then isochronous \n"
	       "transactions will not work.  The host controller card is \n"
	       "cycle master capable, however, most cameras are not.\n"
	       "\n"
	       "The quick solution is to add the parameter \n"
	       "attempt_root=1 when loading the OHCI driver as a \n"
	       "module.  So please do (as root):\n"
	       "\n"
	       "   rmmod ohci1394\n"
	       "   insmod ohci1394 attempt_root=1\n"
	       "\n"
	       "for more information see the FAQ at \n"
	       "http://linux1394.sourceforge.net/faq.html#DCbusmgmt\n"
	       "\n");
		raw1394_destroy_handle(mHandle);
		mHandle=NULL;
		return;

	}

	/*-----------------------------------------------------------------------
	*  setup capture
	*-----------------------------------------------------------------------*/
	if (dc1394_dma_setup_capture(mHandle,mCameraNodes[0],channel,format,
		mode,firewire_speed ,framerate,num_dma_buffers, drop_frames,deviceName.c_str(),&mCamera)!=DC1394_SUCCESS) {
		fprintf( stderr,"unable to setup camera make sure that the video mode,framerate and format are supported by your camera\n");
		dc1394_dma_release_camera(mHandle,&mCamera);
		raw1394_destroy_handle(mHandle);
		mHandle=NULL ;
		return;
	} 
	mpImage = (unsigned char*)malloc(mCamera.dma_frame_size);
  	#ifdef DEBUG
		printf("capture setup executed \n");
	#endif
}

bool FireIAcquisition::isReady(){
	return (mHandle!=NULL);
}

FireIAcquisition::~FireIAcquisition()
{
	if(mHandle!=NULL){
		dc1394_dma_release_camera(mHandle,&mCamera);
		raw1394_destroy_handle(mHandle);
	}
}

void FireIAcquisition::Start()
{
  /*-----------------------------------------------------------------------
   *  have the camera start sending us data
   *-----------------------------------------------------------------------*/
	if (dc1394_start_iso_transmission(mHandle,mCamera.node)!=DC1394_SUCCESS){
		fprintf( stderr, "unable to start camera iso transmission\n");
		dc1394_dma_release_camera(mHandle,&mCamera);
		raw1394_destroy_handle(mHandle);
		exit(1);
	}
  	#ifdef DEBUG
		printf("iso_transmission_started \n");
	#endif

}

void FireIAcquisition::Stop()
{
	/*-----------------------------------------------------------------------
	*  Stop data transmission
	*-----------------------------------------------------------------------*/
	if (dc1394_stop_iso_transmission(mHandle,mCamera.node)!=DC1394_SUCCESS) {
	#ifdef DEBUG
		printf("couldn't stop the camera?\n");
	#endif
	}
}

unsigned char* FireIAcquisition::CaptureFrame()
{
	/*-----------------------------------------------------------------------
	*  capture one frame
	*-----------------------------------------------------------------------*/
	if (dc1394_dma_single_capture(&mCamera)!=DC1394_SUCCESS) 
	{
		fprintf( stderr, "unable to capture a frame\n");
		dc1394_dma_release_camera(mHandle,&mCamera);
		raw1394_destroy_handle(mHandle);
		exit(1);
	}
	memcpy( mpImage, mCamera.capture_buffer, mCamera.dma_frame_size );
	dc1394_dma_done_with_buffer(&mCamera);

	if(saveFrame){
		char s[50];
		char fs[50];
		strcpy(fs,prefix.c_str());
		strcat(fs,iformat.c_str());
		strcat(fs,suffix.c_str());
		sprintf(s,fs,count);
		FILE *f=fopen(s,"w");
		fwrite("P5\n640 480 255\n",15,1,f);
		fwrite(mpImage,640*480,1,f);		
		fclose(f);
	
	}
	count++;
	return (unsigned char*) mpImage;
}
