/*
 * AIRboardUI: graphical interface program to interact with airboard
 * through RS232 serial port.
 *
 * In order to compile this package you have to install the
 * Java Communications API on your system.
 * http://java.sun.com/products/javacomm/downloads/index.html
 * 
 * Bruschieri Stefano
 * Campari PierPaolo
 */

package airboardUI;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.awt.event.*;
import javax.swing.Timer;
import javax.comm.*;
import java.io.*;

/**
 * @author Stefano
 *
 *  This class creates two arrays from two files: <code>first_input.dat</code>
 *  and <code>second_input.dat</code>. From these arrays one element each time 
 *  is transmitted via serial port.
 */
public class FuzzyTable {
	private final int TABLE_LENGTH = 101;

    private AirboardFrame parent;
    
    private Integer[] first_table = new Integer[101];
    private Integer[] second_table = new Integer[101];
    
    private SerialConnection connection;
    private int MAX_ERRORS = 5;
    private int MAX_NO_RESPONSE = 5;
    
    ProgressDialog pd;
    int i=0, k=0;
    Timer timer;
	int noResponseCounter = 0;
	int errorsCounter = 0;
    
    /**
    * Creates a new <code>FuzzyTable</code>: constructor loads data from files.
    *
    * @param parent The main window.
    * @param connection The enabled serial connection.
    */
    public FuzzyTable(AirboardFrame parent, SerialConnection connection)throws IOException {

    	this.parent = parent;
    	this.connection = connection;	
    	File first_inputFile = null;
    	FileInputStream first_in = null;
    	File second_inputFile = null;
    	FileInputStream second_in = null; 
		
    	// Load data from files
        first_inputFile = new File("airboardUI/data/first_input.dat" /*AirboardFrame.FIRST_INPUT_FILE*/);
        first_in = new FileInputStream(first_inputFile);
        second_inputFile = new File("airboardUI/data/second_input.dat" /*AirboardFrame.SECOND_INPUT_FILE*/);
        second_in = new FileInputStream(second_inputFile);
    	
        int first_c, second_c, i=0, init=0;
        byte[] cbuf = new byte[(int)first_inputFile.length()];
        byte[] temp;
            	
        first_c = first_in.read(cbuf);
        //if(c != -1 ) System.out.println("Non va bene");
        
        for(int j=0; j<(int)first_inputFile.length(); j++){
        	if( cbuf[j] == '\n'){ 
        		temp = new byte[(j-init)];
        		for(int k=0; k<(j-init-1); k++){
        			temp[k] = cbuf[init+k];
        		}
        		String stringa = new String(temp);
        		this.first_table[i] = Integer.valueOf(stringa.substring(0,(stringa.length()-2)));
        		i++;
        		init = (j+1);
        	}
        }
        first_in.close();
           
        /* Lettura secondo file */
        i = 0;
        init = 0;
        cbuf = new byte[(int)second_inputFile.length()];
        second_c = second_in.read(cbuf);
        for(int j=0; j<(int)second_inputFile.length(); j++){
        	if( cbuf[j] == '\n'){ 
        		temp = new byte[(j-init)];
        		for(int k=0; k<(j-init-1); k++){
        			temp[k] = cbuf[init+k];
        		}
        		String stringa = new String(temp);
        		this.second_table[i] = Integer.valueOf(stringa.substring(0,(stringa.length()-2)));
        		i++;
        		init = (j+1);
        	}
        }
        second_in.close();
    }
    
    public void setConnection(SerialConnection connection){
    	this.connection = connection;
    }
    
    public void transmitTable(){

    	pd = new ProgressDialog(this.parent, this.TABLE_LENGTH*2 );
 
    	
    	/* timer */
    	timer = new Timer(AirboardFrame.DELAY_TIME, new ActionListener() {		
    		public void actionPerformed(ActionEvent evt) {
    			int mf = 1;
    			SerialPort sPort;
    			OutputStream os;
    			
    			if(!pd.hasBeenStopped() && 
    					i<=(TABLE_LENGTH*2-1)){
    		    	pd.progressBar.setValue(i);
    		    	if( i>100 ) {
    		    		pd.setText("Copying second table");
    		    		mf = 2;
    		    	}
    		    	
    		    	// incremento gli errori
    		    	String lastInput = (connection.getLastInputString());
    		    	if( lastInput.length()>0){
    		    		if(lastInput.lastIndexOf("ERROR")!=-1 ){
    		    			//if( k==0 ) k=1; // 
    		    			errorsCounter++;
    		    			//return;
    		    			System.out.println(i+" An error was occurred "/*+connection.getLastInputString()*/);
    						if(errorsCounter >= MAX_ERRORS) { 
    							pd.setCancelled("AIRLab returns error"); 
    							pd.setCompleted(); 
    						}
    						return;
    		    		} else { 
    		    			/* incremento i solo se non ci sono stati errori */
    	    		    	i++;
    	    		    	/* k is the array counter*/
    	    		    	if( i==102 )  {
    	    		    		k=1;
    	    		    	}
    	    		    	else k++;
    		    			System.out.println(i+" OK"); 
    		    			}
    		    	} else { 
    		    		noResponseCounter++;
    		    		System.out.println("No response");
						if(noResponseCounter >= MAX_NO_RESPONSE) { 
							pd.setCancelled("No response"); 
							pd.setCompleted(); 
						}
						return;
    		    	}
    		    	
    		    	String toOutput;
    		    	if( mf == 1 )
    		    		toOutput = first_table[(k-1)].toString();
    		    	else 
    		    		toOutput = second_table[(k-1)].toString();
    		    	if(toOutput.length()==1) toOutput = "00"+toOutput;
    		    	if(toOutput.length()==2) toOutput = "0"+toOutput;
    		    	if(toOutput.length()==0 || toOutput.length()>3) 
    		    		System.out.println("Error");
    		    	String end;
    		    	end = String.valueOf((k-1));
    		    	if(end.length()==1) end = "00"+end;
    		    	if(end.length()==2) end = "0"+end;
    		    	if(end.length()==0 || end.length()>3) 
    		    		System.out.println("Error");
    		    	toOutput = "ewfuzzymf"+mf+"=+"+toOutput+"&+"+end;
    		    	System.out.println(toOutput);

    		    	connection.sendAString(toOutput);
    		    } else { 
    		    	pd.setCompleted();
    		    	timer.stop(); 
    		    }}
    		});

    		timer.start();
    }
    
}
